/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Dir_H
#define Dir_H

#include <sys/stat.h>
#include <sys/types.h>

#include <dirent.h>

#include "File.h"

#define DirDotChar               '.'
#define DirDotString             "."

#ifndef WIN32

#define DirSlashChar             '/'
#define DirSlashString           "/"

#define DirDotRootString         "/.nx"
#define DirSlashCDashString      "/C-"
#define DirSlashSDashString      "/S-"
#define DirSlashCacheDashString  "/cache-"
#define DirSlashCacheString      "/cache"
#define DirSlashImagesString     "/images"

#else

#define DirSlashChar             '\\'
#define DirSlashString           "\\"

#define DirDotRootString         "\\.nx"
#define DirSlashCDashString      "\\C-"
#define DirSlashSDashString      "\\S-"
#define DirSlashCacheDashString  "\\cache-"
#define DirSlashCacheString      "\\cache"
#define DirSlashImagesString     "\\images"

#endif

typedef DIR DirHandle;

typedef struct dirent DirEntry;
typedef struct stat   DirStats;

#define DirOpen(name)     ::opendir((name))
#define DirRead(handle)   ::readdir((handle))

#define DirRewind(handle) ::rewinddir((handle))
#define DirClose(handle)  ::closedir((handle))

//
// Obtain information about the directory.
// Same function used for normal files.
//

inline Int DirQuery(const Char *name, DirStats *stats)
{
  return FileQuery(name, stats);
}

//
// Create a directory with the given access
// permissions.
//

#ifdef WIN32

Int DirCreate(const Char *name, Int mode);

#else

#define DirCreate(name, mode)  ::mkdir((name), (mode))

#endif

//
// Delete the directory from the filesystem.
//

#define DirRemove(name)  ::rmdir((name))

//
// Change access permissions.
//

#define DirMode(name, mode)  ::chmod((name), (mode))
#define DirMask(name)        ::umask((mode))

//
// Return 1 if the directory exists and is
// actually a directory, 0 if it doesn't,
// -1 if the prefix directory was not found
// or if access to the directory failed.
//

Int DirIsDirectory(const Char *path, const Char *name);

//
// Same as the previous function but the name
// contains the full path.
//

inline Int DirIsDirectory(const Char *name)
{
  return DirIsDirectory(NULL, name);
}

//
// Check if the file or directory exists.
//

inline Int DirIsEntity(const Char *directory, const Char *name)
{
  return FileIsEntity(directory, name);
}

inline Int DirIsEntity(const Char *name)
{
  return FileIsEntity(name);
}

inline Int DirIsDirPrefix(const Char *name)
{
  return FileIsDirPrefix(name);
}

//
// Change ownership of the directory.
//

inline Int DirOwner(const Char *name, Int uid, Int gid)
{
  return FileOwner(name, uid, gid);
}

#endif /* Dir_H */
