/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Log.h"
#include "Error.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// Wrap the error reporting functions
// on Windows.
//
// By convention, we try to translate
// the Windows specific error numbers
// to the values reported for the same
// error on a standard Linux host.
//
// See:
//
// /usr/include/asm-generic/errno-base.h
// /usr/include/asm-generic/errno.h
//

#ifndef WIN32

const Char *GetErrorString()
{
  return GetErrorString(errno);
}

const Char *GetErrorString(Int error)
{
  #ifdef __APPLE__

  switch (error)
  {
    case EBADSLT:
    {
      return "Invalid slot";
    }
    case ECOMM:
    {
      return "Communication error on send";
    }
  }

  #endif

  return (error == 0 ? "No error" : strerror(error));
}

#ifdef __APPLE__

#include <mach/task.h>

//
// Translate a Mach kernel error into a
// standard error number.
//

Int GetKernelError(Int result)
{
  //
  // Not all possible error codes are handled.
  // See /usr/include/mach/kern_return.h for
  // reference.
  //

  switch (result)
  {
    case KERN_SUCCESS:
    {
      return 0;
    }
    case KERN_ABORTED:
    {
      return EINTR;
    }
    case KERN_OPERATION_TIMED_OUT:
    {
      return ETIMEDOUT;
    }
    case KERN_INVALID_NAME:
    {
      return ENOENT;
    }
    case KERN_INVALID_ARGUMENT:
    {
      return EINVAL;
    }
    default:
    {
      return result;
    }
  }
}

#endif // #ifdef __APPLE__ ...
 
#else // #ifndef WIN32 ...

const Char *GetErrorString()
{
  return (errno != 0 ? GetErrorString(errno) :
              SocketGetErrorString(SocketGetError()));
}

const Char *GetErrorString(Int error)
{
  if (error >= 255)
  {
    return SocketGetErrorString(error);
  }

  switch (error)
  {
    case ESRCH:
    {
      return "No such process";
    }
    case EACCES:
    {
      return "Permission denied";
    }
    case ENOMSG:
    {
      return "No message of desired type";
    }
    case EBADE:
    {
      return "Invalid exchange";
    }
    case EBADRQC:
    {
      return "Invalid request code";
    }
    case EBADSLT:
    {
      return "Invalid slot";
    }
    case EPROTO:
    {
      return "Protocol error";
    }
    case EBADMSG:
    {
      return "Not a data message";
    }
    case ENOTSUP:
    {
      return "Operation not supported on transport endpoint";
    }
    case ECONNABORTED:
    {
      return "Software caused connection abort";
    }
    case ECONNRESET:
    {
      return "Connection reset by peer";
    }
    case EISCONN:
    {
      return "Transport endpoint is already connected";
    }
    case ETIMEDOUT:
    {
      return "Connection timed out";
    }
    case ECONNREFUSED:
    {
      return "Connection refused";
    }
    case EADDRINUSE:
    {
      return "Address already in use";
    }
    case EINPROGRESS:
    {
      return "Operation now in progress";
    }
    case EADDRNOTAVAIL:
    {
      return "Address not available";
    }
    case ENOTSOCK:
    {
      return "Socket operation on non-socket";
    }
    case ECOMM:
    {
      return "Communication error on send";
    }
    case EFAULT:
    {
      return "Bad address";
    }
    default:
    {
      return strerror(error);
    }
  }
}

Int GetWaitError(Int result)
{
  if (result == WAIT_TIMEOUT)
  {
    return EAGAIN;
  }
  else if (result == WAIT_ABANDONED)
  {
    return EBADE;
  }
  else
  {
    return GetLastError();
  }
}

Int GetWindowsError()
{
  Int error = GetLastError();

  switch (error)
  {
    case ERROR_ACCESS_DENIED:
    {
      return EACCES;
    }
    case ERROR_INVALID_PARAMETER:
    {
      return EINVAL;
    }
    case ERROR_INVALID_FUNCTION:
    {
      return ENOSYS;
    }
    case ERROR_OPERATION_ABORTED:
    {
      return EINTR;
    }
    case ERROR_FILE_NOT_FOUND:
    {
      return ENOENT;
    }
    case ERROR_PATH_NOT_FOUND:
    {
      return ENOENT;
    }
    case ERROR_BROKEN_PIPE:
    {
      return EPIPE;
    }
    case ERROR_NO_DATA:
    {
      return EPIPE;
    }
    case ERROR_NOT_OWNER:
    {
      return ESRCH;
    }
    case ERROR_INVALID_ADDRESS:
    {
      return EFAULT;
    }
    default:
    {
      return error;
    }
  }
}

#endif // #ifndef WIN32 ... #else ...

const Char *GetErrorName()
{
  #ifdef WIN32

  return (errno != 0 ? GetErrorName(errno) :
              SocketGetErrorName(SocketGetError()));

  #else

  return GetErrorName(errno);

  #endif
}

const Char *GetErrorName(Int error)
{
  #ifdef WIN32

  if (error >= 255)
  {
    return SocketGetErrorName(error);
  }

  #endif

  switch (error)
  {
    case 0:
    {
      return "Success";
    }
    case EPERM:
    {
      return "EPERM";
    }
    case ENOENT:
    {
      return "ENOENT";
    }
    case ESRCH:
    {
      return "ESRCH";
    }
    case EINTR:
    {
      return "EINTR";
    }
    case EIO:
    {
      return "EIO";
    }
    case ENXIO:
    {
      return "ENXIO";
    }
    case E2BIG:
    {
      return "E2BIG";
    }
    case ENOEXEC:
    {
      return "ENOEXEC";
    }
    case EBADF:
    {
      return "EBADF";
    }
    case ECHILD:
    {
      return "ECHILD";
    }
    case EAGAIN:
    {
      return "EAGAIN";
    }
    case ENOMEM:
    {
      return "ENOMEM";
    }
    case EACCES:
    {
      return "EACCES";
    }
    case EFAULT:
    {
      return "EFAULT";
    }
    case EBUSY:
    {
      return "EBUSY";
    }
    case EEXIST:
    {
      return "EEXIST";
    }
    case EXDEV:
    {
      return "EXDEV";
    }
    case ENODEV:
    {
      return "ENODEV";
    }
    case ENOTDIR:
    {
      return "ENOTDIR";
    }
    case EISDIR:
    {
      return "EISDIR";
    }
    case EINVAL:
    {
      return "EINVAL";
    }
    case ENFILE:
    {
      return "ENFILE";
    }
    case EMFILE:
    {
      return "EMFILE";
    }
    case ENOTTY:
    {
      return "ENOTTY";
    }
    case EFBIG:
    {
      return "EFBIG";
    }
    case ENOSPC:
    {
      return "ENOSPC";
    }
    case ESPIPE:
    {
      return "ESPIPE";
    }
    case EROFS:
    {
      return "EROFS";
    }
    case EMLINK:
    {
      return "EMLINK";
    }
    case EPIPE:
    {
      return "EPIPE";
    }
    case EDOM:
    {
      return "EDOM";
    }
    case ERANGE:
    {
      return "ERANGE";
    }
    case EDEADLK:
    {
      return "EDEADLK";
    }

    //
    // Same as EDEADLK for compatibility
    // with older Microsoft C versions.
    //

    #if defined(EDEADLOCK) && (EDEADLOCK != EDEADLOCK)

    case EDEADLOCK:
    {
      return "EDEADLOCK";
    }

    #endif

    case ENAMETOOLONG:
    {
      return "ENAMETOOLONG";
    }
    case ENOLCK:
    {
      return "ENOLCK";
    }
    case ENOSYS:
    {
      return "ENOSYS";
    }
    case ENOTEMPTY:
    {
      return "ENOTEMPTY";
    }
    case EILSEQ:
    {
      return "EILSEQ";
    }

    //
    // Error is 'String was truncated' on
    // Windows, MinGW does not define it.
    //

    #ifdef STRUNCATE
    
    case STRUNCATE:
    {
      return "STRUNCATE";
    }

    #endif

    case ENOMSG:
    {
      return "ENOMSG";
    }
    case EBADE:
    {
      return "EBADE";
    }
    case EBADRQC:
    {
      return "EBADRQC";
    }
    case EBADSLT:
    {
      return "EBADSLT";
    }
    case EPROTO:
    {
      return "EPROTO";
    }
    case EBADMSG:
    {
      return "EBADMSG";
    }
    case ENOTSUP:
    {
      return "ENOTSUP";
    }
    case ECONNABORTED:
    {
      return "ECONNABORTED";
    }
    case ECONNRESET:
    {
      return "ECONNRESET";
    }
    case EISCONN:
    {
      return "EISCONN";
    }
    case ETIMEDOUT:
    {
      return "ETIMEDOUT";
    }
    case ECONNREFUSED:
    {
      return "ECONNREFUSED";
    }
    case EADDRINUSE:
    {
      return "EADDRINUSE";
    }
    case EINPROGRESS:
    {
      return "EINPROGRESS";
    }
    case EADDRNOTAVAIL:
    {
      return "EADDRNOTAVAIL";
    }
    case ENOTSOCK:
    {
      return "ENOTSOCK";
    }
    case ECOMM:
    {
      return "ECOMM";
    }
    default:
    {
      return "Unknown";
    }
  }
}

void SetValueError(const Char *name, Int value,
                       const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Bad value '" << value << "' in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Bad value '" << value << "' in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EINVAL);
}

void SetValueError(const Char *name, void *value,
                       const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Bad value '" << value << "' in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Bad value '" << value << "' in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EINVAL);
}

void SetValueError(const Char *name, const Char *value,
                       const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Bad value '" << value << "' in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Bad value '" << value << "' in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EINVAL);
}

void SetValueError(const Char *name, Int type, Int value,
                       const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Bad value '" << value << "' for type '"
        << type << "' in method '" << name << "::" << method
        << "()' context [" << context << "].\n";
  #endif

  LogError() << "Bad value '" << value << "' for type '" << type
             << "' in method '" << name << "::" << method
             << "()' context [" << context << "].\n";

  SetError(EINVAL);
}

void SetPointerError(const Char *name, const void *pointer,
                         const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Bad pointer '" << pointer << "' in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Bad pointer '" << pointer << "' in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EINVAL);
}

void SetOperationError(const Char *name, const Char *method,
                           const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Operation not permitted in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Operation not permitted in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EPERM);
}

void SetImplementationError(const Char *name, const Char *method,
                                const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Not implemented in method '" << name
        << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Not implemented in method '" << name << "::"
             << method << "()' context [" << context << "].\n";

  SetError(ENOSYS);
}

void SetImplementationError(const Char *name, Int value, const Char *method,
                                const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Not implemented in method '" << name
        << "::" << method << "()' context [" << context
        << "] value '" << value << "'.\n";
  #endif

  LogError() << "Not implemented in method '" << name << "::"
             << method << "()' context [" << context << "] value '"
             << value << "'.\n";

  SetError(ENOSYS);
}

void SetImplementationError(const Char *name, void *value, const Char *method,
                                const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Not implemented in method '" << name
        << "::" << method << "()' context [" << context
        << "] value '" << value << "'.\n";
  #endif

  LogError() << "Not implemented in method '" << name << "::"
             << method << "()' context [" << context << "] value '"
             << value << "'.\n";

  SetError(ENOSYS);
}

void SetAllocationError(const Char *name, const Char *method,
                            const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Memory allocation failed in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Memory allocation failed in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(ENOMEM);
}

void SetStateError(const Char *name, const Char *method,
                       const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Invalid state for object in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Invalid state for object in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EACCES);
}

void SetDescriptorError(const Char *name, const Char *method,
                            const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Bad descriptor state in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Bad descriptor state in method '" << name
             << "::" << method << "()' context [" << context << "].\n";

  SetError(EBADF);
}

void SetBufferError(const Char *name, const Char *method,
                        const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Buffer not found in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Buffer not found in method '" << name
             << "::" << method << "()' context [" << context << "].\n";

  SetError(ENXIO);
}

void SetOwnerError(const Char *name, const Char *method,
                       const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Invalid buffer owner in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Invalid buffer owner in method '" << name
             << "::" << method << "()' context [" << context << "].\n";

  SetError(EDEADLK);
}

void SetEventError(const Char *name, const Char *event,
                       const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Bad event '" << event << "' in method '"
        << name << "::" << method << "()' context [" << context
        << "].\n";
  #endif

  LogError() << "Bad event '" << event << "' in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EBADE);
}

void SetTypeError(const Char *name, const Char *type,
                      const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Invalid operation for type '" << type
        << "' in method '" << name << "::" << method << "()' context "
        << "[" << context << "].\n";
  #endif

  LogError() << "Invalid operation for type '" << type << "' in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(ENOTSUP);
}

void SetProtoError(const Char *name, const Char *proto,
                      const Char *method, const Char *context)
{
  #ifdef ERROR
  Log() << name << ": ERROR! Invalid operation for proto '" << proto
        << "' in method '" << name << "::" << method << "()' context "
        << "[" << context << "].\n";
  #endif

  LogError() << "Invalid operation for proto '" << proto << "' in method '"
             << name << "::" << method << "()' context [" << context
             << "].\n";

  SetError(EPROTO);
}
