/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Error_H
#define Error_H

#include <errno.h>
#include <string.h>

#include "Types.h"

//
// Depending on the platform, some errors
// may be missing. Use arbitrary values
// unused on all the platforms we support.
//

#define ERROR_USE_MISSING

#ifdef  ERROR_USE_MISSING

#define ERROR_MISSING_BASE  140

#ifndef EACCES
#define EACCES         ERROR_MISSING_BASE + 0
#endif

#ifndef ENOMSG
#define ENOMSG         ERROR_MISSING_BASE + 1
#endif

#ifndef EBADE
#define EBADE          ERROR_MISSING_BASE + 2
#endif

#ifndef EBADRQC
#define EBADRQC        ERROR_MISSING_BASE + 3
#endif

#ifndef EBADSLT
#define EBADSLT        ERROR_MISSING_BASE + 4
#endif

#ifndef EPROTO
#define EPROTO         ERROR_MISSING_BASE + 5
#endif

#ifndef EBADMSG
#define EBADMSG        ERROR_MISSING_BASE + 6
#endif

#ifndef ENOTSUP
#define ENOTSUP        ERROR_MISSING_BASE + 7
#endif

#ifndef ECONNABORTED
#define ECONNABORTED   ERROR_MISSING_BASE + 8
#endif

#ifndef ECONNRESET
#define ECONNRESET     ERROR_MISSING_BASE + 9
#endif

#ifndef EISCONN
#define EISCONN        ERROR_MISSING_BASE + 10
#endif

#ifndef ETIMEDOUT
#define ETIMEDOUT      ERROR_MISSING_BASE + 11
#endif

#ifndef ECONNREFUSED
#define ECONNREFUSED   ERROR_MISSING_BASE + 12
#endif

#ifndef EADDRINUSE
#define EADDRINUSE     ERROR_MISSING_BASE + 13
#endif

#ifndef EINPROGRESS
#define EINPROGRESS    ERROR_MISSING_BASE + 14
#endif

#ifndef EADDRNOTAVAIL
#define EADDRNOTAVAIL  ERROR_MISSING_BASE + 15
#endif

#ifndef ENOTSOCK
#define ENOTSOCK       ERROR_MISSING_BASE + 16
#endif

#ifndef ECOMM
#define ECOMM          ERROR_MISSING_BASE + 17
#endif

#endif // ERROR_USE_MISSING

//
// Error handling macros.
//

#ifndef WIN32

#define GetError()      (errno)
#define SetError(error) (errno = (error))

#ifdef __APPLE__

//
// Translate a Mach kernel error into a
// standard error number.
//

Int GetKernelError(Int result);

#endif // #ifdef __APPLE__ ...
 
#else // #ifndef WIN32 ...

#include "Socket.h"

#define GetError()      (errno != 0 ? errno : SocketGetError())
#define SetError(error) if ((error) < 255) { errno = (error); SocketSetError(0); } \
                        else { errno = 0; SocketSetError((error)); }

//
// Translate the result returned by any
// of the Windows wait functions into a
// standard error number.
//

Int GetWaitError(Int result);

//
// Translate the last Windows  error to
// the corresponding Posix error number.
//

Int GetWindowsError();

#endif // #ifndef WIN32 ... #else ...

const Char *GetErrorString();
const Char *GetErrorString(Int error);

//
// Return a conventional name for the
// error, like for example EINTR or
// EINVAL.
//

const Char *GetErrorName();
const Char *GetErrorName(Int error);

//
// Print the descriptive string of the
// system error.
//

#define Err(value) GetErrorString((value))

//
// Error reporting functions.
//

void SetValueError(const Char *name, Int value,
                       const Char *method, const Char *context);

void SetValueError(const Char *name, void *value,
                       const Char *method, const Char *context);

void SetValueError(const Char *name, const Char *value,
                       const Char *method, const Char *context);

void SetValueError(const Char *name, Int type, Int value,
                       const Char *method, const Char *context);

void SetPointerError(const Char *name, const void *pointer,
                         const Char *method, const Char *context);

void SetOperationError(const Char *name, const Char *method,
                           const Char *context);

void SetImplementationError(const Char *name, const Char *method,
                                const Char *context);

void SetImplementationError(const Char *name, Int value,
                                const Char *method, const Char *context);

void SetImplementationError(const Char *name, void *value,
                                const Char *method, const Char *context);

void SetAllocationError(const Char *name, const Char *method,
                            const Char *context);

void SetStateError(const Char *name, const Char *method,
                       const Char *context);

void SetDescriptorError(const Char *name, const Char *method,
                            const Char *context);

void SetBufferError(const Char *name, const Char *method,
                        const Char *context);

void SetOwnerError(const Char *name, const Char *method,
                       const Char *context);

void SetEventError(const Char *name, const Char *event,
                       const Char *method, const Char *context);

void SetTypeError(const Char *name, const Char *type,
                      const Char *method, const Char *context);

void SetProtoError(const Char *name, const Char *proto,
                      const Char *method, const Char *context);

#endif /* Error_H */
