/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Log.h"
#include "Dir.h"
#include "Error.h"
#include "String.h"
#include "File.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// Linux specific and present only
// on kernels since Linux 2.6.35.
//

#ifdef __linux__
#ifndef F_SETPIPE_SZ
#define F_SETPIPE_SZ  1031
#endif
#endif

Int FileStartup()
{
  //
  // Nothing to do on Windows or Unix-like
  // environments.
  //

  return 1;
}
  
Int FileCleanup()
{
  //
  // Nothing to do on Windows or Unix-like
  // environments.
  //

  return 1;
}

//
// Implement the file I/O operation.
//

Int FilePipe(Int fds[2], Int size)
{
  #ifndef WIN32

  if (pipe(fds) == -1)
  {
    return -1;
  }

  //
  // Set the size of the underlying
  // operating system buffer.
  //

  if (FileReceiveBuffer(fds[0], size) != -1)
  {
    FileSendBuffer(fds[1], size);
  }

  return 1;

  #else // #ifndef WIN32 ...

  //
  // Create the pipe with the given
  // size.
  //

  return ::_pipe(fds, size, _O_BINARY);

  #endif // #ifndef WIN32 ... #else ...
}

#ifdef WIN32

Int FileIdRead(Int id, void *buffer, Size size)
{
  DWORD total;

  if (ReadFile(HandleCast(id), buffer, size, &total, NULL) != 0)
  {
    return total;
  }

  //
  // Check if the I/O operation was
  // cancelled.
  //

  if (GetLastError() == ERROR_OPERATION_ABORTED)
  {
    SetError(EINTR);
  }
  else
  {
    FileTranslateError();
  }

  return -1;
}

Int FileIdWrite(Int id, const void *buffer, Size size)
{
  DWORD total;

  if (WriteFile(HandleCast(id), buffer, size, &total, NULL) != 0)
  {
    return total;
  }

  //
  // Check if the I/O operation was
  // cancelled.
  //

  if (GetLastError() == ERROR_OPERATION_ABORTED)
  {
    SetError(EINTR);
  }
  else
  {
    FileTranslateError();
  }

  return -1;
}

//
// Emulate the Unix scatter/gather
// I/O operation.
//

Int FileIdReadVector(Int id, const FileVector *vector, Size count)
{
  Char *base;

  Int length;
  Int result;

  Int total = 0;

  for (Size i = 0;  i < count;  i++, vector++)
  {
    length = vector -> iov_len;
    base = (Char *) vector -> iov_base;

    while (length > 0)
    {
      result = FileIdRead(id, base, length);

      if (result == -1 && total == 0)
      {
        return result;
      }
      else if (result <= 0)
      {
        return total;
      }

      length -= result;
      total  += result;
      base   += result;
    }
  }

  return total;
}

Int FileIdWriteVector(Int id, const FileVector *vector, Size count)
{
  Char *base;

  Int length;
  Int result;

  Int total = 0;

  for (Size i = 0;  i < count;  i++, vector++)
  {
    length = vector -> iov_len;
    base = (Char *) vector -> iov_base;

    while (length > 0)
    {
      result = FileIdWrite(id, base, length);

      if (result == -1 && total == 0)
      {
        return result;
      }
      else if (result <= 0)
      {
        return total;
      }

      length -= result;
      total  += result;
      base   += result;
    }
  }

  return total;
}

void FileTranslateError()
{
  errno = GetWindowsError();
}

#endif // #ifdef WIN32 ...

//
// File I/O options.
//

Int FileNoBlocking(Int fd, Int value)
{
  #ifdef WIN32

  #ifdef ERROR
  Log() << "File: ERROR! Can't set no-blocking "
        << "to " << value << " on FD#" << fd 
        << ".\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #else

  return SocketNoBlocking(fd, value);

  #endif
}

Int FileNoInherit(Int fd, Int value)
{
  #ifdef TEST
  Log() << "File: Setting no-inherit to " << value
        << " on FD#" << fd << ".\n";
  #endif

  #ifdef WIN32

  Int flag = (value != 0 ? 0 : HANDLE_FLAG_INHERIT);

  //
  // Convert the file descriptor to a
  // Windows handle.
  //

  Int id = FileGetId(fd);

  if (SetHandleInformation(HandleCast(id), HANDLE_FLAG_INHERIT, flag) != 0)
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "File: ERROR! Can't set no-inherit "
        << "to " << value << " on FD#" << fd 
        << ".\n";
  #endif

  #ifdef ERROR
  Log() << "File: ERROR! Error is " << GetError()
        << " " << Stk(GetErrorString()) << ".\n";
  #endif

  return -1;

  #else

  return SocketNoInherit(fd, value);

  #endif
}

//
// Internal implementation of
// the buffer size functions.
//

static Int FileBuffer(Int fd, Int size, const Char *what)
{
  #ifdef __linux__

  if (fcntl(fd, F_SETPIPE_SZ, size) != -1)
  {
    #ifdef TEST
    Log() << "File: Set " << what << " buffer "
          << "to " << size << " for FD#" << fd
          << ".\n";
    #endif

    return 1;
  }

  #else // #ifdef __linux__ ...

  SetError(ENOSYS);

  #endif // #ifdef__linux__ ... #else ...

  #ifdef TEST
  Log() << "File: WARNING! Can't set " << what
        << " buffer to " << size << " for FD#"
        << fd << ".\n";
  #endif

  #ifdef __linux__

  //
  // We may check the kernel version
  // or test the command using a get.
  //

  if (GetError() == EINVAL)
  {
    #ifdef TEST
    Log() << "File: WARNING! Control command "
          << "may not be implemented.\n";
    #endif

    return 0;
  }

  #ifdef TEST
  Log() << "File: WARNING! Error is " << GetError()
        << ", " << Stk(GetErrorString())
        << ".\n";
  #endif

  #endif // #ifdef __linux__ ...

  return -1;
}

Int FileReceiveBuffer(Int fd, Int size)
{
  return FileBuffer(fd, size, "receive");
}

Int FileSendBuffer(Int fd, Int size)
{
  return FileBuffer(fd, size, "send");
}

//
// Query the file buffer information.
//

Int FileBytesReadable(Int fd)
{
  #ifdef WIN32

  #ifdef WARNING
  Log() << "File: WARNING! Can't query the bytes "
        << "readable from FD#" << fd << ".\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #else

  return SocketBytesReadable(fd);

  #endif
}

Int FileBytesReadable(Int fd, Int *readable)
{
  #ifdef WIN32

  #ifdef WARNING
  Log() << "File: WARNING! Can't query the bytes "
        << "readable from FD#" << fd << ".\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #else

  return SocketBytesReadable(fd, readable);

  #endif
}

Int FileBytesWritable(Int fd)
{
  //
  // This operation currently depends on
  // being able to determine the buffer
  // size, only implemented for sockets.
  //

  #ifdef WARNING
  Log() << "File: WARNING! Can't query the bytes "
        << "writable to FD#" << fd << ".\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int FileBytesQueued(Int fd)
{
  #ifdef WIN32

  #ifdef WARNING
  Log() << "File: WARNING! Can't query the bytes "
        << "queued to FD#" << fd << ".\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #else

  return SocketBytesQueued(fd);

  #endif
}

Char *FilePath(const Char *directory, const Char *name)
{
  Char *fileName;

  if (name == NULL || *name == '\0')
  {
    return NULL;
  }
  else if (directory != NULL && *directory != '\0')
  {
    fileName = new Char[StringLength(directory) +
                            StringLength(DirSlashString) +
                                StringLength(name) + 1];

    StringCopy(fileName, directory);
    StringCat(fileName, DirSlashString);
    StringCat(fileName, name);
  }
  else
  {
    fileName = new Char[StringLength(name) + 1];

    StringCopy(fileName, name);
  }

  return fileName;
}

Int FileIsRegular(const FileStats *stats)
{
  if (S_ISREG(stats -> st_mode) != 0)
  {
    #ifdef TEST
    Log() << "File: Entity is a regular file.\n";
    #endif

    return 1;
  }
  else if (S_ISDIR(stats -> st_mode) != 0)
  {
    #ifdef TEST
    Log() << "File: Entity is a directory.\n";
    #endif

    return 0;
  }
  else
  {
    #ifdef TEST
    Log() << "File: Entity not a regular file.\n";
    #endif

    return -1;
  }
}

Int FileIsDirectory(const FileStats *stats)
{
  if (S_ISDIR(stats -> st_mode) != 0)
  {
    #ifdef TEST
    Log() << "File: File is a directory.\n";
    #endif

    return 1;
  }
  else
  {
    #ifdef TEST
    Log() << "File: File is not a directory.\n";
    #endif

    return 0;
  }
}

Int FileIsRegular(const Char *directory, const Char *name)
{
  Int result;

  FileStats fileStats;

  Char *fileName = FilePath(directory, name);

  #ifdef TEST
  Log() << "File: Checking whether " << Stk(fileName)
        << " is a regular file.\n";
  #endif

  result = FileQuery(fileName, &fileStats);

  if (result > 0)
  {
    //
    // Check if the path is actually a 
    // file.
    //

    if (FileIsRegular(&fileStats) == 1)
    {
      result = 1;
    }
    else
    {
      result = 0;
    }
  }

  delete [] fileName;

  return result;
}

Int FileIsDirectory(const Char *directory, const Char *name)
{
  Int result;

  FileStats fileStats;

  Char *fileName = FilePath(directory, name);

  #ifdef TEST
  Log() << "File: Checking whether " << Stk(fileName)
        << " is a directory.\n";
  #endif

  result = FileQuery(fileName, &fileStats);

  if (result > 0)
  {
    //
    // Check if the path is actually a 
    // file.
    //

    if (FileIsDirectory(&fileStats) == 1)
    {
      result = 1;
    }
    else
    {
      result = 0;
    }
  }

  delete [] fileName;

  return result;
}

Int FileIsEntity(const Char *directory, const Char *name)
{
  Int result;

  FileStats fileStats;

  Char *fileName = FilePath(directory, name);

  #ifdef TEST
  Log() << "File: Checking existence of file "
        << Stk(fileName) << ".\n";
  #endif

  result = FileQuery(fileName, &fileStats);

  delete [] fileName;

  return result;
}

Int FileIsAccessible(const Char *directory, const Char *name, Int mode)
{
  Int result;

  Char *fileName = FilePath(directory, name);

  #ifdef TEST
  Log() << "File: Checking accessibility of file "
        << Stk(fileName) << " with mode "
        << Itk(mode) << ".\n";
  #endif

  result = access(fileName, mode);

  if (result == 0)
  {
    result = 1;
  }
  else
  {
    result = 0;
  }
  
  delete [] fileName;

  return result;
}

//
// Implement the fgets() and fputs() functions
// using a file descriptor instead of a stream
// file handle.
//

Char *FileGet(Int fd, Char *string, Int size)
{
  Char *start = string;

  Int result;

  #ifdef TEST
  Log() << "File: Reading string with size "
        << size << " from FD#" << fd << ".\n";
  #endif

  while (size > 1)
  {
    result = FileRead(fd, string, 1);

    if (result == 1)
    {
      if (*string == '\n')
      {
        #ifdef TEST
        Log() << "File: Read newline from FD#" << fd
              << " with " << string - start
              << " in the buffer.\n";
        #endif

        string++;

        break;
      }

      #ifdef TEST
      Log() << "File: Read more from FD#" << fd
            << " with " << string - start
            << " in the buffer.\n";
      #endif

      string++;
      size--;
    }
    else
    {
      if (result < 0)
      {
        if (GetError() == EAGAIN)
        {
          #ifdef TEST
          Log() << "File: Read from FD#" << fd
                << " would block.\n";
          #endif

          break;
        }
        else if (GetError() == EINTR)
        {
          #ifdef TEST
          Log() << "File: Read from FD#" << fd
                << " was interrupted.\n";
          #endif

          continue;
        }
      }

      #ifdef TEST
      Log() << "File: Read from FD#" << fd
            << " failed.\n";
      #endif

      if (result == 0)
      {
        SetError(EIO);
      }

      break;
    }
  }

  if (string - start > 0)
  {
    #ifdef TEST
    Log() << "File: Returning " << string - start
          << " bytes read from FD#" << fd
          << ".\n";
    #endif

    *string = '\0';

    return string;
  }
  else
  {
    #ifdef TEST
    Log() << "File: Returning no data read from "
          << "FD#" << fd << ".\n";
    #endif

    return NULL;
  }
}

Int FilePut(Int fd, const Char *string)
{
  Int size = StringLength(string);

  const Char *start = string;

  Int result;

  #ifdef TEST
  Log() << "File: Writing string with size "
        << size << " to FD#" << fd << ".\n";
  #endif

  while (size > 0)
  {
    result = FileWrite(fd, string, size);

    if (result > 0)
    {
      #ifdef TEST
      Log() << "File: Written " << result << " bytes "
            << "to FD#" << fd << ".\n";
      #endif

      string += result;
      size   -= result;
    }
    else if (result < 0)
    {
      if (GetError() == EAGAIN)
      {
        #ifdef TEST
        Log() << "File: Write to FD#" << fd
              << " would block.\n";
        #endif

        break;
      }
      else if (GetError() == EINTR)
      {
        #ifdef TEST
        Log() << "File: Write to FD#" << fd
              << " was interrupted.\n";
        #endif

        continue;
      }
      else
      {
        #ifdef TEST
        Log() << "File: Write to FD#" << fd
              << " failed.\n";
        #endif

        if (string - start > 0)
        {
          #ifdef TEST
          Log() << "File: Delaying failure with "
                << string - start << " bytes written.\n";
          #endif

          break;
        }

        return -1;
      }
    }
    else
    {
      #ifdef TEST
      Log() << "File: No data written to FD#" << fd
            << " after " << string - start << " bytes "
            << "with " << size << " remaining.\n";
      #endif

      break;
    }
  }

  #ifdef TEST
  Log() << "File: Returning " << string - start
        << " bytes written to FD#" << fd
        << ".\n";
  #endif

  return string - start;
}

Long Long FileSize(const Char *directory, const Char *name)
{
  Long Long result;

  FileStats fileStats;

  Char *fileName = FilePath(directory, name);

  #ifdef TEST
  Log() << "File: Checking size of file "
        << Stk(fileName) << ".\n";
  #endif

  result = FileQuery(fileName, &fileStats);

  if (result > 0)
  {
    //
    // Check if the path is actually a 
    // file.
    //

    if (FileIsRegular(&fileStats) == 1)
    {
      result = fileStats.st_size;

      #ifdef TEST
      Log() << "File: Returning a size of "
            << result << " bytes.\n";
      #endif
    }
    else
    {
      SetError(EINVAL);

      result = -1;
    }
  }

  delete [] fileName;

  return result;
}

Int FileQuery(const Char *directory, const Char *name, FileStats *stats)
{
  Char *fileName = FilePath(directory, name);

  Int result = FileQuery(fileName, stats);

  delete [] fileName;

  return result;
}

Int FileQuery(const Char *name, FileStats *stats)
{
  #ifdef TEST
  Log() << "File: Getting stats of entity "
        << Stk(name) << ".\n";
  #endif

  Int result = stat(name, stats);

  if (result == -1)
  {
    if (GetError() == ENOENT)
    {
      #ifdef TEST
      Log() << "File: Entity " << Stk(name)
            << " does not exist.\n";
      #endif

      result = 0;
    }
    else
    {
      #ifdef TEST
      Log() << "File: Can't access entity "
            << Stk(name) << ". Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      result = -1;
    }
  }
  else
  {
    #ifdef TEST
    Log() << "File: Got stats for entity "
          << Stk(name) << ".\n";
    #endif

    result = 1;
  }

  return result;
}

//
// Change the file ownership.
//

Int FileOwner(const Char *name, Int uid, Int gid)
{
  #ifndef WIN32

  if (chown(name, (uid_t) uid, (gid_t) gid) == 0)
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "File: ERROR! Can't change ownership "
        << "of " << Stk(name) << ".\n";
  #endif

  #ifdef ERROR
  Log() << "File: ERROR! Error is " << GetError()
        << ", " << GetErrorString() << ".\n";
  #endif

  #else // #ifndef WIN32 ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function FileOwner() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif // #ifndef WIN32 ... #else ...

  return -1;
}

Int FileIsDirPrefix(const Char *name)
{
  #if defined(__CYGWIN32__) || defined(WIN32)

  //
  // Try to determine if the file name contains
  // a directory specification. On Windows check
  // if a drive letter was given. This is flaky,
  // at least, and may not work.
  //

  if (StringLength(name) >= 2 && CharIsAlpha(*name) && *(name + 1) == ':')
  {
    return 1;
  }

  #endif

  return (*name == DirSlashChar ||
              (StringLength(name) > 3 && *name == DirDotChar &&
                   *(name + 1) == DirDotChar &&
                       *(name + 2) == DirSlashChar));
}
