/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef File_H
#define File_H

#ifdef WIN32

#include <io.h>

#ifdef ERROR
#undef ERROR
#endif

#else

#ifdef __linux__
#ifndef _LARGEFILE64_SOURCE
#define _LARGEFILE64_SOURCE
#endif
#define _FILE_OFFSET_BITS 64
#endif

#include <unistd.h>

#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <fcntl.h>

#ifdef _MSC_VER

#include <sys/utime.h>

#else

#include <utime.h>

#endif

#include "Socket.h"

//
// Initialize the file I/O stack.
//

Int FileStartup();
Int FileCleanup();

typedef struct stat    FileStats;
typedef struct utimbuf FileTime;

//
// Portable file open modes.
//

#ifdef WIN32

#define OpenModeRead         _O_RDONLY
#define OpenModeWrite        _O_WRONLY
#define OpenModeReadWrite    _O_RDWR

#define OpenModeRandom       _O_RANDOM
#define OpenModeSequential   _O_SEQUENTIAL

#define OpenModeAppend       _O_APPEND
#define OpenModeCreate       _O_CREAT
#define OpenModeTruncate     _O_TRUNC
#define OpenModeExclusive    _O_EXCL
#define OpenModeLarge         0

#define OpenModeText         _O_TEXT
#define OpenModeBinary       _O_BINARY

#define OpenModeNoBlocking    0
#define OpenModeNoInherit    _O_NOINHERIT

#else

#define OpenModeRead         O_RDONLY
#define OpenModeWrite        O_WRONLY
#define OpenModeReadWrite    O_RDWR

#define OpenModeRandom       0
#define OpenModeSequential   0

#define OpenModeAppend       O_APPEND
#define OpenModeCreate       O_CREAT
#define OpenModeTruncate     O_TRUNC
#define OpenModeExclusive    O_EXCL

#ifndef __APPLE__
#define OpenModeLarge        O_LARGEFILE
#else
#define OpenModeLarge        0
#endif

#define OpenModeText         0
#define OpenModeBinary       0

#define OpenModeNoBlocking   O_NONBLOCK

#ifdef O_CLOEXEC
#define OpenModeNoInherit    O_CLOEXEC
#else
#define OpenModeNoInherit    0
#endif

#endif

//
// File creation permission masks.
//

#ifdef WIN32

#define FileUserAll          _S_IREAD | _S_IWRITE
#define FileUserRead         _S_IREAD
#define FileUserWrite        _S_IWRITE
#define FileUserReadWrite    _S_IREAD | _S_IWRITE
#define FileUserExecute       0

#define FileGroupAll          0
#define FileGroupRead         0
#define FileGroupWrite        0
#define FileGroupReadWrite    0
#define FileGroupExecute      0

#define FileOthersAll         0
#define FileOthersRead        0
#define FileOthersWrite       0
#define FileOthersReadWrite   0
#define FileOthersExecute     0

#else

#define FileUserAll          S_IRWXU
#define FileUserRead         S_IRUSR
#define FileUserWrite        S_IWUSR
#define FileUserReadWrite    S_IRUSR | S_IWUSR
#define FileUserExecute      S_IXUSR

#define FileGroupAll         S_IRWXG
#define FileGroupRead        S_IRGRP
#define FileGroupWrite       S_IWGRP
#define FileGroupReadWrite   S_IRGRP | S_IWGRP
#define FileGroupExecute     S_IXGRP

#define FileOthersAll        S_IRWXO
#define FileOthersRead       S_IROTH
#define FileOthersWrite      S_IWOTH
#define FileOthersReadWrite  S_IROTH | S_IWOTH
#define FileOthersExecute    S_IXOTH

#endif

//
// File descriptors manipulation
// functions.
//

#ifdef WIN32

#define FileOpen(path, mode, permissions)  ::_open((path), (mode), (permissions))

#define FileDuplicate(fd)   ::_dup((fd))
#define FileClone(fd1, fd2) ::_dup2((fd1), (fd2))

#define FileClose(fd)       ::_close((fd))
#define FileEof(fd)         ::_eof((fd))

#else

#define FileOpen(path, mode, permissions)  ::open((path), (mode), (permissions))

#define FileDuplicate(fd)   ::dup((fd))
#define FileClone(fd1, fd2) ::dup2((fd1), (fd2))

#define FileClose(fd)       ::close((fd))
#define FileEof(fd)         ::eof((fd))

#endif

//
// Open the pipe in binary mode using
// the requested buffer size.
// 

Int FilePipe(Int fds[2], Int size);

//
// Return the operating system's handle
// associated to the file descriptor.
//

#ifdef WIN32

inline Int FileGetId(Int fd)
{
  HANDLE handle = HandleCast(::_get_osfhandle(fd));

  return (handle == INVALID_HANDLE_VALUE ? fd : IntCast(handle));
}

#else

#define FileGetId(fd)  (fd)

#endif

//
// File I/O operations.
//

typedef SocketVector FileVector;

#ifdef WIN32

Int FileIdRead(Int id, void *buffer, Size size);
Int FileIdWrite(Int id, const void *buffer, Size size);

Int FileIdReadVector(Int id, const FileVector *vector, Size count);
Int FileIdWriteVector(Int id, const FileVector *vector, Size count);

#define FileRead(fd, buffer, size)   FileIdRead(FileGetId(fd), (buffer), (size))
#define FileWrite(fd, buffer, size)  FileIdWrite(FileGetId(fd), (buffer), (size))

#define FileReadVector(fd, vector, count)   FileIdReadVector(FileGetId(fd), (vector), (count))
#define FileWriteVector(fd, vector, count)  FileIdWriteVector(FileGetId(fd), (vector), (count))

//
// Map the Windows system error in
// the Posix error having the same
// meaning.
//

void FileTranslateError();

#else // #ifdef WIN32 ...

#define FileRead(fd, buffer, size)   ::read((fd), (buffer), (size))
#define FileWrite(fd, buffer, size)  ::write((fd), (buffer), (size))

#define FileReadVector(fd, vector, count)   ::readv((fd), (vector), (count))
#define FileWriteVector(fd, vector, count)  ::writev((fd), (vector), (count))

#endif // #ifdef WIN32 ... #else ...

//
// Setting a file in non-blocking mode is not
// supported on Windows. The function will set
// errno and will ignore the operation.
//

Int FileNoBlocking(Int fd, Int value);
Int FileNoInherit(Int fd, Int value);

//
// When supported, change the underlying size
// of the operating system buffer.
//

Int FileReceiveBuffer(Int fd, Int size);
Int FileSendBuffer(Int fd, Int size);

//
// Query the file buffer information. Will set
// errno and ignore the operation on Windows.
//

Int FileBytesReadable(Int fd);
Int FileBytesReadable(Int fd, Int *readable);
Int FileBytesWritable(Int fd);
Int FileBytesQueued(Int fd);

//
// The following macros must be available on
// the host operating system:
//
// SEEK_SET  The offset is set to offset bytes.
//
// SEEK_CUR  The offset is set to its current
//           location plus offset bytes.
//
// SEEK_END  The offset is set to the size of
//           the file plus offset bytes.
//

typedef off_t FileOffset;

#define FileSeekBegin    SEEK_SET
#define FileSeekCurrent  SEEK_CUR
#define FileSeekEnd      SEEK_END

#define FileSeek(fd, offset)  ::lseek((fd), (offset), SEEK_SET)
#define FileSkip(fd, offset)  ::lseek((fd), (offset), SEEK_CUR)

//
// Position to the beginning or to the end of
// the file.
//

#define FileBegin(fd)  ::lseek((fd), 0, SEEK_SET)
#define FileEnd(fd)    ::lseek((fd), 0, SEEK_END)

//
// Use the same interface as of the file
// handle based functions but use a plain
// file descriptor.
//

Char *FileGet(Int fd, Char *string, Int size);
Int   FilePut(Int fd, const Char *string);

//
// Return the size of the file or -1 if the
// file doesn't exist.
//

Long Long FileSize(const Char *directory, const Char *name);

//
// Same as the previous function but the name
// contains the full path.
//

inline Long Long FileSize(const Char *name)
{
  return FileSize(NULL, name);
}

//
// Load the stat structure for the file.
// Return 1 if the stats were copied, 0 if
// the file doesn't exist, -1 otherwise.
//

Int FileQuery(const Char *name, FileStats *stats);

//
// Concatenate path and name and get the
// information about the file.
//

Int FileQuery(const Char *directory, const Char *name, FileStats *stats);

//
// Combine path and name to create a fully
// qualified filename. The name is allocated
// in the heap and must be deallocated by
// the caller using delete [].
//

Char *FilePath(const Char *directory, const Char *name);

//
// Return 1 if the file exists, 0 if it doesn't,
// -1 if the file was not found or if trying to
// access the file or directory caused an error.
//

Int FileIsEntity(const Char *directory, const Char *name);

//
// Same as the previous function but the name
// contains the full path.
//

inline Int FileIsEntity(const Char *name)
{
  return FileIsEntity(NULL, name);
}

//
// Return 1 if the file exists and it is possible
// to access the file with the given mode, 0 if
// access would be denied, -1 if the file was not
// found. Testing the file for execute permissions
// is not available on Windows. The function will
// only test the file for readadability.
//

#ifdef WIN32

#define FileAccessExist        00
#define FileAccessRead         04
#define FileAccessWrite        02
#define FileAccessExecute      04

#else

#define FileAccessExist        F_OK
#define FileAccessRead         R_OK
#define FileAccessWrite        W_OK
#define FileAccessExecute      X_OK

#endif

Int FileIsAccessible(const Char *directory, const Char *name, Int mode);

//
// Same as the previous function but the name
// contains the full path.
//

inline Int FileIsAccessible(const Char *name, Int mode)
{
  return FileIsAccessible(NULL, name, mode);
}

//
// Return 1 if the file is a regular file,
// 0 if it is a directory -1 if it is any
// other type.
//

Int FileIsRegular(const FileStats *stats);

//
// Return 1 if the file is a directory or
// 0 otherwise.
//

Int FileIsDirectory(const FileStats *stats);

//
// Same functions accepting a file name
// instead of the file stats.
//

Int FileIsRegular(const Char *directory, const Char *name);

inline Int FileIsRegular(const Char *name)
{
  return FileIsRegular(NULL, name);
}

Int FileIsDirectory(const Char *directory, const Char *name);

inline Int FileIsDirectory(const Char *name)
{
  return FileIsDirectory(NULL, name);
}

//
// Delete the file from the filesystem.
//

#define FileRemove(name)  ::unlink((name))

//
// Rename the file.
//

#define FileRename(first, second)  ::rename((first), (second))

//
// Change access permissions.
//

#define FileMode(name, mode)  ::chmod((name), (mode))
#define FileMask(mode)        ::umask((mode))

//
// Change the file ownership.
//

Int FileOwner(const Char *name, Int uid, Int gid);

//
// Change the access and modification time
// of the file.
//

#define FileTouch(name, time)  ::utime((name), (time))

//
// Return true if the file name contains a
// directory specification.
//

Int FileIsDirPrefix(const Char *name);

//
// Define a handle as a file structure of
// the standard I/O library.
//

typedef FILE FileHandle;

//
// File handle based functions from the
// standard I/O library.
//

#define FileHandleOpen(name, mode)            ::fopen((name), (mode))
#define FileHandleReopen(path, mode, handle)  ::freopen((path), (mode), (handle))
#define FileHandleClose(handle)               ::fclose((handle))
#define FileHandleFlush(handle)               ::fflush((handle))
#define FileHandleFdOpen(fd, mode)            ::fdopen((fd), (mode))

//
// On Windows and Android can be a macro
// so don't use the namespace specifier.
//

#ifdef WIN32

#define FileHandleFd(handle)  _fileno((handle))

#else

#define FileHandleFd(handle)  fileno((handle))

#endif

//
// Read and write a single line of text
// using the standard I/O library.
//

#define FileHandleGet(handle, string, size)  ::fgets((string), (size), (handle))
#define FileHandlePut(handle, string)        ::fputs((string), (handle))

#endif /* File_H */
