/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Log.h"
#include "Error.h"
#include "Library.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

LibraryHandle LibraryOpen(const Char *path)
{
  #ifdef TEST
  Log() << "Library: Loading library " << Stk(path)
        << ".\n";
  #endif

  if (path == NULL)
  {
    #ifdef ERROR
    Log() << "Library: ERROR! Invalid libary path "
          << Stk(path) << ".\n";
    #endif

    SetError(EINVAL);

    return NULL;
  }

  LibraryHandle handle;

  #ifdef WIN32

  if ((handle = LoadLibrary(path)) == NULL)
  {
    Char message[1024] = { 0 };

    Int error = GetLastError();

    if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL, error, MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
                            (LPTSTR) message, 1024, NULL ) != 0)
    {
      if (error == ERROR_MOD_NOT_FOUND)
      {
        #ifdef TEST
        Log() << "Library: Could not load library "
              << Stk(path) << ".\n";

        Log() << "Library: Error is " << error << ", "
              << Stk(message) << ".\n";
        #endif

        SetError(ENOENT);
      }
      else
      {
        #ifdef WARNING
        Log() << "Library: WARNING! Could not load library "
              << Stk(path) << ".\n";

        Log() << "Library: WARNING! Error is " << error
              << ", " << Stk(message) << ".\n";
        #endif

        SetError(error);
      }
    }
    else
    {
      SetError(error);
    }
  }

  #else // #ifdef WIN32 ...

  if ((handle = dlopen(path, RTLD_LAZY)) == NULL)
  {
    //
    // We can't distinguish among the
    // different errors. Assume that
    // the library was not found.
    //

    #ifdef TEST

    const Char *message = dlerror();

    Log() << "Library: WARNING! Could not load "
          << "library " << Stk(path) << ".\n";

    Log() << "Library: WARNING! Error is "
          << Stk(message) << ".\n";

    #endif

    SetError(ENOENT);
  }

  #endif // #ifdef WIN32 ... #else ...

  return handle;
}

LibraryFunction LibrarySymbol(LibraryHandle handle, const Char *name)
{
  #ifdef TEST
  Log() << "Library: Loading symbol " << Stk(name)
        << ".\n";
  #endif

  LibraryFunction function;

  #ifdef WIN32

  if ((function = GetProcAddress(handle, name)) == NULL)
  {
    Int error = GetLastError();

    #ifdef WARNING

    Char message[1024] = { 0 };

    if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL, error, MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
                            (LPTSTR) message, 1024, NULL ) != 0)
    {
      Log() << "Library: WARNING! Could not resolve symbol "
            << Stk(name) << ".\n";

      Log() << "Library: WARNING! Error is " << error
            << ", " << Stk(message) << ".\n";
    }

    #endif

    SetError(error);
  }

  #else // #ifdef WIN32

  if ((function = dlsym(handle, name)) == NULL)
  {
    #ifdef WARNING

    const Char *message = dlerror();

    Log() << "Library: WARNING! Could not resolve symbol "
          << Stk(name) << ".\n";

    Log() << "Library: WARNING! Error is " << Stk(message)
          << "\n";

    #endif

    SetError(ENOMSG);
  }

  #endif // #ifdef WIN32

  return function;
}

Int LibraryClose(LibraryHandle handle)
{
  #ifdef TEST
  Log() << "Library: Closing library " << handle
        << ".\n";
  #endif

  #ifdef WIN32

  if (FreeLibrary(handle) == 0)
  {
    Int error = GetLastError();

    #ifdef WARNING

    Char message[1024] = { 0 };

    if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL, error, MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
                            (LPTSTR) message, 1024, NULL ) != 0)
    {
      Log() << "Library: WARNING! Could not free handle "
            << VoidCast(handle) << ".\n";

      Log() << "Library: WARNING! Error is " << error
            << ", " << Stk(message) << ".\n";
    }

    #endif

    SetError(error);

    return -1;
  }

  #else // #ifdef WIN32 ...

  if (dlclose(handle) != 0)
  {
    const Char *message = dlerror();

    #ifdef WARNING
    Log() << "Library: WARNING! Could not close library "
          << VoidCast(handle) << ".\n";

    Log() << "Library: WARNING! Error is " << Stk(message)
          << ".\n";
    #endif

    SetError(ENOMSG);

    return -1;
  }

  #endif // #ifdef WIN32

  return 0;
}

