/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef List_H
#define List_H

#include <list>
#include <algorithm>

#include "Types.h"

class List : private list < void * >
{
  //
  // Members are protected to force the
  // available lists to be specialized.
  //

  public:

  typedef list < void * >::iterator Iterator;

  typedef list < void * >::const_iterator ConstIterator;

  List()
  {
  }

  ~List()
  {
  }

  Int getSize() const
  {
    return size();
  }

  static Int getSize(List *list)
  {
    return (list == NULL ? 0 : list -> size());
  }

  List *copyList() const
  {
    return new List(*this);
  }

  Iterator findValue(const void *value)
  {
    return find(begin(), end(), value);
  }

  void addValue(void *value)
  {
    push_back(value);
  }

  void addValue(Iterator iterator)
  {
    push_back(*iterator);
  }

  void insertValue(void *value, Iterator iterator)
  {
    insert(iterator, value);
  }

  void moveValue(Iterator iterator, List *list)
  {
    list -> splice(list -> end(), *this, iterator);
  }

  void moveValue(void *value, List *list)
  {
    list -> splice(list -> end(), *this, find(begin(), end(), value));
  }

  void moveValue(Iterator value, Iterator iterator);

  void moveValue(void *value, Iterator iterator)
  {
    moveValue(find(begin(), end(), value), iterator);
  }

  void *getValue(Iterator iterator) const
  {
    return *iterator;
  }

  void *getValue(Int index) const;

  Iterator getBegin()
  {
    return begin();
  }

  Iterator getEnd()
  {
    return end();
  }

  ConstIterator getBegin() const
  {
    return begin();
  }

  ConstIterator getEnd() const
  {
    return end();
  }

  void *getFirst() const
  {
    return (size() > 0 ? *(begin()) : NULL);
  }

  void *getLast() const
  {
    return (size() > 0 ? *(--end()) : NULL);
  }

  void removeValue(Iterator iterator)
  {
    erase(iterator);
  }

  void removeValue(void *value);

  void removeValues()
  {
    erase(begin(), end());
  }

  void dumpList() const;

  void rotateValues(Iterator first, Iterator middle, Iterator last)
  {
    rotate(first, middle, last);
  }

  void rotateValues()
  {
    if (size() > 1)
    {
      rotateValues(begin(), ++begin(), end());
    }
  }
};

class VoidList : protected List
{
  public:

  typedef List::Iterator Iterator;
  typedef List::ConstIterator ConstIterator;

  VoidList()
  {
  }

  ~VoidList()
  {
  }

  Int getSize() const
  {
    return List::getSize();
  }

  Iterator getBegin()
  {
    return List::getBegin();
  }

  Iterator getEnd()
  {
    return List::getEnd();
  }

  ConstIterator getBegin() const
  {
    return List::getBegin();
  }

  ConstIterator getEnd() const
  {
    return List::getEnd();
  }

  VoidList *copyList() const
  {
    return new VoidList(*this);
  }

  Iterator findValue(const void *value)
  {
    return List::findValue(value);
  }

  void addValue(void *value)
  {
    List::addValue(value);
  }

  void insertValue(void *value, Iterator iterator)
  {
    List::insertValue(value, iterator);
  }

  void removeValue(void *value)
  {
    List::removeValue(value);
  }

  void removeValue(Iterator iterator)
  {
    List::removeValue(iterator);
  }

  void removeValues()
  {
    List::removeValues();
  }

  void *getValue(Iterator iterator) const
  {
    return *iterator;
  }

  void *getFirst() const
  {
    return List::getFirst();
  }

  void *getLast() const
  {
    return List::getLast();
  }

  void dumpList() const
  {
    List::dumpList();
  }
};

class IntList : protected List
{
  public:

  typedef List::Iterator Iterator;

  IntList()
  {
  }

  ~IntList()
  {
  }

  Int getSize() const
  {
    return List::getSize();
  }

  Iterator getBegin()
  {
    return List::getBegin();
  }

  Iterator getEnd()
  {
    return List::getEnd();
  }

  IntList *copyList() const
  {
    return new IntList(*this);
  }

  void addValue(Int value)
  {
    List::addValue(VoidCast(value));
  }

  void insertValue(Int value, Iterator iterator)
  {
    List::insertValue(VoidCast(value), iterator);
  }

  void removeValue(Int value)
  {
    List::removeValue(VoidCast(value));
  }

  void removeValue(Iterator iterator)
  {
    List::removeValue(iterator);
  }

  void removeValues()
  {
    List::removeValues();
  }

  Int getValue(Iterator iterator) const
  {
    return IntCast(*iterator);
  }

  Int getFirst() const
  {
    return (List::getSize() > 0 ? IntCast(List::getFirst()) : -1);
  }

  Int getLast() const
  {
    return (List::getSize() > 0 ? IntCast(List::getLast()) : -1);
  }

  void rotateValues(Iterator first, Iterator middle, Iterator last)
  {
    List::rotateValues(first, middle, last);
  }

  void rotateValues()
  {
    List::rotateValues();
  }
};

#endif /* List_H */
