/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Log.h"
#include "Time.h"
#include "String.h"
#include "Thread.h"
#include "Process.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// If defined, monitor the state
// of the errno variable.
//

#undef  LOG_USE_ERRNO

//
// Define this to run the program
// with Valgrind when the log can
// be generated handling a signal.
//

#undef  LOG_USE_VALGRIND

//
// Overloaded insertion operator.
//

#ifdef LOG_USE_OVERLOADED_INSERT

LogStream &LogStream::insert(const Char *string)
{
  #ifdef DEBUG

  ((OutputStream &) *this) << "<BEGIN>";

  #endif

  ((OutputStream &) *this) << string;

  #ifdef DEBUG

  ((OutputStream &) *this) << "<END>";

  #endif

  return *this;
}

#endif

//
// Generally available log facilities
// using the static system logger.
//

LogStream &Log()
{
  return Log(GetLogger());
}

LogStream &LogSession()
{
  return LogSession(GetLogger());
}

LogStream &LogError()
{
  return LogError(GetLogger());
}

LogStream &LogWarning()
{
  return LogWarning(GetLogger());
}

LogStream &LogInfo()
{
  return LogInfo(GetLogger());
}

LogStream &LogUser()
{
  return LogUser(GetLogger());
}

LogStream &LogMore()
{
  return LogMore(GetLogger());
}

LogStream &LogSessionMore()
{
  return LogSessionMore(GetLogger());
}

LogStream &LogErrorMore()
{
  return LogErrorMore(GetLogger());
}

LogStream &LogWarningMore()
{
  return LogWarningMore(GetLogger());
}

LogStream &LogInfoMore()
{
  return LogInfoMore(GetLogger());
}

LogStream &LogUserMore()
{
  return LogUserMore(GetLogger());
}

void LogFlush()
{
  LogFlush(GetLogger());
}

//
// Make a log header with the process id,
// the thread id and the current time.
//

#ifndef LOG_USE_VALGRIND

void LogHeader(Char *header, Int size)
{
  Char time[17];

  Int pid = ProcessGetPid();
  Int tid = ThreadGetPid();

  StringCopyMsTimeval(time, getTimeval());

  StringSizePrint(header, size - 1, "%d %d %s ", pid, tid, time);
}

#else // LOG_USE_VALGRIND

//
// Don't put the time in the header since
// this makes Valgrind unhappy. Note that,
// without Valgrind, everything works ok.
//

void LogHeader(Char *header, Int size)
{
  const Char *time = "--:--:-- ---.---";

  Int pid = ProcessGetPid();
  Int tid = ThreadGetPid();

  StringSizePrint(header, size - 1, "%d %d %s ", pid, tid, time);
}

#endif // LOG_USE_VALGRIND

//
// Internal implementation of the log
// facilities.
//

LogStream &Log(Logger *logger)
{
  LogStream *stream;

  if (logger -> isEnabled())
  {
    Char header[128];

    LogHeader(header, 128);

    Int exclusive = logger -> isExclusive();

    if (exclusive == 1)
    {
      logger -> lock();
    }

    stream = logger -> getLogOut();

    #ifdef LOG_USE_ERRNO
/*
FIXME: This variable is per-thread and must be
       moved to Thread Local Storage.
*/
    static SYSTEM Int error;

    if (GetError() != error)
    {
      if ((error = GetError()) != 0)
      {
        const Char *name = GetErrorName();

        const Char *string = GetErrorString();

        *stream << header << "Log: WARNING! Detected "
                << "error " << error << " " << name
                << ", " << Stk(string) << ".\n";
      }
    }

    #endif

    stream -> flush();

    *stream << header;

    *stream << dec;

    if (exclusive == 1)
    {
      logger -> unlock();
    }
  }
  else
  {
    stream = logger -> getNullOut();
  }

  return *stream;
}

LogStream &LogSession(Logger *logger)
{
  LogStream *stream;

  if (logger -> isEnabled())
  {
    logger -> lock();

    stream = logger -> getUserOut();

    stream -> flush();

    *stream << "Session: ";

    logger -> unlock();
  }
  else
  {
    stream = logger -> getNullOut();
  }

  return *stream;
}

LogStream &LogError(Logger *logger)
{
  LogStream *stream;

  if (logger -> isEnabled())
  {
    logger -> lock();

    stream = logger -> getUserOut();

    stream -> flush();

    *stream << "Error: ";

    logger -> unlock();
  }
  else
  {
    stream = logger -> getNullOut();
  }

  return *stream;
}

LogStream &LogWarning(Logger *logger)
{
  LogStream *stream;

  if (logger -> isEnabled())
  {
    logger -> lock();

    stream = logger -> getUserOut();

    stream -> flush();

    *stream << "Warning: ";

    logger -> unlock();
  }
  else
  {
    stream = logger -> getNullOut();
  }

  return *stream;
}

LogStream &LogInfo(Logger *logger)
{
  LogStream *stream;

  if (logger -> isEnabled())
  {
    logger -> lock();

    stream = logger -> getUserOut();

    stream -> flush();

    *stream << "Info: ";

    logger -> unlock();
  }
  else
  {
    stream = logger -> getNullOut();
  }

  return *stream;
}

LogStream &LogUser(Logger *logger)
{
  LogStream *stream;

  if (logger -> isEnabled())
  {
    logger -> lock();

    stream = logger -> getUserOut();

    stream -> flush();

    logger -> unlock();
  }
  else
  {
    stream = logger -> getNullOut();
  }

  return *stream;
}

//
// Append to the stream.
//

LogStream &LogMore(Logger *logger)
{
  LogStream *stream;

  if (logger -> isEnabled())
  {
    logger -> lock();

    stream = logger -> getLogOut();

    stream -> flush();

    logger -> unlock();
  }
  else
  {
    stream = logger -> getNullOut();
  }

  return *stream;
}

//
// Include the class name.
//

LogStream &Log(Logger *logger, const Char *name)
{
  LogStream &stream = Log(logger);

  if (logger -> isExtended())
  {
    stream << name << "/";
  }

  return stream;
}

//
// Return the default log stream for
// this system based on the buffering
// mode set in the logger.
//

LogStream &LogDefault()
{
  return *Logger::getDefaultOut();
}

//
// Safe versions not requiring a lock
// or a logger instance.
//

LogStream &LogSafe()
{
  Char header[128];

  LogHeader(header, 128);

  LogDefault().flush();

  LogDefault() << header;

  return LogDefault();
}

LogStream &LogSafeMore()
{
  LogDefault().flush();

  return LogDefault();
}

LogStream &LogSafeError()
{
  Char header[128];

  LogHeader(header, 128);

  LogDefault().flush();

  LogDefault() << header;

  LogDefault() << "Error: ";

  return LogDefault();
}

//
// Flush the log output.
//

void LogFlush(Logger *logger)
{
  logger -> lock();

  LogStream *stream;

  stream = logger -> getLogOut();

  *stream << dec;

  stream -> flush();

  stream = logger -> getUserOut();

  *stream << dec;

  stream -> flush();

  logger -> unlock();
}
