/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Log.h"
#include "File.h"
#include "String.h"
#include "Process.h"
#include "Logger.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// Default log mode.
//

#define LOGGER_DEFAULT_MODE  ModeSystem

//
// Default extended mode.
//

#define LOGGER_DEFAULT_EXTENDED   1

//
// Default buffering mode.
//

#define LOGGER_DEFAULT_BUFFERING  1

//
// Whether the exclusive mode is
// enabled by default.
//

#define LOGGER_DEFAULT_EXCLUSIVE  1

SYSTEM Logger logger_;

SYSTEM Mutex  Logger::mutex_;

Logger *GetLogger()
{
  return &logger_;
}

//
// The static logger mutex is defined in
// System.cpp or Io.cpp, to deal with the
// different initialization order of the
// static members on different platforms.
//

Logger::Logger()
{
  logOut_  = getDefaultOut();
  userOut_ = getDefaultOut();
  statOut_ = getDefaultOut();

  logOutSave_  = NULL;
  userOutSave_ = NULL;
  statOutSave_ = NULL;

  logOutName_  = NULL;
  userOutName_ = NULL;
  statOutName_ = NULL;

  nullOut_ = (LogStream *) new OutputFileStream();

  defaultBuffer_ = getDefaultOut() -> rdbuf();

  enabled_ = 1;

  mode_ = LOGGER_DEFAULT_MODE;

  extended_  = LOGGER_DEFAULT_EXTENDED;
  buffered_  = LOGGER_DEFAULT_BUFFERING;
  exclusive_ = LOGGER_DEFAULT_EXCLUSIVE;
}

Logger::~Logger()
{
  if (userOutName_ != NULL)
  {
    #ifdef TEST
    log() << "Logger: Restoring the standard error.\n";
    #endif

    getDefaultOut() -> rdbuf(defaultBuffer_);
  }

  #ifdef TEST
  log() << "Logger: Resetting all streams.\n";
  #endif

  resetStream(logOut_,  logOutSave_,  logOutName_); 
  resetStream(userOut_, userOutSave_, userOutName_); 
  resetStream(statOut_, statOutSave_, statOutName_); 

  #ifdef TEST
  log() << "Logger: Set streams to the standard error.\n";
  #endif

  delete logOutSave_;
/*
FIXME: This doesn't work on Windows.
*/
  #ifndef WIN32

  delete userOutSave_;

  #endif

  delete statOutSave_;

  #ifdef TEST
  log() << "Logger: Deleted all the saved streams.\n";
  #endif

  delete nullOut_;

  nullOut_ = getDefaultOut();

  #ifdef TEST
  log() << "Logger: Deleted the null output stream.\n";
  #endif
}

void Logger::setLogOut(const Char *directory, const Char *file)
{
  if (mode_ != ModeSystem)
  {
    #ifdef TEST
    log() << "Logger: Replacing the log output stream.\n";
    #endif

    openStream(logOut_, logOutSave_, logOutName_, directory, file);
  }
  #ifdef TEST
  else
  {
    log() << "Logger: Ignoring log output stream "
          << "with mode " << Itk(mode_) << ".\n";
  }
  #endif
}

void Logger::setUserOut(const Char *directory, const Char *file)
{
  if (mode_ != ModeSystem)
  {
    #ifdef TEST
    log() << "Logger: Replacing the user output stream.\n";
    #endif

    openStream(userOut_, userOutSave_, userOutName_, directory, file);

    //
    // Redirect the standard error to the file.
    //

    if (userOutName_ != NULL)
    {
      #ifdef TEST
      log() << "Logger: Redirecting the standard error to "
            << Stk(userOutName_) << ".\n";
      #endif

      getDefaultOut() -> rdbuf(userOut_ -> rdbuf());
    }

    LogDefault() << "Logger: Redirected the standard error to "
                 << Stk(userOutName_) << ".\n";
  }
  #ifdef TEST
  else
  {
    log() << "Logger: Ignoring user output stream "
          << "with mode " << Itk(mode_) << ".\n";
  }
  #endif
}

void Logger::setStatOut(const Char *directory, const Char *file)
{
  #ifdef TEST
  log() << "Logger: Replacing the stat output stream.\n";
  #endif

  openStream(statOut_, statOutSave_, statOutName_, directory, file);
}

void Logger::setLogOut(OutputStream *stream)
{
  if (mode_ != ModeSystem)
  {
    #ifdef TEST
    log() << "Logger: Setting the log output stream.\n";
    #endif

    setStream(logOut_, logOutSave_, logOutName_, stream);
  }
  #ifdef TEST
  else
  {
    log() << "Logger: Ignoring log output stream "
          << "with mode " << Itk(mode_) << ".\n";
  }
  #endif
}

void Logger::setUserOut(OutputStream *stream)
{
  if (mode_ != ModeSystem)
  {
    #ifdef TEST
    log() << "Logger: Setting the user output stream.\n";
    #endif

    setStream(userOut_, userOutSave_, userOutName_, stream);

    #ifdef TEST
    log() << "Logger: Redirecting the standard error to "
          << "the stream.\n";
    #endif

    getDefaultOut() -> rdbuf(userOut_ -> rdbuf());

    LogDefault() << "Logger: Redirected the standard error to "
                 << "the stream.\n";
  }
  #ifdef TEST
  else
  {
    log() << "Logger: Ignoring user output stream "
          << "with mode " << Itk(mode_) << ".\n";
  }
  #endif
}

void Logger::setStatOut(OutputStream *stream)
{
  #ifdef TEST
  log() << "Logger: Setting the stat output stream.\n";
  #endif

  setStream(statOut_, statOutSave_, statOutName_, stream);
}

void Logger::resetLogOut()
{
  #ifdef TEST
  log() << "Logger: Resetting the log output stream.\n";
  #endif

  resetStream(logOut_, logOutSave_, logOutName_);
}

void Logger::resetUserOut()
{
  #ifdef TEST
  log() << "Logger: Resetting the user output stream.\n";
  #endif

  if (userOutName_ != NULL)
  {
    #ifdef TEST
    log() << "Logger: Restoring the standard error.\n";
    #endif

    getDefaultOut() -> rdbuf(defaultBuffer_);
  }

  resetStream(userOut_, userOutSave_, userOutName_);
}

void Logger::resetStatOut()
{
  #ifdef TEST
  log() << "Logger: Resetting the stat output stream.\n";
  #endif

  resetStream(statOut_, statOutSave_, statOutName_);
}

void Logger::truncateLogOut(Int limit)
{
  #ifdef TEST
  log() << "Logger: Checking the log output stream.\n";
  #endif

  truncateStream(logOut_, logOutSave_, logOutName_, limit);
}

void Logger::truncateUserOut(Int limit)
{
  #ifdef TEST
  log() << "Logger: Checking the user output stream.\n";
  #endif

  truncateStream(userOut_, userOutSave_, userOutName_, limit);
}

void Logger::truncateStatOut(Int limit)
{
  #ifdef TEST
  log() << "Logger: Checking the stat output stream.\n";
  #endif

  truncateStream(statOut_, statOutSave_, statOutName_, limit);
}

LogStream *Logger::getDefaultOut()
{
  #if (LOGGER_DEFAULT_BUFFERING == 0)

  //
  // Force the standard error to be unbufferd.
  // This should not be required but when the
  // output is redirected, Windows appears to
  // behave differently.
  //

  #ifdef WIN32

  setbuf(stderr, NULL);

  #endif

  return (LogStream *) &cerr;

  #else

  return (LogStream *) &clog;

  #endif
}

void Logger::openStream(LogStream *&stream, LogStream *&save,
                            Char *&name, const Char *directory, const Char *file)
{
  Char *node = makeNode(directory, file);

  if (node == NULL)
  {
    return;
  }

  LogStream *replace = makeStream(FileStreamAppend, node);

  if (replace == NULL)
  {
    StringReset(node);

    return;
  }

  StringReset(name);

  name = node;

  LogStream *previous = stream;

  stream = replace;

  deleteStream(previous, save);
}

void Logger::truncateStream(LogStream *&stream, LogStream *&save,
                                Char *&name, Int limit)
{
  if (name == NULL)
  {
    return;
  }
  else if (limit < 0)
  {
    return;
  }

  FileStats stats;

  if (FileQuery(name, &stats) != 1)
  {
    return;
  }
  else if (stats.st_size < (Long) limit)
  {
    return;
  }

  #ifdef TEST
  log() << "Logger: Truncating file " << Stk(name)
        << " with size " << stats.st_size << ".\n";
  #endif

  //
  // Delete the old stream and use the standard
  // error until the file is reopened.
  //

  LogStream *replace = stream;

  stream = getDefaultOut();

  deleteStream(replace, save);

  replace = makeStream(FileStreamTruncate, name);

  if (replace == NULL)
  {
    StringReset(name);

    return;
  }

  stream = replace;

  #ifdef TEST
  log() << "Logger: Reopened file " << Stk(name) << ".\n";
  #endif
}

void Logger::setStream(LogStream *&stream, LogStream *&save,
                           Char *&name, OutputStream *requested)
{
  LogStream *previous = stream;

  stream = (LogStream *) requested;

  deleteStream(previous, save);

  StringReset(name);
}

void Logger::resetStream(LogStream *&stream, LogStream *&save, Char *&name)
{
  LogStream *previous = stream;

  stream = getDefaultOut();

  deleteStream(previous, save);

  StringReset(name);
}

Char *Logger::makeNode(const Char *directory, const Char *file)
{
  Char *node;

  if (file == NULL || *file == '\0')
  {
    return NULL;
  }
  else if ((node = FilePath(directory, file)) == NULL)
  {
    return NULL;
  }

  #ifdef TEST

  if (directory != NULL)
  {
    log() << "Logger: Using compound node name "
          << Stk(node) << ".\n";
  }
  else
  {
    log() << "Logger: Using simple node name "
          << Stk(node) << ".\n";
  }

  #endif

  return node;
}

LogStream *Logger::makeStream(FileStreamMode mode, const Char *node)
{
  #ifdef TEST
  log() << "Logger: Opening output file " << Stk(node)
        << ".\n";
  #endif

  Int mask = FileMask(0077);

  //
  // The file can be temporarily unavailable.
  // This can happen at random under Cygwin.
  //

  #ifndef __CYGWIN32__

  Int retry = 5;

  #else

  Int retry = 20;

  #endif

  LogStream *stream;

  for (Int i = 0; i < retry; i++)
  {
    stream = (LogStream *) new OutputFileStream();

    if (stream != NULL)
    {
      //
      // Check whether we need to make the
      // output unbuffered.
      //

      if (buffered_ == 0)
      {
        #ifdef TEST
        log() << "Logger: Making the output unbuffered.\n";
        #endif

        #if __GNUC__ < 3

        stream -> rdbuf() -> setbuf(0, 0);

        #else

        stream -> rdbuf() -> pubsetbuf(0, 0);

        #endif
      }
      #ifdef TEST
      else
      {
        log() << "Logger: Making the output buffered.\n";
      }
      #endif

      ((OutputFileStream *) stream) -> open(node, mode);

      if (stream -> fail() == 0)
      {
        break;
      }
      else
      {
        #ifdef TEST

        log() << "Logger: Failed to open file " << Stk(node)
              << ".";

        if (i + 1 < retry)
        {
          log() << " Trying again...\n";
        }
        else
        {
          log() << "\n";
        }
          
        #endif

        delete stream;

        stream = NULL;        
      }
    }

    ProcessSleep(200);
  }

  FileMask(mask);

  if (stream == NULL)
  {
    return NULL;
  }
  #ifdef TEST
  else
  {
    log() << "Logger: Opened file " << Stk(node) << ".\n";
  }
  #endif

  return stream;
}

void Logger::deleteStream(LogStream *stream, LogStream *&save)
{
  if (stream != getDefaultOut())
  {
    //
    // Don't delete the stream now since it
    // can be in use by some other thread.
    //

    if (save != NULL)
    {
      #ifdef TEST
      log() << "Logger: Deleting the saved stream.\n";
      #endif

      delete save;
    }

    #ifdef TEST
    log() << "Logger: Saving the old stream.\n";
    #endif

    stream -> flush();

    stream -> rdbuf(defaultBuffer_);

    save = stream;
  }
}
