/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Logger_H
#define Logger_H

#include "Mutex.h"

//
// Define this to overload the insert operator
// of the log output stream. This can be useful
// to add a global lock to the log output with
// the deferred flush at the time the log line
// is completed.
//

#undef LOG_USE_OVERLOADED_INSERT

#ifdef LOG_USE_OVERLOADED_INSERT

class LogStream : public OutputStream
{
  public:

  //
  // Implement overloading of the string
  // insertion operator.
  //

  LogStream &insert(const Char *string);
};

//
// Overloaded insertion operator.
//

#define LOG_INSERT_STRING(stream, string) \
\
  (stream).insert((const Char *) (string)); \
\
  return (stream)

inline LogStream &operator<<(LogStream &stream, Char *string)
{
  LOG_INSERT_STRING(stream, string);
}

inline LogStream &operator<<(LogStream &stream, const Char *string)
{
  LOG_INSERT_STRING(stream, string);
}

inline LogStream &operator<<(LogStream &stream, signed Char *string)
{
  LOG_INSERT_STRING(stream, string);
}

inline LogStream &operator<<(LogStream &stream, const signed Char *string)
{
  LOG_INSERT_STRING(stream, string);
}

inline LogStream &operator<<(LogStream &stream, unsigned Char *string)
{
  LOG_INSERT_STRING(stream, string);
}

inline LogStream &operator<<(LogStream &stream, const unsigned Char *string)
{
  LOG_INSERT_STRING(stream, string);
}

#define LogStreamBuffer OutputStreamBuffer

#else // #ifdef LOG_USE_OVERLOADED_INSERT ...

#define LogStream       OutputStream
#define LogStreamBuffer OutputStreamBuffer

#endif // #ifdef LOG_USE_OVERLOADED_INSERT ... #else ...

/*
FIXME: This must become a Lockable class. In this
       way we can remove the static mutex object.
*/
class Logger
{
  public:

  //
  // Available logging modes:
  //
  // ModeSystem: The output will always go to the
  //             system logger and any redirection
  //             attempt will be ignored.
  //
  // ModeThread: The output occurring in the parent
  //             thread of an object will go to the
  //             stream selected by the parent, and
  //             not to the stream selected by the
  //             object's logger.
  //
  // ModeObject: The output will go to the actual
  //             stream set for the object's logger.
  //

  typedef enum
  {
    ModeSystem,
    ModeThread,
    ModeObject
  }
  Mode;

  //
  // Constructors and destructors.
  //

  Logger();

  ~Logger();

  //
  // Set the logging mode.
  //

  void setMode(Mode mode)
  {
    mode_ = mode;
  }

  Mode getMode() const
  {
    return mode_;
  }

  //
  // Enable or disable the log output.
  //

  void setEnabled(Int value)
  {
    enabled_ = (value != 0);
  }

  Int isEnabled() const
  {
    return enabled_;
  }

  //
  // Tell the logger whether to include
  // the name of the derived class in
  // the log output.
  //

  void setExtended(Int value)
  {
    extended_ = (value != 0);
  }

  Int isExtended() const
  {
    return extended_;
  }

  //
  // Enable or disable buffering.
  //

  void setBuffered(Int value)
  {
    buffered_ = (value != 0);
  }

  //
  // Tell the logger if a global lock
  // on the log output is requested.
  //

  void setExclusive(Int value)
  {
    exclusive_ = (value != 0);
  }

  Int isExclusive() const
  {
    return exclusive_;
  }

  //
  // Acquire or release a global lock
  // on the log output.
  //

  static void lock()
  {
    mutex_.lock();
  }

  static void unlock()
  {
    mutex_.unlock();
  }

  static Int unwind()
  {
    return mutex_.unwind();
  }

  static void reset()
  {
    mutex_.reset();
  }

  //
  // Return the stream to use for the
  // selected output.
  //

  LogStream *getLogOut() const
  {
    return (enabled_ == 1 ? logOut_ : nullOut_);
  }

  LogStream *getUserOut() const
  {
    return userOut_;
  }

  LogStream *getStatOut() const
  {
    return statOut_;
  }

  LogStream *getNullOut() const
  {
    return nullOut_;
  }

  //
  // Let the user redirect the selected
  // output to a different file.
  //

  void setLogOut(const Char *directory, const Char *file);
  void setUserOut(const Char *directory, const Char *file);
  void setStatOut(const Char *directory, const Char *file);

  void setLogOut(const Char *file)
  {
    setLogOut(NULL, file);
  }

  void setUserOut(const Char *file)
  {
    setUserOut(NULL, file);
  }

  void setStatOut(const Char *file)
  {
    setStatOut(NULL, file);
  }

  void setLogOut(OutputStream *stream);
  void setUserOut(OutputStream *stream);
  void setStatOut(OutputStream *stream);

  const Char *getLogOutName() const
  {
    return logOutName_;
  }

  const Char *getUserOutName() const
  {
    return userOutName_;
  }

  const Char *getStatOutName() const
  {
    return statOutName_;
  }

  //
  // Let the output go again to the standard
  // error.
  //

  void resetLogOut();
  void resetUserOut();
  void resetStatOut();

  //
  // Truncate the log file if its size exceeds
  // the limit.
  //

  void truncateLogOut(Int limit);
  void truncateUserOut(Int limit);
  void truncateStatOut(Int limit);

  //
  // Return either cerr or clog, depending on
  // the default buffering mode set at compile
  // time.
  //

  static LogStream *getDefaultOut();

  protected:

  //
  // Utility functions used to concatenate the
  // directory and file name and replace the
  // old stream with the new.
  //

  void openStream(LogStream *&stream, LogStream *&save,
                      Char *&name, const Char *directory, const Char *file);

  void truncateStream(LogStream *&stream, LogStream *&save,
                          Char *&name, Int limit);

  void setStream(LogStream *&stream, LogStream *&save,
                     Char *&name, OutputStream *requested);

  void resetStream(LogStream *&stream, LogStream *&save, Char *&name);

  void deleteStream(LogStream *stream, LogStream *&save);

  Char *makeNode(const Char *directory, const Char *file);

  LogStream *makeStream(FileStreamMode mode, const Char *node);

  private:

  //
  // Log file for the session.
  //

  LogStream *logOut_;
  LogStream *logOutSave_;

  //
  // User-level messages printed for the
  // session.
  //

  LogStream *userOut_;
  LogStream *userOutSave_;

  //
  // Output file where we need to print
  // the session statistics.
  //

  LogStream *statOut_;
  LogStream *statOutSave_;

  //
  // Names of files where the output was
  // redirected.
  //

  Char *logOutName_;
  Char *userOutName_;
  Char *statOutName_;

  //
  // Null stream used to discard the output.
  // 

  LogStream *nullOut_;

  //
  // Save the original stream buffer in the
  // case the standard error is redirected.
  //

  LogStreamBuffer *defaultBuffer_;

  //
  // The current logging mode.
  //

  Mode mode_;

  //
  // Whether the disagnostic output
  // is or not enabled.
  //

  Int enabled_;

  //
  // Whether to include extended class
  // name in the output.
  //

  Int extended_;

  //
  // Whether the output is buffered.
  //

  Int buffered_;

  //
  // Whether a global lock is required
  // upon accessing the output.
  //

  Int exclusive_;

  //
  // Global mutex arbitrating the access
  // to the log.
  //

  static SYSTEM Mutex mutex_;
};

//
// Return a global logger for the system.
//

Logger *GetLogger();

#endif /* Logger_H */
