/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Log.h"
#include "Error.h"
#include "Memory.h"
#include "String.h"
#include "Random.h"
#include "List.h"

#define WARNING
#define ERROR
#undef  TEST
#undef  DEBUG

//
// Implement the standard Unix shared memory
// functions on Windows.
//

#ifdef WIN32

#define FILE_MAP_NAME_LENGTH  64
#define FILE_MAP_HEADER_SIZE  4

//
// Try to tanslate the Windows error to the
// POSIX error having the same meaning.
//

void MemoryTranslateError()
{
  errno = GetWindowsError();
}

class ShmHandleInfoList : private List
{
  public:

  Iterator getBegin()
  {
    return List::getBegin();
  }

  Iterator getEnd()
  {
    return List::getEnd();
  }

  Int getSize()
  {
    return List::getSize();
  }

  void addValue(struct SharedMemoryHandleInfo *info)
  {
    return List::addValue(info);
  }

  void *getValue(Int index)
  {
    return List::getValue(index);
  }

  void removeValue(void *value)
  {
    return List::removeValue(value);
  }
};

static class ShmHandleInfoList SharedMemoryHandleInfoList;

Int SharedMemoryCreate(Int key, Size size, Int shmflg)
{
  Int userFlags;
  Int flags;
  Int index = 0;
  Int found = 0;
  Int id, close;

  HANDLE mapping;

  struct SharedMemoryHandleInfo *mappingInfo = new struct SharedMemoryHandleInfo;

  void *mappingAddr = NULL;

  Char shmname[FILE_MAP_NAME_LENGTH];
 
  #ifdef TEST
  Log() << "Memory: Going to create shared memory "
        << "with size " << size << ".\n";
  #endif

  userFlags = shmflg & 0700;

  switch (userFlags)
  {
    case 0700:
    case 0600:
    {
      //
      // case (PROT_EXEC | PROT_READ | PROT_WRITE):
      // PROT_EXEC is not supported by Windows 2000,
      // Server 2003 with SP1, XP until SP2. 
      // In this case, 'case 0600:' is used.
      //
      // case (PROT_READ | PROT_WRITE):
      //

      flags = PAGE_READWRITE;

      break;
    }
    case 0500:
    case 0400:
    {
      //
      // case (PROT_EXEC | PROT_READ):
      // PROT_EXEC is not supported by Windows 2000,
      // Server 2003 with SP1, XP until SP2.
      // In this case, 'case 0400:' is used.
      // case PROT_READ:
      //

      flags = PAGE_READONLY;

      break;
    }
    case 0300:
    case 0200:
    {
      //
      // case (PROT_EXEC | PROT_WRITE):
      // PROT_EXEC is not supported by Windows 2000,
      // Server 2003 with SP1, XP until SP2.
      // In this case, 'case 0200:' is used.
      // case PROT_WRITE:
      //

      flags = PAGE_WRITECOPY;

      break;
    }
    default:
    {
      #ifdef WARNING
      Log() << "Memory: WARNING! Supported flags are [2-7].\n";
      #endif
     
      return -1;
    }
  }

  found = 1;
  id    = 0;

/*
FIXME: Force the session id to 0 in order to
       use the global mapping.

  DWORD pSessionId;

  ProcessIdToSessionId (GetCurrentProcessId(), &pSessionId);
*/

  DWORD pSessionId = 0;

  do
  {
    close = 0;

    while (found == 1)
    {
      id = RandomGet();

      if ((id % 2 == 0 && pSessionId != 0) ||
              (id % 2 == 1 && pSessionId == 0))
      {
        id++;
      }
 
      index = 0;
      found = 0;
 
      while (index < SharedMemoryHandleInfoList.getSize() && found == 0)
      {
        mappingInfo = (struct SharedMemoryHandleInfo *)
                          SharedMemoryHandleInfoList.getValue(index);
 
        mappingInfo -> id != id ? index++ : found = 1;
      }
    }

    if (pSessionId == 0)
    {
      #ifdef TEST
      Log() << "Memory: Creating global mapping.\n";
      #endif
/*
FIXME: Convert to StringSizePrint() and add the
       final terminator to handle the overflow.
*/
      StringPrint(shmname, "Global\\ShmMap-%d", id);
    }
    else
    {
      #ifdef TEST
      Log() << "Memory: Creating local mapping.\n";
      #endif
/*
FIXME: Convert to StringSizePrint() and add the
       final terminator to handle the overflow.
*/
      StringPrint(shmname, "Local\\ShmMap-%d", id);
    }

/*
FIXME: Making the mapping accessible by everyone.

    mapping = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, flags,
                                    0, size + FILE_MAP_HEADER_SIZE, shmname);
*/
    SECURITY_DESCRIPTOR sd;
/*
FIXME: Add the error handling.
*/
    if (InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION) == 0)
    {
    }

    if (SetSecurityDescriptorDacl(&sd, true, 0, false) == 0)
    {
    }

    SECURITY_ATTRIBUTES sa;

    sa.nLength = sizeof (SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = &sd;
    sa.bInheritHandle = FALSE;

    //
    // If the function succeeds, the return value is a
    // handle to the newly created file mapping object.
    // If fails, the return value is NULL.
    // GetLastError() specified error information.
    // On the contrary, shmget returns a valid segment
    // identifier (shmid), or -1 on error.   
    //

    mapping = CreateFileMapping(INVALID_HANDLE_VALUE, &sa, flags,
                                    0, size + FILE_MAP_HEADER_SIZE, shmname);
 
    if (mapping == NULL)
    {
      #ifdef WARNING
      Log() << "Memory: WARNING! Failed to create file "
            << "mapping with error " << GetLastError()
            << ".\n";
      #endif
 
      return -1;
    }

    if (GetLastError() == ERROR_ALREADY_EXISTS)
    {
      close = 1;

      #ifdef TEST
      Log() << "Memory: ID#" << id << " already exists.\n";
      #endif

      CloseHandle(mapping);
    }

    SetHandleInformation(mapping, HANDLE_FLAG_INHERIT, 0);
  }
  while (close == 1);

  mappingInfo -> id = id;

  #if defined(__x86_64__) || defined(__aarch64__)

  mappingInfo -> shmid = (Long Long) mapping;

  #else

  mappingInfo -> shmid = (Int) mapping;

  #endif

  mappingInfo -> info.shm_segsz   = size;
  mappingInfo -> info.shm_nattch  = 0;

  SharedMemoryHandleInfoList.addValue(mappingInfo);

  #ifdef TEST
  Log() << "Memory: Created shared memory "
        << "ID#" << id << " with "
        << "handle " << mapping << ".\n";

  Log() << "Memory: Going to attach header in mapped file.\n";
  #endif

  mappingAddr = MapViewOfFile(mapping, FILE_MAP_ALL_ACCESS, 0, 0, 0);

  if (mappingAddr != NULL)
  {
    MemoryCopy(mappingAddr, &size, sizeof(Size));

    #ifdef TEST
    Log() << "Memory: Saved mapped file "
          << "with size " << size << ".\n";
    #endif

    UnmapViewOfFile(mappingAddr);
  }

  return (Int) id;
}

void *SharedMemoryAttach(Int shmid, const void *shmaddr, Int shmflg)
{
  HANDLE mapping = 0;
  HANDLE id = 0;
  void *mappingAddr;
  Int userFlags, flags, index, found, listSize;
  struct SharedMemoryHandleInfo *mappingInfo = NULL;
  Size size;

  index = 0;
  found = 0;

  userFlags = shmflg & 0700;

  //
  // shmat: if SHM_RDONLY is specified in shmflg, the
  // segment is attached for reading and the process
  // must have read permission for the segment.
  // Otherwise the segment is attached for read and
  // write and the process must have read and write
  // permission for the segment. There is no notion
  // of a write-only shared memory segment.
  //

  flags = (userFlags == 0500 || userFlags == 0400) ? FILE_MAP_READ
                                                       : FILE_MAP_ALL_ACCESS;
 
  #ifdef TEST
  Log() << "Memory: Going to attach shared memory "
        << "with ID# " << shmid << ".\n";
  #endif

  listSize = SharedMemoryHandleInfoList.getSize();

  while (index < listSize && found == 0)
  {
    mappingInfo = (struct SharedMemoryHandleInfo*)
                      SharedMemoryHandleInfoList.getValue(index);

    mappingInfo -> id != shmid ? index++ : found = 1;
  }

  if (found == 0)
  {
    mappingInfo = new struct SharedMemoryHandleInfo;

    Char shmname[FILE_MAP_NAME_LENGTH];

    if (shmid % 2 == 0)
    {
      #ifdef TEST
      Log() << "Memory: Attaching to the global mapping.\n";
      #endif
/*
FIXME: Convert to StringSizePrint() and add the
       final terminator to handle the overflow.
*/
      StringPrint(shmname, "Global\\ShmMap-%d", shmid);
    }
    else
    {
      #ifdef TEST
      Log() << "Memory: Attaching to the local mapping.\n";
      #endif
/*
FIXME: Convert to StringSizePrint() and add the
       final terminator to handle the overflow.
*/
      StringPrint(shmname, "Local\\ShmMap-%d", shmid);
    }

    #ifdef TEST
    Log() << "Memory: Opening file mapping " << shmname
          << ".\n";
    #endif

    mapping = OpenFileMapping(flags, 1, shmname);

    if (mapping == NULL)
    {
      #ifdef WARNING
      Log() << "Memory: WARNING! Invalid "
            << "shared memory ID#" << shmid
            << ".\n";
      #endif

      return (void *) -1;
    }
  }

  if (found == 0)
  {
    id = mapping;
  }
  else
  {
    id = HandleCast(mappingInfo -> shmid);
  }

  //
  // If the operation succeeds, the return value is
  // the starting address of the mapped view. If it
  // fails, the return value is NULL and we can call
  // GetLastError() to get error information. On the
  // contrary, shmat() on success returns the address
  // of the attached shared memory segment and on a
  // failure (void *) -1 is returned.
  //

  mappingAddr = MapViewOfFile(id, flags, 0, 0, 0);

  if (mappingAddr == NULL)
  {
    mappingAddr = (void *) -1;
  }
/*
FIXME: Not implemented yet.

  else
  {
    mappingInfo -> info.shm_nattach++;
  }
*/
  #ifdef WARNING

  if (mappingAddr == (void *) -1)
  {
    MemoryTranslateError();

    Log() << "Memory: WARNING! Failed to map the file "
          << "in memory.\n";

    Log() << "Memory: WARNING! Error is " << GetError()
          << ", " << Stk(GetErrorString()) << ".\n";
  }

  #endif

  if (found == 0 && mappingAddr != (void *) -1)
  {
    MemoryCopy(&size, mappingAddr, sizeof(Size));

    mappingInfo -> id = shmid;

    #if defined(__x86_64__) || defined(__aarch64__)

    mappingInfo -> shmid = (Long Long) mapping;

    #else

    mappingInfo -> shmid = (Int) mapping;

    #endif

    mappingInfo -> info.shm_segsz   = size;
    mappingInfo -> info.shm_nattch  = 0;

    SharedMemoryHandleInfoList.addValue(mappingInfo);
  }

  #ifdef TEST
  Log() << "Memory: Shared memory attached "
        << "to ID#" << shmid << " with "
        << "result " << mappingAddr << ".\n";
  #endif

  if (mapping != NULL)
  {
/*
FIXME: This breaks segment reattaching on windows.

    CloseHandle(mapping);
*/
  }

  return ((Char *) mappingAddr) + FILE_MAP_HEADER_SIZE;
}

Int SharedMemoryDetach(const void *shmaddr)
{
  if (shmaddr == (void *) -1)
  {
    shmaddr = NULL;
  }

  shmaddr = ((Char *) shmaddr) - FILE_MAP_HEADER_SIZE;

  //
  // UnmapViewOfFile: if succeds, the return value
  // is nonzero, if fails, the return value is 0.
  // GetLastError gets error information.
  // shmdt: on success returns 0, on error -1.
  //

  Int result = UnmapViewOfFile(shmaddr);

  result = (result != 0) ? 0 : -1;

  #ifdef WARNING

  if (result == -1)
  {
    MemoryTranslateError();

    Log() << "Memory: WARNING! Failed to unmap the file "
          << "in memory.\n";

    Log() << "Memory: WARNING! Error is " << GetError()
          << ", " << Stk(GetErrorString()) << ".\n";
  }

  #endif

  #ifdef TEST
  Log() << "Memory: Detached shared memory "
        << "address " << shmaddr  << " with "
        << "result " << result << ".\n";
  #endif

  return result;
}

Int SharedMemoryControl(Int shmid, Int cmd, struct shmid_ds* buf)
{
  Int result;

  SharedMemoryHandleInfo *mappingInfo = NULL;
  Int index = 0, found = 0, listSize = SharedMemoryHandleInfoList.getSize();

  while (found == 0 && index < listSize) 
  {
    mappingInfo = (struct SharedMemoryHandleInfo*)
                      SharedMemoryHandleInfoList.getValue(index);

    mappingInfo -> id != shmid ? index++ : found = 1;
  }

  if (found == 0)
  {
    #ifdef WARNING
    Log() << "Memory: WARNING! Invalid ID#" << shmid << ".\n";
    #endif

    return -1;
  }

  switch (cmd)
  {
    case IPC_RMID:
    {
      //
      // Remove shared memory segment shmid and
      // destroy corresponding shmid_ds.
      //

      result = CloseHandle(HandleCast(mappingInfo -> shmid));

      result = (result != 0 ? 0 : -1);

      #ifdef WARNING

      if (result == -1)
      {
        MemoryTranslateError();

        Log() << "Memory: WARNING! Failed to close the file "
              << "handle.\n";

        Log() << "Memory: WARNING! Error is " << GetError()
             << ", " << Stk(GetErrorString()) << ".\n";

        return result;
      }

      #endif

      SharedMemoryHandleInfoList.removeValue(mappingInfo);

      #ifdef TEST
      Log() << "Memory: Closed shared memory"
            << "ID#" << shmid << " with "
            << "result: " << result << ".\n";
      #endif

      break;
    }
    case IPC_STAT:
    {
      //
      // Copy current values for shared memory
      // segment shmid into buffer.
      //

      buf -> shm_segsz  = mappingInfo -> info.shm_segsz;
      buf -> shm_nattch = mappingInfo -> info.shm_nattch;

      #ifdef TEST
      Log() << "Memory: Shared memory ID#" << shmid
            << " has size " << buf -> shm_segsz
            << " and " << buf -> shm_nattch
            << " attached.\n";
      #endif

      result = 0;

      break;
    }
    case IPC_SET:
    {
      //
      // Set values of fields for shared memory
      // segment shmid from values found in buffer.
      // At moment IPC_SET on windows does nothing.
      // Permission 0777 changed in 0600 (see
      // CreateFileMapping).
      //

      result = 0;

      break;
    }
    default:
    {
      #ifdef WARNING
      Log() << "Memory: WARNING! Unknow command " << cmd
            << " to control shared memory.\n";
      #endif

      result = -1;
    }
  }

  return result;
}

#endif // #ifdef WIN32 ...

//
// Print a warning on Android and iOS
// and make the function fail.
//

#if defined(__BIONIC__) || defined(__IPHONEOS__)

Int  SharedMemoryCreate(Int key, Size size, Int flags)
{
  #ifdef WARNING
  Log() << "Memory: WARNING! Function SharedMemoryCreate() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

void *SharedMemoryAttach(Int id, const void *address, Int flags)
{
  #ifdef WARNING
  Log() << "Memory: WARNING! Function SharedMemoryAttach() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return (void *) -1;
}

Int SharedMemoryDetach(const void *address)
{
  #ifdef WARNING
  Log() << "Memory: WARNING! Function SharedMemoryDetach() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int SharedMemoryControl(Int id, Int command, struct shmid_ds *info)
{
  #ifdef WARNING
  Log() << "Memory: WARNING! Function SharedMemoryControl() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

#endif // #if defined(__BIONIC__) || defined(__IPHONEOS__) ...
