/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Memory_H
#define Memory_H

#ifdef WIN32

#ifdef ERROR
#undef ERROR
#endif

#endif

#include <stdlib.h>
#include <string.h>

#ifdef __APPLE__
#include <sys/types.h>
#include <malloc/malloc.h>
#else
#include <malloc.h>
#endif

#ifndef SHM_MEMORY
#include "Types.h"
#else
#define Int  int
#define Size size_t
#endif

#define MemoryAlloc   malloc
#define MemoryRealloc realloc
#define MemoryCopy    memcpy
#define MemoryMove    memmove
#define MemorySet     memset

#define MemoryCompare memcmp

#ifdef WIN32
#define MemoryZero(destination, size)  memset((destination), 0, (size))
#else
#define MemoryZero(destination, size)  bzero((destination), (size))
#endif

#define MemoryFree free

//
// This call is not standard but it
// is present in different form on
// all operating sytems.
//

#ifdef WIN32

#define MemorySize(buffer) _msize((void *) buffer)

#elif defined(__linux__)

//
// Can be defined on some systems as
// size_t malloc_usable_size(void *).
//

#define MemorySize(buffer) malloc_usable_size((void *) buffer)

#elif defined(__APPLE__)

#define MemorySize(buffer) malloc_size(buffer)

#else

#define MemorySize(buffer) 0

#endif

//
// Shared memory creation functions.
//

#if !defined(WIN32) && !defined(__BIONIC__) && !defined(__IPHONEOS__)

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#define SharedMemoryId  Int

#define SharedMemoryCreate  shmget
#define SharedMemoryAttach  shmat
#define SharedMemoryDetach  shmdt
#define SharedMemoryControl shmctl

#else // #if !defined(WIN32) && !defined(__BIONIC__) && ...

#define SharedMemoryId  HANDLE

#define shmget SharedMemoryCreate
#define shmat  SharedMemoryAttach
#define shmdt  SharedMemoryDetach
#define shmctl SharedMemoryControl

#define IPC_PRIVATE  0
#define IPC_CREAT    00001000
#define IPC_RMID     0
#define IPC_SET      1
#define IPC_STAT     2

#define SHM_RDONLY   010000

#define ushort       unsigned short

struct ipc_perm
{
  ushort uid;
  int    mode;
};

//
// The shmid_ds struct is used in shmctl on
// Unix and defined here the Windows case.
// At the moment only the shm_segsz is used.
//

struct shmid_ds
{
  size_t          shm_segsz;
  struct ipc_perm shm_perm;
  int             shm_nattch;
};

//
// The SharedMemoryHandleInfo struct is used
// to associate the the Windows handle of the
// shared memory segment to an integer id.
//

struct SharedMemoryHandleInfo
{
  Int             id;
  Int             shmid;
  struct shmid_ds info;
};

#ifdef __cplusplus
extern "C"
{
#endif

Int  SharedMemoryCreate(Int key, Size size, Int flags);
void *SharedMemoryAttach(Int id, const void *address, Int flags);
Int  SharedMemoryDetach(const void *address);
Int  SharedMemoryControl(Int id, Int command, struct shmid_ds *info);

#ifdef __cplusplus
}
#endif

#endif // #if !defined(WIN32) && !defined(__BIONIC__) && ... #else ...

#endif /* Memory_H */
