/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Time.h"
#include "Error.h"
#include "Mutex.h"
#include "Process.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

#ifndef RELEASE

//
// If defined, assert the consistency
// of the operations performed on the
// object.
//

#define ASSERT

#endif

//
// Print a warning if there is a count
// inconsistency unwinding the mutex.
//

#undef  INSPECT

//
// Issue a warning if the number of
// locks granted on the object goes
// beyond the limit.
//

#define MUTEX_LOCK_LIMIT        1
#define MUTEX_UNLOCK_LIMIT      0

#define MUTEX_TRYLOCK_LIMIT     2
#define MUTEX_TRYUNLOCK_LIMIT   1

//
// Constructors and destructors.
//

Mutex::Mutex()
{
  #ifndef WIN32

  //
  // Explicitly make the mutex recursive.
  //

  pthread_mutexattr_t attributes;

  pthread_mutexattr_init(&attributes);

  pthread_mutexattr_settype(&attributes, PTHREAD_MUTEX_RECURSIVE);

  if (pthread_mutex_init(&mutex_, &attributes) != 0)
  {
    #ifdef ERROR
    LogSafe() << "Mutex: ERROR! Failed to initialize "
              << "the mutex.\n";
    #endif

    LogSafe() << "Failed to initialize the mutex.\n";
  }

  pthread_mutexattr_destroy(&attributes);

  #else

  //
  // Make the mutex anonymous and created
  // not owned.
  //

  mutex_ = CreateMutex(NULL, FALSE, NULL);

  if (mutex_ == NULL)
  {
    #ifdef ERROR
    LogSafe() << "Mutex: ERROR! Failed to initialize "
              << "the mutex.\n";
    #endif

    LogSafe() << "Failed to initialize the mutex.\n";
  }

  #endif

  #ifdef ASSERT

  count_ = 0;
  trace_ = 0;

  #endif
}

Mutex::~Mutex()
{
  //
  // Free the mutex resources.
  //

  #ifndef WIN32

  pthread_mutex_destroy(&mutex_);

  #else

  CloseHandle(mutex_);

  #endif
}

#ifdef ASSERT

//
// Can't use the logger since the logger
// itself may need to acquire a mutex.
//

void Mutex::lock()
{
  #ifdef DEBUG

  if (trace_ == 1)
  {
    LogSafe() << "Mutex: Locking mutex at " << this
              << " with count " << count_ << ".\n";
  }

  #endif

  #ifndef WIN32

  Int result;

  if ((result = pthread_mutex_lock(&mutex_)) != 0)

  #else

  DWORD result;

  if ((result = WaitForSingleObject(mutex_, INFINITE)) != WAIT_OBJECT_0)
  
  #endif

  {
    #ifdef WIN32

    result = GetWaitError(result);

    #endif

    const Char *string = GetErrorString(result);

    LogSafe() << "Mutex: ERROR! Lock of mutex at "
              << this << " failed.\n";

    LogSafe() << "Mutex: ERROR! Error is " << result
              << " " << Stk(string) << ".\n";
  }
  else
  {
    count_++;

    if (count_ > MUTEX_LOCK_LIMIT)
    {
      #if defined(TEST) || defined(INSPECT)
      LogSafe() << "Mutex: WARNING! Locked mutex at "
                << this << " with count " << count_
                << ".\n";
      #endif
    }
    #ifdef DEBUG
    else if (trace_ == 1)
    {
      LogSafe() << "Mutex: Locked mutex at " << this
                << " with count " << count_ << ".\n";
    }
    #endif
  }
}

void Mutex::unlock()
{
  #ifdef DEBUG

  if (trace_ == 1)
  {
    LogSafe() << "Mutex: Unlocking mutex at " << this
              << " with count " << count_ << ".\n";
  }

  #endif

  //
  // Decrement the counter while the mutex
  // is still presumably locked.
  //

  count_--;

  if (count_ < 0)
  {
    #if defined(TEST) || defined(INSPECT)
    LogSafe() << "Mutex: WARNING! Unlock inconsistency "
              << "in mutex at " << this << " with "
              << "count " << count_ << ".\n";
    #endif

    count_ = 0;
  }
  else if (count_ > MUTEX_UNLOCK_LIMIT)
  {
    #if defined(TEST) || defined(INSPECT)
    LogSafe() << "Mutex: WARNING! Unlocked mutex at "
              << this << " with count " << count_
              << ".\n";
    #endif
  }
  #ifdef DEBUG
  else if (trace_ == 1)
  {
    LogSafe() << "Mutex: Unlocked mutex at " << this
              << " with count " << count_ << ".\n";
  }
  #endif

  #ifndef WIN32

  Int result;

  if ((result = pthread_mutex_unlock(&mutex_)) != 0)

  #else

  DWORD result;

  if ((result = ReleaseMutex(mutex_)) == 0)
 
  #endif

  {
    #ifdef WIN32

    result = GetLastError();

    #endif

    const Char *string = GetErrorString(result);

    LogSafe() << "Mutex: ERROR! Unlock of mutex at "
              << this << " failed.\n";

    LogSafe() << "Mutex: ERROR! Error is " << result
              << " " << Stk(string) << ".\n";
  }
}

Int Mutex::trylock()
{
  #ifndef WIN32

  Int result;

  if ((result = pthread_mutex_trylock(&mutex_)) == 0)

  #else

  DWORD result;

  if ((result = WaitForSingleObject(mutex_, 0)) == WAIT_OBJECT_0)

  #endif

  {
    count_++;

    if (count_ > MUTEX_TRYLOCK_LIMIT)
    {
/*
FIXME: Due to a concurrency problem in libnxslavcf,
       the slave server needs to lock the session
       multiple times.

      #ifdef WARNING
      LogSafe() << "Mutex: WARNING! Trylocked mutex at "
                << this << " with count " << count_
                << ".\n";
      #endif
*/
      #ifdef TEST
      LogSafe() << "Mutex: WARNING! Trylocked mutex at "
                << this << " with count " << count_
                << ".\n";
      #endif
    }
    #ifdef DEBUG
    else
    {
      LogSafe() << "Mutex: Trylocked mutex at " << this
                << " with count " << count_ << ".\n";
    }
    #endif

    return 1;
  }

  #ifdef TEST

  #ifdef WIN32

  result = GetWaitError(result);

  #endif

  const Char *string = GetErrorString(result);

  LogSafe() << "Mutex: WARNING! Trylock of mutex at "
            << this << " failed.\n";

  LogSafe() << "Mutex: WARNING! Error is " << result
            << " " << Stk(string) << ".\n";

  #endif

  return 0;
}

Int Mutex::tryunlock()
{
  count_--;

  if (count_ < 0)
  {
    #if defined(TEST) || defined(INSPECT)
    LogSafe() << "Mutex: WARNING! Tryunlock inconsistency "
              << "in mutex at " << this << " with count "
              << count_ << ".\n";
    #endif

    count_ = 0;
  }

  #ifndef WIN32

  Int result;

  if ((result = pthread_mutex_unlock(&mutex_)) == 0)

  #else

  DWORD result;

  if ((result = ReleaseMutex(mutex_)) != 0)

  #endif

  {
    if (count_ > MUTEX_TRYUNLOCK_LIMIT)
    {
/*
FIXME: Due to a concurrency problem in libnxslavcf,
       the slave server needs to lock the session
       multiple times.

      #ifdef WARNING
      LogSafe() << "Mutex: WARNING! Tryunlocked mutex at "
                << this << " with count " << count_
                << ".\n";
      #endif
*/
      #ifdef TEST
      LogSafe() << "Mutex: WARNING! Tryunlocked mutex at "
                << this << " with count " << count_
                << ".\n";
      #endif
    }
    #ifdef DEBUG
    else
    {
      LogSafe() << "Mutex: Tryunlocked mutex at " << this
                << " with count " << count_ << ".\n";
    }
    #endif

    return 1;
  }

  #ifdef WIN32

  result = GetWaitError(result);

  #endif

  count_++;

  const Char *string = GetErrorString(result);

  #ifdef WARNING

  LogSafe() << "Mutex: WARNING! Tryunlock of mutex at "
            << this << " failed.\n";

  LogSafe() << "Mutex: WARNING! Error is " << result
            << " " << Stk(string) << ".\n";

  #endif

  return 0;
}

Int Mutex::unwind()
{
  Int result = 0;

  #ifndef WIN32

  while (pthread_mutex_unlock(&mutex_) == 0)

  #else

  while (ReleaseMutex(mutex_) != 0)

  #endif

  {
    count_--;

    #ifdef DEBUG
    LogSafe() << "Mutex: Unwound mutex at " << this
              << " with count " << count_ << ".\n";
    #endif

    result++;
  }

  if (count_ != 0)
  {
    #if defined(TEST) || defined(INSPECT)
    LogSafe() << "Mutex: WARNING! Unwind inconsistency "
              << "in mutex at " << this << " with "
              << "count " << count_ << ".\n";
    #endif

    count_ = 0;
  }

  return result;
}

#endif // #ifdef ASSERT ...

Int Mutex::timedlock(Int timeout)
{
  #ifndef WIN32

  #if defined(__APPLE__) || defined(__BIONIC__)

  //
  // The call is not available on MacOSX.
  // Emulate it by trying multiple times
  // after a short timeout.
  //

  Timeval start = getTimeval();

  Int result;

  Int interval = timeout / 5;

  if (interval < 1)
  {
    interval = 1;
  }
  else if (interval > 1000)
  {
    interval = 1000;
  }

  for (;;)
  {
    #ifdef TEST
    LogSafe() << "Mutex: Checking the mutex after "
              << diffMsTimeval(start, getTimeval())
              << " Ms.\n";
    #endif

    if ((result = trylock()) == 1)
    {
      //
      // Decrement the counter since it
      // will be incremented in the test
      // code checking the result.
      //

      #ifdef ASSERT

      count_--;

      #endif

      result = 0;

      break;
    }
    else if (diffMsTimeval(start, getTimeval()) + interval > timeout)
    {
      result = ETIMEDOUT;

      break;
    }

    ProcessSleep(interval);
  }

  if (result == 0)

  #else // #if defined(__APPLE__) || ...

  //
  // The function requires an absolute
  // time expressed through a timespec
  // structure.
  //

  Timespec ts;

  clock_gettime(CLOCK_REALTIME, &ts);

  if (timeout > 0)
  {
    addMsTimespec(ts, timeout);
  }

  Int result;

  if ((result = pthread_mutex_timedlock(&mutex_, &ts)) == 0)

  #endif // #if defined(__APPLE__) || ... #else ...

  #else // #ifndef WIN32 ...

  DWORD result;

  if ((result = WaitForSingleObject(mutex_, 0)) == WAIT_OBJECT_0)

  #endif // #ifndef WIN32 ... #else ...

  #ifndef ASSERT

  {
    return 1;
  }

  return 0;

  #else // #ifndef ASSERT ...

  {
    count_++;

    if (count_ > MUTEX_TRYLOCK_LIMIT)
    {
      #ifdef WARNING
      LogSafe() << "Mutex: WARNING! Timedlocked mutex at "
                << this << " with count " << count_
                << ".\n";
      #endif
    }
    #ifdef DEBUG
    else
    {
      LogSafe() << "Mutex: Timedlocked mutex at " << this
                << " with count " << count_ << ".\n";
    }
    #endif

    return 1;
  }

  #ifndef WIN32

  else if (result == ETIMEDOUT)

  #else

  else if (result == WAIT_TIMEOUT)

  #endif

  {
    #ifdef WARNING
    LogSafe() << "Mutex: WARNING! Timedlock of mutex at "
              << this << " failed with timeout.\n";
    #endif

    return 0;
  }

  //
  // The lock failed with an error.
  //

  #ifdef WIN32

  result = GetWaitError(result);

  #endif

  const Char *string = GetErrorString(result);

  LogSafe() << "Mutex: ERROR! Timedlock of mutex at "
            << this << " failed.\n";

  LogSafe() << "Mutex: ERROR! Error is " << result
            << " " << Stk(string) << ".\n";

  return 0;

  #endif // #ifndef ASSERT ... #else ...
}

void Mutex::spinlock()
{
  #ifdef TEST

  Int attempt = 1;

  Timeval start = getTimeval();

  #endif

  for (;;)
  {
    if (trylock() == 1)
    {
      #ifdef TEST

      Int elapsed = diffMsTimeval(start, getTimeval());

      if (elapsed > 1)
      {
        LogSafe() << "Mutex: Spinlocked mutex at attempt "
                  << attempt << " after " << elapsed
                  << " Ms.\n";
      }

      #endif

      return;
    }

    #ifdef TEST

    attempt++;

    #endif
  }
}

//
// Allocate a new mutex and replace the
// old one. This will implicitly release
// it, regardless who is owning the lock. 
//

void Mutex::reset()
{
  #ifdef TEST
  LogSafe() << "Mutex: WARNING! Resetting mutex at "
            << this << " with count " << count_
            << ".\n";
  #endif

  MutexHandle mutex;

  #ifndef WIN32

  //
  // This function is normally called after a fork
  // to clean up an orphaned mutex. Trying to lock
  // the previous mutex can cause a deadlock, so
  // we need to destroy the old mutex and create a
  // new one.
  //

  pthread_mutexattr_t attributes;

  pthread_mutexattr_init(&attributes);

  pthread_mutexattr_settype(&attributes, PTHREAD_MUTEX_RECURSIVE);

  if (pthread_mutex_init(&mutex, &attributes) != 0)
  {
    #ifdef ERROR
    LogSafe() << "Mutex: ERROR! Failed to reset the mutex.\n";
    #endif

    LogSafe() << "Failed to reset the mutex.\n";

    return;
  }

  pthread_mutexattr_destroy(&attributes);

  #else

  mutex = CreateMutex(NULL, FALSE, NULL);

  if (mutex == NULL)
  {
    #ifdef ERROR
    LogSafe() << "Mutex: ERROR! Failed to reset "
              << "the mutex.\n";
    #endif

    LogSafe() << "Failed to reset the mutex.\n";

    return;
  }

  #endif

  MutexHandle temporary = mutex_;

  mutex_ = mutex;

  #ifndef WIN32

  pthread_mutex_destroy(&temporary);

  #else

  CloseHandle(temporary);

  #endif

  #ifdef ASSERT

  count_ = 0;

  #endif
}
