/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Mutex_H
#define Mutex_H

#ifndef WIN32

#include <unistd.h>
#include <pthread.h>

#else

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>

#ifdef ERROR
#undef ERROR
#endif

#endif

#include "Types.h"

#ifndef RELEASE

//
// If defined, do additional logging
// and assert the consistency of the
// operations performed on the object.
//

#define ASSERT

#endif

class Mutex
{
  //
  // Make the definitions independent
  // from the platform.
  //

  #ifndef WIN32

  typedef pthread_mutex_t MutexHandle;

  #define MutexNone { { -1, -1, -1 } }

  #else

  typedef HANDLE MutexHandle;

  #define MutexNone ((MutexHandle) -1)

  #endif

  public:

  Mutex();

  ~Mutex();

  //
  // Inline versions of the mutex
  // functions.
  //
  
  #ifndef ASSERT

  void lock()
  {
    #ifndef WIN32

    pthread_mutex_lock(&mutex_);

    #else

    WaitForSingleObject(mutex_, INFINITE);

    #endif
  }

  void unlock()
  {
    #ifndef WIN32

    pthread_mutex_unlock(&mutex_);

    #else

    ReleaseMutex(mutex_);

    #endif
  }

  Int trylock()
  {
    #ifndef WIN32

    return (pthread_mutex_trylock(&mutex_) == 0);

    #else

    return (WaitForSingleObject(mutex_, 0) == WAIT_OBJECT_0);

    #endif
  }

  Int tryunlock()
  {
    #ifndef WIN32

    return (pthread_mutex_unlock(&mutex_) == 0);

    #else

    return (ReleaseMutex(mutex_) != 0);

    #endif
  }

  //
  // Repetedly unlock the mutex until
  // it is successfully released.
  //

  Int unwind()
  {
    Int result = 0;

    #ifndef WIN32

    while (pthread_mutex_unlock(&mutex_) == 0) result++;

    #else

    while (ReleaseMutex(mutex_) != 0) result++;

    #endif

    return result;
  }

  //
  // Stub versions of the test utility
  // functions.
  //

  Int getCount() const
  {
    return -1;
  }

  void setTrace(Int value)
  {
  }

  Int getTrace() const
  {
    return 0;
  }

  #else // #ifndef ASSERT ...

  //
  // Logging versions of the normally
  // inlined functions.
  //

  void lock();

  void unlock();

  Int trylock();

  Int tryunlock();

  Int unwind();

  //
  // Stub versions of the test utility
  // functions.
  //

  Int getCount()
  {
    return -1;
  }

  void setTrace(Int value)
  {
  }

  Int getTrace()
  {
    return 0;
  }

  #endif // #ifndef ASSERT ... #else ...

  //
  // Try to lock the mutex. Return 0 if
  // the mutex can't be locked before
  // the timeout.
  //

  Int timedlock(Int timeout);

  //
  // Add to the mutex the given number
  // of locks.
  //

  void relock(Int count)
  {
    while (count-- > 0)
    {
      lock();
    }
  }

  //
  // Keep the thread busy until it can
  // acquire the mutex.
  //

  void spinlock();

  //
  // Reset the mutex to the unlocked
  // state.
  //

  void reset();

  //
  // Data members.
  // 

  private:

  #ifdef ASSERT

  //
  // Count the current number of locks.
  //

  volatile Int count_;

  //
  // Tell the class to print a log when
  // the mutex is locked or unlocked.
  //

  volatile Int trace_;

  #endif // #ifdef ASSERT ...

  //
  // Handle of the mutex in its system
  // dependent form.
  //

  MutexHandle mutex_;
};

//
// Ensure that the assert definition
// does not propagate to other files.
//

#ifdef ASSERT
#undef ASSERT
#endif

#endif /* Mutex_H */
