/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include <stdio.h>
#include <unistd.h>

//
// Fool the compiler and get the deprecated
// daemon() function on MacOSX even if we
// compile with -Werror. The same trick is
// used in some Apple products.
//

#ifdef __APPLE__

#define daemon nowarndaemon
#include <stdlib.h>
#undef daemon
extern "C" { extern int daemon(int, int); }

#else // #ifdef __APPLE__ ...

#include <stdlib.h>

#endif // #ifdef __APPLE__ ... #else ...

//
// Deal with the Windows environment.
//

#ifdef WIN32

#ifdef STRING
#undef STRING
#endif

#if __GNUC__ > 3
#include <winternl.h>
#else
#include <windows.h>
#include <ddk/ntapi.h>
#endif

#include <ntstatus.h>

#ifdef ASSERT
#undef ASSERT
#endif

#else // #ifdef WIN32 ...

#include <sys/time.h>
#include <sys/resource.h>

#include <grp.h>
#include <pwd.h>

#endif // #ifdef WIN32 ... #else ...

#include "Arg.h"
#include "Dir.h"
#include "Error.h"
#include "String.h"
#include "Memory.h"
#include "Signal.h"
#include "Thread.h"
#include "Process.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// Only on Cygwin, retry n times waiting a
// given amount of milliseconds after each
// attempt.
//

#define PROCESS_RETRY_LIMIT           30
#define PROCESS_RETRY_TIMEOUT         1000

//
// Define the priorities to use on a Unix
// host.
//

#define PROCESS_PRIORITY_LOW          19
#define PROCESS_PRIORITY_NORMAL        0
#define PROCESS_PRIORITY_HIGH        -10
#define PROCESS_PRIORITY_REALTIME    -20

Int ProcessFork()
{
  #ifndef WIN32

  #ifdef __CYGWIN32__

  Int limit   = PROCESS_RETRY_LIMIT;
  Int timeout = PROCESS_RETRY_TIMEOUT;

  #else

  Int limit   = 1;
  Int timeout = 0;

  #endif

  Int id = 0;

  //
  // Prevent the other threads from locking
  // the low-level file stream resources.
  //

  GetLogger() -> lock();

  #ifdef TEST
  Log() << "Process: Doing a fork in process with "
        << "id " << ProcessGetPid() << ".\n";
  #endif

  for (Int i = 0; i < limit; i++)
  {
    //
    // It could possibly try again only if the
    // error code is 11, 'Resource temporarily
    // unavailable'.
    //

    while ((id = fork()) == -1 && errno == EINTR);

    if (id >= 0)
    {
      if (id == 0)
      {
        GetLogger() -> reset();
      }

      break;
    }
    else if (i < limit - 1)
    {
      #ifdef WARNING
      Log() << "Process: WARNING! Function fork() "
            << "failed.\n";
      #endif

      #ifdef WARNING
      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString())
            << ". Retrying...\n";
      #endif

      ProcessSleep(timeout);
    }
  }

  //
  // Unlock the logger in the parent.
  //
/*
FIXME: The following attempt to lock the logger
       before forking is ineffective at solving
       the problem.
*/
  if (id != 0)
  {
    GetLogger() -> unlock();
  }

  #ifdef TEST
  Log() << "Process: Returning from fork with "
        << "result " << id << ".\n";
  #endif

  return id;

  #else

  #ifdef WARNING
  Log() << "Process: WARNING! Function ProcessFork() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #endif
}

Int ProcessCreate(const Char *const file, const Char *const parameters[],
                      const Char *const environment[], Int fdin, Int fdout, Int fderr,
                          Int priority, Int preserve, Int interactive, Int inherit,
                              Int group)
{
  //
  // Count the process parameters. This is
  // mostly for error reporting since the
  // function requires that the first and
  // second parameters are the name of the
  // program. Most exec() implementations
  // ignore this count.
  //

  Int count;

  for (count = 0; parameters[count] != NULL; count++);

  if (count < 2)
  {
    #ifdef WARNING
    Log() << "Process: WARNING! Invalid number of "
          << "parameters in process creation.\n";
    #endif

    SetError(EINVAL);

    return -1;
  }

  #ifndef WIN32

  //
  // Do a fork, redirect the descriptors
  // and exec the requested program.
  //

  Int id = ProcessFork();

  if (id != 0)
  {
    if (id == -1)
    {
      #ifdef WARNING
      Log() << "Process: WARNING! Function fork() "
            << "failed.\n";
      #endif

      #ifdef WARNING
      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif
    }
    #ifdef TEST
    else
    {
      Log() << "Process: Created child with id "
            << id << ".\n";
    }
    #endif

    return id;
  }

  //
  // Disable the log output or we may
  // send it to the connected process.
  //

  #ifndef TEST

  GetLogger() -> setEnabled(0);

  #endif

  //
  // Child.
  //

  id = ProcessGetId();

  #ifdef TEST

  Int parent = ProcessGetParentPid();

  Log() << "Process: Executing child with id "
        << id << " and parent " << parent
        << ".\n";
  #endif

  if (preserve == 0)
  {
    #ifdef TEST
    Log() << "Process: Dropping the permissions "
          << "in the child process.\n";
    #endif

    ProcessSetGroupId(ProcessGetGroupId());
    ProcessSetUserId(ProcessGetUserId());
  }

  //
  // Set up stdin.
  //

  if (fdin != -1 && fdin != 0)
  {
    #ifdef TEST
    Log() << "Process: Using redirected input "
          << "on FD#" << fdin << ".\n";
    #endif

    FileClone(fdin, 0);

    if (fdin != fdout && fdin != fderr)
    {
      FileClose(fdin);
    }

    fdin = 0;
  }
  #ifdef TEST
  else
  {
    Log() << "Process: Using standard input "
          << "on FD#0.\n";
  }
  #endif

  //
  // Set up stdout.
  //

  if (fdout != -1 && fdout != 1)
  {
    #ifdef TEST
    Log() << "Process: Using redirected output "
          << "on FD#" << fdout << ".\n";
    #endif

    FileClone(fdout, 1);

    if (fdout != fdin && fdout != fderr)
    {
      FileClose(fdout);
    }

    fdout = 1;
  }
  #ifdef TEST
  else
  {
    Log() << "Process: Using standard output "
          << "on FD#1.\n";
  }
  #endif

  //
  // Set up stderr.
  //

  if (fderr != -1 && fderr != 2)
  {
    #ifdef TEST
    Log() << "Process: Using redirected error "
          << "on FD#" << fderr << ".\n";
    #endif

    FileClone(fderr, 2);

    if (fderr != fdin && fderr != fdout)
    {
      FileClose(fderr);
    }

    fderr = 2;
  }
  #ifdef TEST
  else
  {
    Log() << "Process: Using standard error "
          << "on FD#2.\n";
  }
  #endif

  #ifdef TEST
  Log() << "Process: Changing priority to "
        << Itk(priority) << ".\n";
  #endif

  ProcessSetPriority(priority);

  #ifdef DEBUG
  Log() << "Process: Yielding to the requested "
        << "command.\n";
  #endif

  ProcessExec(parameters[0], count, parameters + 1, environment);

  ProcessExit(GetError());

  #else // #ifndef WIN32 ...

  //
  // It is not possible to pass the command
  // parameters in a argument vector, so we
  // must create a single command string.
  //

  if (StringCompare(file, parameters[1]) != 0)
  {
    #ifdef ERROR
    Log() << "Process: ERROR! Command file must match "
          << "the second parameter.\n";
    #endif

    #ifdef ERROR
    Log() << "Process: ERROR! Running command " << file
          << ".\n";
    #endif

    SetError(ENOENT);

    return -1;
  }

  Int length = 32768;

  Char command[length];

  StringSizePrint(command, length, "\"%s\"", parameters[1]);

  for (Int i = 2; i < count && parameters[i] != NULL; i++)
  {
    #ifdef DEBUG
    Log() << "Process: Adding parameter " << Stk(parameters[i])
          << ".\n";
    #endif

    if (*parameters[i] == '\"')
    {
      StringSizePrint(command, length, "%s %s", command, parameters[i]);
    }
    else
    {
      StringSizePrint(command, length, "%s \"%s\"", command, parameters[i]);
    }
  }

  *(command + length - 1) = '\0';

  #ifdef DEBUG
  Log() << "Process: Using command " << Stk(command)
        << ".\n";
  #endif
 
  STARTUPINFO si;

  PROCESS_INFORMATION pi;
  
  DWORD flags = ProcessTranslatePriority(priority);

  MemoryZero(&si, sizeof(si));
  MemoryZero(&pi, sizeof(pi));

  si.cb = sizeof(si);

  //
  // Ensure that the inherit flag is set to true.
  // The application must handle the case where
  // any descriptor is not inherited as it would
  // handle the case in the Unix world. This may
  // imply the use of the special file device NUL
  // or creating a pipe and closing the opposite
  // end.
  //

  #ifdef WARNING

  if (inherit == 0)
  {
    Log() << "Process: WARNING! The inherit flag "
          << "must be true from this version.\n";
  }

  #endif

  si.dwFlags = STARTF_USESTDHANDLES;

  //
  // A casual Windows programmer may believe
  // that 0, 1 and 2 are OK, but that's not
  // the case.
  //

  if (fdin != -1 && fdin != 0)
  {
    si.hStdInput = HandleCast(FileGetId(fdin));

    #ifdef TEST

    Int idin = IntCast(si.hStdInput);

    Log() << "Process: Using redirected input "
          << "on FD#" << fdin << " ID#" << idin
          << ".\n";
    #endif
  }
  else
  {
    si.hStdInput = GetStdHandle(STD_INPUT_HANDLE);

    #ifdef TEST

    Int idin = IntCast(si.hStdInput);

    Log() << "Process: Using standard input "
          << "on FD#0 ID#" << idin << ".\n";
    #endif
  }

  if (fdout != -1 && fdout != 1)
  {
    si.hStdOutput = HandleCast(FileGetId(fdout));

    #ifdef TEST

    Int idout = IntCast(si.hStdOutput);

    Log() << "Process: Using redirected output "
          << "on FD#" << fdout << " ID#" << idout
          << ".\n";
    #endif
  }
  else
  {
    si.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);

    #ifdef TEST

    Int idout = IntCast(si.hStdOutput);

    Log() << "Process: Using standard output "
          << "on FD#1 ID#" << idout << ".\n";
    #endif
  }

  if (fderr != -1 && fderr != 2)
  {
    si.hStdError = HandleCast(FileGetId(fderr));

    #ifdef TEST

    Int iderr = IntCast(si.hStdError);

    Log() << "Process: Using redirected error "
          << "on FD#" << fderr << " ID#" << iderr
          << ".\n";
    #endif
  }
  else
  {
    si.hStdError = GetStdHandle(STD_ERROR_HANDLE);

    #ifdef TEST

    Int iderr = IntCast(si.hStdError);

    Log() << "Process: Using standard error "
          << "on FD#2 ID#" << iderr << ".\n";
    #endif
  }

  //
  // Use the following flags to hide the
  // process' console window.
  //

  if (interactive == 0)
  {
    #ifdef DEBUG
    Log() << "Process: Hiding the window of the "
          << "child process.\n";
    #endif

    si.dwFlags |= STARTF_USESHOWWINDOW;

    si.wShowWindow = SW_HIDE;
  }
  
  //
  // Create the process in a new process
  // group, if requested.
  //
/*
FIXME: Find a better way to handle extra system specific
       flags. CREATE_NEW_CONSOLE is needed for webplayer,
       but not needed in general case.
*/
  if (group == 1)
  {
    flags |= (CREATE_NEW_PROCESS_GROUP | CREATE_NEW_CONSOLE);
  }

  //
  // Start the child process with the following
  // parameters:
  //
  // - NULL, use the command line.
  //
  // - The command line.
  //
  // - NULL, the process handle is not inheritable.
  //
  // - NULL, the thread handle is not inheritable.
  //
  // - Handle inheritance is enabled.
  //
  // - The default priority class and other flags.
  //
  // - The environment block. If NULL, use the parent's environment.
  //
  // - NULL, use the parent's starting directory.
  //
  // - Address of the STARTUPINFO structure.
  //
  // - Address of the PROCESS_INFORMATION structure.
  //

  Int id;

  #ifdef TEST

  Int pid;

  #endif

  //
  // Build up the environment block.
  //

  Char *block;

  if (environment != NULL)
  {
    length = 0;

    for (Int i = 0; environment[i] != NULL; i++)
    {
      length += StringLength(environment[i]) + 1;
    }

    length++;

    StringAlloc(block, length);

    Char *value = block;

    for (Int i = 0; environment[i] != NULL; i++)
    {
      StringCopy(value, environment[i]);

      value += StringLength(value) + 1;
    }

    *value = '\0';
  }
  else
  {
    block = NULL;
  }

  HANDLE handle;

  if (CreateProcess(NULL, command, NULL, NULL, 1, flags,
                        (LPVOID) block, NULL, &si, &pi) == 0)
  {
    goto ProcessCreateError;
  }

  handle = pi.hProcess;

  if (handle == NULL)
  {
    goto ProcessCreateError;
  }

  //
  // Close the thread handle since we
  // don't need it.
  //

  CloseHandle(pi.hThread);

  //
  // The id returned by the function
  // is a handle, not the actual pid.
  //

  id = IntCast(handle);

  #ifdef TEST

  pid = ProcessGetPid(id);

  Log() << "Process: Created child with id "
        << id << " system pid " << pid << ".\n";
  #endif

  goto ProcessCreateEnd;

ProcessCreateError:

  //
  // Set the error explicitly since
  // CreateProcess() does not set
  // errno.
  //

  FileTranslateError();

  #ifdef WARNING
  Log() << "Process: WARNING! Process creation failed "
        << "for " << Stk(file) << ".\n";

  Log() << "Process: WARNING! Error is " << GetError()
        << " " << Stk(GetErrorString()) << ".\n";
  #endif

  id = -1;

ProcessCreateEnd:

  if (block != NULL)
  {
    StringReset(block);
  }

  return id;

  #endif // #ifndef WIN32 ... #else ...
}

Int ProcessExec(const Char *const file, Int count, const Char *const parameters[],
                    const Char *const environment[])
{
  if (environment == NULL)
  {
    if (execvp(file, (Char *const *) parameters) == -1)
    {
      #ifdef WARNING
      Log() << "Process: WARNING! Function execvp() "
            << "for " << Stk(file) << " failed.\n";

      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      return -1;
    }
  }
  else
  {
    if (execve(file, (Char *const *) parameters,
            (Char *const *) environment) == -1)
    {
      #ifdef WARNING
      Log() << "Process: WARNING! Function execve() "
            << "for " << Stk(file) << " failed.\n";

      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      return -1;
    }
  }

  //
  // Add a return statement to make the
  // compiler happy even if, in the case
  // of success, this function doesn't
  // return.
  //

  return 0;
}

Int ProcessDaemon()
{
  #if defined(__sun) || defined(WIN32)
/*
FIXME: Don't know how to detach from terminal on Windows
       and Solaris, but probably this is not required,
       except for testing. In fact we are going to start
       the daemon from init or as a service.
*/
  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessDaemon() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #else // #if defined(__sun) || defined(WIN32) ...

  if (daemon(1, 1) == 0)
  {
    return 1;
  }

  #ifdef WARNING
  Log() << "Process: WARNING! Detach from terminal "
        << "failed.\n";
  #endif

  #ifdef WARNING
  Log() << "Provess: WARNING! Error is " << GetError()
        << " " << Stk(GetErrorString()) << ".\n";
  #endif

  #endif // #if defined(__sun) || defined(WIN32) ... #else ...

  return -1;
}

Int ProcessSignal(Int id, Int signal)
{
  return SignalSend(id, signal);
}

Int ProcessRaise(Int signal)
{
  return SignalRaise(signal);
}

void ProcessSleep(Int ms)
{
  ThreadSleep(ms);
}

void ProcessExit(Int code)
{
  ::exit(code);
}

void ProcessAbort()
{
  ::abort();
}

void ProcessKill()
{
  //
  // In a multithreaded program raise() is
  // equivalent to pthread_kill(), that is
  // not what we want.
  //

  #ifndef WIN32

  SignalSend(ProcessGetId(), SIGKILL);

  #else // #ifndef WIN32 ...

  TerminateProcess(GetCurrentProcess(), EINTR);

  #endif // #ifndef WIN32 ... #else ...

  for (;;)
  {
    ProcessSleep(1000000);
  }
}

//
// Query the NT kernel for information about
// the process. The code is based on a sample
// published on CodeGuru.com by Jurgen Kraus.
//

#ifdef WIN32

#define DYNLOADED_FPTR(ptrname, procname, dllname) \
            FPTR_##procname ptrname = (FPTR_##procname) \
                GetProcAddress(GetModuleHandle(TEXT(#dllname)),#procname);

#define CREATE_DYNFUNC_4(ptrname, procname, dllname, rettype, callconv, a1, a2, a3, a4) \
            typedef rettype (callconv *FPTR_##procname)(a1, a2, a3, a4); \
                DYNLOADED_FPTR(ptrname, procname, dllname);

#define CREATE_DYNFUNC_5(ptrname, procname, dllname, rettype, callconv, a1, a2, a3, a4, a5) \
            typedef rettype (callconv *FPTR_##procname)(a1, a2, a3, a4, a5); \
                DYNLOADED_FPTR(ptrname, procname, dllname);

Int ProcessGetInformation(Int id, PROCESS_BASIC_INFORMATION *information)
{
  HANDLE handle;
  ULONG  length;

  length = sizeof(PROCESS_BASIC_INFORMATION);

  //
  // Create an entry point for the WIN32
  // function NtQueryInformationProcess().
  //

  CREATE_DYNFUNC_5(NtQueryInformationProcess, NtQueryInformationProcess,
                       ntdll, NTSTATUS, __stdcall, HANDLE, PROCESSINFOCLASS,
                           PVOID, ULONG, PULONG);

  //
  // Get the process handle.
  //

  handle = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, (DWORD) id);

  //
  // It may fail due to an invalid id
  // or insufficient privileges.
  //

  if (handle == NULL)
  {
    //
    // Don't print an error if Windows
    // can't find the process handle.
    //

    FileTranslateError();

    if (GetError() != EINVAL || id <= 0)
    {
      #ifdef WARNING
      Log() << "Process: WARNING! Call to OpenProcess() "
            << "failed.\n";
      #endif

      #ifdef WARNING
      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif
    }

    return 0;
  }
  else
  {
    //
    // Gather the requested information.
    //

    Int result = 1;

    if (NtQueryInformationProcess(handle, ProcessBasicInformation,
                                      (void *) information, length, &length) != 0)
    {
      FileTranslateError();

      #ifdef WARNING
      Log() << "Process: WARNING! Query process information "
            << "failed.\n";
      #endif

      #ifdef WARNING
      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      result = -1;
    }

    CloseHandle(handle);

    return result;
  }
}

Int ProcessGetPid(Int id)
{
  PROCESS_BASIC_INFORMATION information;

  HANDLE handle;
  ULONG  length;

  length = sizeof(PROCESS_BASIC_INFORMATION);

  //
  // Create an entry point for the WIN32
  // function NtQueryInformationProcess().
  //

  CREATE_DYNFUNC_5(NtQueryInformationProcess, NtQueryInformationProcess,
                       ntdll, NTSTATUS, __stdcall, HANDLE, PROCESSINFOCLASS,
                           PVOID, ULONG, PULONG);

  //
  // Gather the requested information.
  //

  if (id == ProcessGetId())
  {
    handle = GetCurrentProcess(); 
  }
  else
  {
    handle = HandleCast(id);
  }

  if (NtQueryInformationProcess(handle, ProcessBasicInformation,
                                    (void *) &information, length, &length) != 0)
  {
    //
    // Don't print an error if Windows
    // can't find the process handle.
    //

    FileTranslateError();

    if (GetError() != ESRCH)
    {
      #ifdef WARNING
      Log() << "Process: WARNING! Query pid information "
            << "failed.\n";
      #endif

      #ifdef WARNING
      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif
    }

    return -1;
  }
  else
  {
    return IntCast(information.UniqueProcessId);
  }
}

HANDLE ProcessGetHandle(Int pid)
{
  HANDLE handle;

  handle = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, (DWORD) pid);

  //
  // It may fail due to an invalid id
  // or insufficient privileges.
  //

  if (handle == NULL)
  {
    //
    // Don't print an error if Windows
    // can't find the process handle.
    //

    FileTranslateError();

    if (GetError() != EINVAL)
    {
      #ifdef WARNING
      Log() << "Process: WARNING! Can't open process "
            << "with pid " << pid << ".\n";
      #endif

      #ifdef WARNING
      Log() << "Process: WARNING! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif
    }
  }

  return handle;
}

Int ProcessGetParentPid()
{
  PROCESS_BASIC_INFORMATION information;

  Int id = ProcessGetId();

  Int result = ProcessGetInformation(id, &information);

  if (result == 1)
  {
    return IntCast(information.InheritedFromUniqueProcessId);
  }
  else
  {
    return result;
  }
}

#endif // #ifdef WIN32 ...

Int ProcessIsRunning(Int id)
{
  #ifndef WIN32

  //
  // On Unix use the same implementation
  // as with a child process.
  //

  return ProcessIsChildRunning(id);

  #else // #ifndef WIN32 ...

  //
  // On Windows loop across all processes
  // in the system searching a matching
  // pid.
  //

  Int running = 0;

  CREATE_DYNFUNC_4(NtQuerySystemInformation, NtQuerySystemInformation,
                       ntdll, NTSTATUS, __stdcall, SYSTEM_INFORMATION_CLASS,
                           PVOID, ULONG, PULONG);

  NTSTATUS status;
  ULONG    unused;

  Size size = 65536;

  SYSTEM_PROCESS_INFORMATION *nspi;

  SYSTEM_PROCESS_INFORMATION *spi = (SYSTEM_PROCESS_INFORMATION *) MemoryAlloc(size);

  if (spi == NULL)
  {
    goto ProcessIsRunningError;
  }

  for (;;)
  {
    status = NtQuerySystemInformation(SystemProcessInformation, spi, size, &unused);

    if (status == STATUS_SUCCESS)
    {
      break;
    }
    else if (status != STATUS_INFO_LENGTH_MISMATCH)
    {
      goto ProcessIsRunningError;
    }

    //
    // Grow the buffer to accomodate
    // all the output data.
    //

    size <<= 1;

    if ((spi = (SYSTEM_PROCESS_INFORMATION *) MemoryRealloc(spi, size)) == NULL)
    {
      goto ProcessIsRunningError;
    }

    #ifdef WARNING
    Log() << "Process: WARNING! Reallocated buffer "
          << "with size " << size << ".\n";
    #endif
  }

  //
  // Save the base address so that
  // we can free it later.
  //

  nspi = spi;

  for (;;)
  {
    #ifdef TEST
    Log() << "Process: Checking process id "
          << IntCast(nspi -> UniqueProcessId)
          << ".\n";
    #endif

    if (IntCast(nspi -> UniqueProcessId) == id)
    {
      running = 1;

      break;
    }
    else if (nspi -> NextEntryOffset == 0)
    {
      break;
    }

    nspi = (SYSTEM_PROCESS_INFORMATION *)(nspi -> NextEntryOffset + (PBYTE) nspi);
  }

  #ifdef TEST
  Log() << "Process: Process id " << id << " "
        << (running == 1 ? "found" : "not found")
        << ".\n";
  #endif

  MemoryFree(spi);

  return running;

ProcessIsRunningError:

  FileTranslateError();
  
  #ifdef ERROR
  Log() << "Process: ERROR! Can't list the running "
        << "processes.\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << " " << Stk(GetErrorString()) << ".\n";
  #endif

  MemoryFree(spi);

  return -1;

  #endif // #ifndef WIN32 ... #else ...
}

#ifdef WIN32

Int ProcessWait(Int id, Int *status, Int options)
{
  //
  // Simulate the non-blocking behaviour.
  //

  if ((options & WNOHANG) != 0)
  {
    if (WaitForSingleObject(HandleCast(id), 0) == WAIT_TIMEOUT)
    {
      return 0;
    }
  }

  //
  // This will call GetExitCodeProcess()
  // and store the result in status as
  // the corresponding Unix call.
  //

  return cwait(status, id, 0);
}

#endif // #ifdef WIN32 ...

Int ProcessIsChildRunning(Int id)
{
  #ifndef WIN32

  Int result = SignalSend(id, 0);

  if (result == -1)
  {
    if (GetError() == ESRCH)
    {
      return 0;
    }
    else
    {
      #ifdef ERROR
      Log() << "Process: ERROR! Can't signal the pid "
            << id << ".\n";
      #endif

      #ifdef ERROR
      Log() << "Process: ERROR! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      return -1;
    }
  }

  return 1;

  #else // #ifndef WIN32 ...

  //
  // On Windows check if we can acquire
  // a handle to the process.
  //

  HANDLE handle = ProcessGetHandle(id);

  if (handle == NULL)
  {
    return 0;
  }

  CloseHandle(handle);

  return 1;

  #endif // #ifndef WIN32 ... #else ...
}

Int ProcessGetError(Int status)
{
  if (ProcessIsExited(status) == 1)
  {
    Int code = ProcessGetExitCode(status);

    #ifdef TEST
    Log() << "Process: Process exited with code "
          << Itk(code) << ".\n";
    #endif

    return code;
  }
  else if (ProcessIsSignaled(status) == 1)
  {
    #ifdef TEST

    Int signal = ProcessGetExitSignal(status);

    const Char *name = ProcessGetExitSignalName(signal);

    Log() << "Process: Process died with signal "
          << Itk(signal) << ", " << Stk(name)
          << ".\n";
    #endif

    return EINTR;
  }
  else
  {
    #ifdef WARNING
    Log() << "Process: WARNING! Unrecognized exit "
          << "code " << Itk(status) << ".\n";
    #endif

    return status;
  }
}

//
// We should identify the most common
// Windows exceptions and simulate the
// Unix behavior.
//

Int ProcessIsExited(Int status)
{
  #ifndef WIN32

  return WIFEXITED(status);

  #else

  return 1;

  #endif
}

Int ProcessGetExitCode(Int status)
{
  #ifndef WIN32

  return WEXITSTATUS(status);

  #else

  return status;

  #endif
}

Int ProcessIsSignaled(Int status)
{
  #ifndef WIN32

  return WIFSIGNALED(status);

  #else

  return 0;

  #endif
}

Int ProcessGetExitSignal(Int status)
{
  #ifndef WIN32

  return WTERMSIG(status);

  #else

  return -1;

  #endif
}

const Char *ProcessGetExitSignalName(Int status)
{
  #ifndef WIN32

  return SignalGetName(ProcessGetExitSignal(status));

  #else

  return "Unknown";

  #endif
}

Int ProcessIsStopped(Int status)
{
  #ifndef WIN32

  return WIFSTOPPED(status);

  #else

  return 0;

  #endif
}

Int ProcessGetStopSignal(Int status)
{
  #ifndef WIN32

  return WSTOPSIG(status);

  #else

  return -1;

  #endif
}

const Char *ProcessGetStopSignalName(Int status)
{
  #ifndef WIN32

  return SignalGetName(ProcessGetStopSignal(status));

  #else

  return "Unknown";

  #endif
}

#if defined(__sun) || defined(WIN32)

Int ProcessSetEnvironment(const Char *name, const Char *value)
{
  Int length = StringLength(name) + StringLength(value) + 2;

  Char *environment = new Char[length];

  if (environment == NULL)
  {
    SetError(ENOMEM);

    return -1;
  }

  StringCopy(environment, name);
  StringCat(environment, "=");
  StringCat(environment, value);

  length = putenv(environment);

  //
  // The putenv() function doesn't copy the
  // string for us, so free it only if the
  // call fails.
  //

  if (length != 0)
  {
    delete [] environment;
  }

  return (length != 0 ? -1 : 0); 
}

#endif // #if defined(__sun) || defined(WIN32) ...

Int ProcessTranslatePriority(Int value)
{
  #ifndef WIN32

  switch (value)
  {
    case ProcessPriorityLow:
    {
      return PROCESS_PRIORITY_LOW;
    }
    case ProcessPriorityNormal:
    {
      return PROCESS_PRIORITY_NORMAL;
    }
    case ProcessPriorityHigh:
    {
      return PROCESS_PRIORITY_HIGH;
    }
    case ProcessPriorityRealtime:
    {
      return PROCESS_PRIORITY_REALTIME;
    }
    default:
    {
      //
      // Use whatever value was provided
      // by the user. Fail accordingly
      // if not supported by the OS.
      //

      return value;
    }
  }

  #else // #ifndef WIN32 ...

  switch (value)
  {
    case ProcessPriorityLow:
    {
      return IDLE_PRIORITY_CLASS;
    }
    case ProcessPriorityNormal:
    {
      return NORMAL_PRIORITY_CLASS;
    }
    case ProcessPriorityHigh:
    {
      return HIGH_PRIORITY_CLASS;
    }
    case ProcessPriorityRealtime:
    {
      return REALTIME_PRIORITY_CLASS;
    }
    default:
    {
      SetError(EINVAL);

      #ifdef ERROR
      Log() << "Process: ERROR! Can't set priority to "
            << value << ".\n";
      #endif

      #ifdef ERROR
      Log() << "Process: ERROR! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      return GetPriorityClass(GetCurrentProcess());
    }
  }

  #endif
}

Int ProcessSetPriority(Int value)
{
  #ifndef WIN32

  value = ProcessTranslatePriority(value);

  if (setpriority(PRIO_PROCESS, 0, value) == -1)
  {
    #ifdef ERROR
    Log() << "Process: ERROR! Can't set priority to "
          << value << ".\n";
    #endif

    #ifdef ERROR
    Log() << "Process: ERROR! Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    return -1;
  }

  return 1;

  #else // #ifndef WIN32 ...

  #ifdef WARNING
  Log() << "Process: WARNING! Function ProcessSetPriority() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #endif // #ifndef WIN32 ... #else ...
}

Int ProcessChangePriority(Int value)
{
  //
  // The following applies to Cygwin: "Cygwin processes can be 
  // set to IDLE_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS, HIGH_-
  // PRIORITY_CLASS, or REALTIME_PRIORITY_CLASS with the nice
  // call. If you pass a positive number to nice(), then the
  // priority level will decrease by one (within the above list
  // of priorities). A negative number would make it increase
  // by one. It is not possible to change it by more than one
  // at a time without making repeated calls".
  //

  #ifndef WIN32

  SetError(0);

  if (nice(value) < 0 && errno != 0)
  {
    #ifdef ERROR
    Log() << "Process: ERROR! Can't change priority by "
          << value << ".\n";
    #endif

    #ifdef ERROR
    Log() << "Process: ERROR! Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    return -1;
  }

  return 1;

  #else // #ifndef WIN32 ...

  #ifdef WARNING
  Log() << "Process: WARNING! Function ProcessChangePriority() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #endif // #ifndef WIN32 ... #else ...
}

const Char *ProcessGetComputerName()
{
  //
  // In some Windows versions the SMB service does
  // not bind to the localhost interface. In this
  // case we must find the actual IP of the local
  // computer and fall back to the localhost name
  // only if can't find the computer name in the
  // environment. A better approach would require
  // that we bind to localhost and, on failure, we
  // try the other IPs.
  //

  #if defined(__CYGWIN32__) || defined(WIN32)

  Char *hostname = ProcessGetEnvironment("COMPUTERNAME");

  #else

  Char *hostname = ProcessGetEnvironment("HOSTNAME");

  #endif

  if (hostname != NULL)
  {
    return hostname;
  }

  return ProcessGetLocalhostName();
}

const Char *ProcessGetLocalhostName()
{
  #if defined(__CYGWIN32__) || defined(WIN32)

  return "127.0.0.1";

  #else

  return "localhost";

  #endif
}

Int ProcessNameIsLocal(const Char *name)
{
  if (StringCompare(name, "localhost") == 0 ||
          StringCompare(name, "localhost.localdomain") == 0 ||
              StringCompare(name, "127.0.0.1") == 0)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

const Char *ProcessGetTemporaryDirectory()
{
  //
  // Check the temporary directory name
  // in the environment.
  //

  const Char *temp = ProcessGetEnvironment("TEMP");

  if (temp == NULL || *temp == '\0')
  {
    #ifdef TEST
    Log() << "Process: No environment for TEMP.\n";
    #endif

    temp = ProcessGetEnvironment("TMP");

    if (temp == NULL || *temp == '\0')
    {
      #ifdef TEST
      Log() << "Process: No environment for TMP.\n";
      #endif

      #ifndef WIN32

      temp = "/tmp";

      #else // #ifndef WIN32 ...

      temp = ProcessGetEnvironment("USERPROFILE");

      if (temp == NULL || *temp == '\0')
      {
        #ifdef TEST
        Log() << "Process: No environment for USERPROFILE.\n";
        #endif

        temp = ProcessGetEnvironment("HOMEDRIVE");

        if (temp == NULL || *temp == '\0')
        {
          #ifdef TEST
          Log() << "Process: No environment for HOMEDRIVE.\n";
          #endif

          //
          // Put the temporary files in the root
          // directory of the 'C:' drive. This
          // should never happen.
          //

          temp = "C:\\";
        }
      }

      #ifdef WARNING
      Log() << "Process: WARNING! Assuming temporary "
            << "directory in " << Stk(temp) << ".\n";
      #endif

      #endif // #ifndef WIN32 ... #else ...
    }
  }

  #ifdef TEST
  Log() << "Process: Returning temporary directory "
        << Stk(temp) << ".\n";
  #endif

  return temp;
}

Char *ProcessGetTemporaryName(const Char *directory, const Char *prefix)
{
  #ifndef WIN32

  if (prefix == NULL)
  {
    prefix = "T-";
  }

  Char name[1024];

  Timeval ts = getTimeval();

  unsigned Long suffix = ts.tv_usec * ThreadGetPid();

  StringSizePrint(name, 1024, "%s%lu", prefix, suffix);

  *(name + 1024 - 1) = '\0';

  #ifdef TEST
  Log() << "Process: Generated name " << Stk(name) << ".\n";
  #endif

  Int length = StringLength(name) + 1;

  //
  // Determine where to create the file.
  // If the directory string is empty,
  // then use the current directory.
  //

  Char *path;

  if (directory == NULL)
  {
    const Char *temp;

    temp = ProcessGetTemporaryDirectory();

    length += StringLength(temp) + 1;

    path = (Char *) MemoryAlloc(length);

    if (path == NULL)
    {
      goto ProcessGetTemporaryNameError;
    }

    #ifdef TEST
    Log() << "Process: Returning temporary name "
          << "in directory " << Stk(temp)
          << ".\n";
    #endif

    StringCopy(path, temp);
    StringCat(path, DirSlashString);
    StringCat(path, name);
  }
  else if (*directory == '\0')
  {
    #ifdef WARNING
    Log() << "Process: WARNING! Returning temporary "
          << "name in current directory.\n";
    #endif

    path = (Char *) MemoryAlloc(length);

    if (path == NULL)
    {
      goto ProcessGetTemporaryNameError;
    }

    StringCopy(path, name);
  }
  else
  {
    #ifdef TEST
    Log() << "Process: Returning temporary name "
          << "in directory " << Stk(directory)
          << ".\n";
    #endif

    length += StringLength(directory) + 1;

    path = (Char *) MemoryAlloc(length);

    if (path == NULL)
    {
      goto ProcessGetTemporaryNameError;
    }

    StringCopy(path, directory);
    StringCat(path, DirSlashString);
    StringCat(path, name);
  }

  return path;

ProcessGetTemporaryNameError:

  #ifdef ERROR
  Log() << "Process: ERROR! Can't create temporary "
        << "name.\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << " " << Stk(GetErrorString()) << ".\n";
  #endif

  return NULL;

  #else // #ifndef WIN32 ...

  //
  // Use the ISO C++ conformant function.
  //

  return ::_tempnam(directory, prefix);

  #endif // #ifndef WIN32 ... #else ...
}

Int ProcessParseArgs(Int argc, Char * const *argv, const Char *args)
{
  return BSDGetopt(argc, argv, args);
}

void ProcessPurgeArg(Char *arg)
{
  Char *next = arg;

  while (*next != '\0')
  {
    //
    // Use the following replacements:
    //
    // ' '   0x1f   ASCII US (unit separator).
    // '='   0x1d   ASCII GS (group separator).
    // ','   0x1e   ASCII RS (record separator).
    //

    switch (*next)
    {
      case ' ':
      {
        *next = 0x1f;

        break;
      }
      case '=':
      {
        *next = 0x1d;

        break;
      }
      case ',':
      {
        *next = 0x1e;

        break;
      }
    }

    next++;
  }
}

void ProcessUnpurgeArg(Char *arg)
{
  Char *next = arg;

  while (*next != '\0')
  {
    switch (*next)
    {
      case 0x1f:
      {
        *next = ' ';

        break;
      }
      case 0x1d:
      {
        *next = '=';

        break;
      }
      case 0x1e:
      {
        *next = ',';

        break;
      }
    }

    next++;
  }
}

Int ProcessCheckArg(const Char *arg, Int length)
{
  if (arg == NULL)
  {
    return -1;
  }
  else if (StringFind(arg, "=") != NULL ||
               StringFind(arg, ",") != NULL)
  {
    return -1;
  }
  else if (length != -1)
  {
    if (length == 0)
    {
      length = 1024;
    }

    if (StringLength(arg) >= (Unsigned) length)
    {
      return -1;
    }
  }

  return 1;
}

Int ProcessCheckArg(const Char *name, const Char *arg, Int length)
{
  if (ProcessCheckArg(name) != 1 ||
          ProcessCheckArg(arg, length) != 1)
  {
    return -1;
  }

  return 1;
}

Int ProcessParseArg(const Char *arg)
{
  Int number;

  if (CharIsDigit(*arg) == 0 ||
          (number = StringToInt(arg)) < 0)
  {
    return -1;
  }

  return number;
}

Int ProcessCompareArg(const Char *first, const Char *second)
{
  #ifdef TEST
  Log() << "Process: First '" << *first
        << "' second '" << *second << "'.\n";
  #endif

  while (*second != '\0')
  {
    #ifdef TEST
    Log() << "Process: Next first '" << *first
          << "' next second '" << *second << "'.\n";
    #endif

    if (*first != *second || *first == '\0' ||
            *first == ',')
    {
      return 0;
    }

    first++;
    second++;
  }

  #ifdef TEST
  Log() << "Process: Last first '" << *first
        << "' next second '" << *second << "'.\n";
  #endif

  if (*first != '\0' && *first != ',')
  {
    return 0;
  }

  return 1;
}

Int ProcessReplaceArg(Char *&value, const Char *symbol, const Char *replace)
{
  #ifdef TEST
  Log() << "Process: Replacing " << Stk(symbol)
        << " with " << Stk(replace) << ".\n";

  Log() << "Process: In string " << Stk(value)
        << ".\n";
  #endif

  if (StringEmpty(symbol) == 0 &&
          StringEmpty(replace) == 0)
  {
    return StringHeadReplace(value, symbol, replace);
  }

  return 0;
}

Int ProcessGetUserId()
{
  #ifndef WIN32

  return getuid();

  #else// #ifndef WIN32 ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessGetUserId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #endif // #ifndef WIN32 ... #else ...
}

Int ProcessGetGroupId()
{
  #ifndef WIN32

  return getgid();

  #else // #ifndef WIN32 ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessGetGroupId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #endif // #ifndef WIN32 ... #else ...
}

Int ProcessSetUserId(Int id)
{
  #ifndef WIN32

  if (setuid(id) == 0)
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "Process: ERROR! Can't set user id to "
        << id << ".\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << ", " << GetErrorString() << ".\n";
  #endif

  #else // #ifndef WIN32 ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessSetUserId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif // #ifndef WIN32 ... #else ...

  return -1;
}

Int ProcessSetGroupId(Int id)
{
  #ifndef WIN32

  if (setgid(id) == 0)
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "Process: ERROR! Can't set group id to "
        << id << ".\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << ", " << GetErrorString() << ".\n";
  #endif

  #else // #ifndef WIN32 ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessSetGroupId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif // #ifndef WIN32 ... #else ...

  return -1;
}

//
// Return the system information about the
// user or the group.
//

#if !defined(WIN32) && !defined(__BIONIC__)

typedef struct passwd ProcessUserInformation;
typedef struct group  ProcessGroupInformation;

Int ProcessGetUserInformation(const Char *name, ProcessUserInformation *ui, Char *&buffer)
{
  //
  // Get the actual value from the
  // system or allocate a sensible
  // default.
  //

  Long result;

  Size size = 16384;

  if ((result = sysconf(_SC_GETPW_R_SIZE_MAX)) != -1)
  {
    size = result;
  }

  buffer = new Char[size];

  ProcessUserInformation *pui;

  result = getpwnam_r(name, ui, buffer, size, &pui);

  if (pui == NULL)
  {
    if (result == 0)
    {
      SetError(ENOENT);
    }
    else
    {
      SetError(result);
    }

    delete [] buffer;
  }
  else
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "Process: ERROR! Can't get information "
        << "for user " << Stk(name) << ".\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << ", " << GetErrorString() << ".\n";
  #endif

  return -1;
}

Int ProcessGetGroupInformation(const Char *name, ProcessGroupInformation *gi, Char *&buffer)
{
  Long result;

  Size size = 16384;

  if ((result = sysconf(_SC_GETGR_R_SIZE_MAX)) != -1)
  {
    size = result;
  }

  buffer = new Char[size];

  ProcessGroupInformation *pgi;

  result = getgrnam_r(name, gi, buffer, size, &pgi);

  if (pgi == NULL)
  {
    if (result == 0)
    {
      SetError(ENOENT);
    }
    else
    {
      SetError(result);
    }

    delete [] buffer;
  }
  else
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "Process: ERROR! Can't get information "
        << "for group " << Stk(name) << ".\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << ", " << GetErrorString() << ".\n";
  #endif

  return -1;
}

Int ProcessGetUserInformation(Int id, ProcessUserInformation *ui, Char *&buffer)
{
  Long result;

  Size size = 16384;

  if ((result = sysconf(_SC_GETPW_R_SIZE_MAX)) != -1)
  {
    size = result;
  }

  buffer = new Char[size];

  ProcessUserInformation *pui;

  result = getpwuid_r(id, ui, buffer, size, &pui);

  if (pui == NULL)
  {
    if (result == 0)
    {
      SetError(ENOENT);
    }
    else
    {
      SetError(result);
    }

    delete [] buffer;
  }
  else
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "Process: ERROR! Can't get information "
        << "for user id " << Itk(id) << ".\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << ", " << GetErrorString() << ".\n";
  #endif

  return -1;
}

Int ProcessGetGroupInformation(Int id, ProcessGroupInformation *gi, Char *&buffer)
{
  Long result;

  Size size = 16384;

  if ((result = sysconf(_SC_GETGR_R_SIZE_MAX)) != -1)
  {
    size = result;
  }

  buffer = new Char[size];

  ProcessGroupInformation *pgi;

  result = getgrgid_r(id, gi, buffer, size, &pgi);

  if (pgi == NULL)
  {
    if (result == 0)
    {
      SetError(ENOENT);
    }
    else
    {
      SetError(result);
    }

    delete [] buffer;
  }
  else
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "Process: ERROR! Can't get information "
        << "for group id " << Itk(id) << ".\n";
  #endif

  #ifdef ERROR
  Log() << "Process: ERROR! Error is " << GetError()
        << ", " << GetErrorString() << ".\n";
  #endif

  return -1;
}

#endif // #if !defined(WIN32) && !defined(__BIONIC__) ...
 
//
// Get the numeric id of the user or group,
// given the name string.
//

Int ProcessGetUserId(const Char *name)
{
  #ifdef TEST
  Log() << "Process: Getting user id for "
        << "user " << Stk(name) << ".\n";
  #endif

  #if !defined(WIN32) && !defined(__BIONIC__)

  Char *buffer;

  ProcessUserInformation ui;

  if (ProcessGetUserInformation(name, &ui, buffer) == 1)
  {
    Int uid = ui.pw_uid;

    #ifdef TEST
    Log() << "Process: Returning uid " << uid
          << " for user " << Stk(name) << ".\n";
    #endif

    delete [] buffer;

    return uid;
  }

  #else // #if !defined(WIN32) && !defined(__BIONIC__) ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessGetUserId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif // #if !defined(WIN32) && !defined(__BIONIC__) ... #else ...

  return -1;
}

Int ProcessGetGroupId(const Char *name)
{
  #ifdef TEST
  Log() << "Process: Getting group id for "
        << "group " << Stk(name) << ".\n";
  #endif

  #if !defined(WIN32) && !defined(__BIONIC__)

  Char *buffer;

  ProcessGroupInformation gi;

  if (ProcessGetGroupInformation(name, &gi, buffer) == 1)
  {
    Int gid = gi.gr_gid;

    #ifdef TEST
    Log() << "Process: Returning gid " << gid
          << " for group " << Stk(name) << ".\n";
    #endif

    delete [] buffer;

    return gid;
  }

  #else // #if !defined(WIN32) && !defined(__BIONIC__) ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessGetGroupId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif // #if !defined(WIN32) && !defined(__BIONIC__) ... #else ...

  return -1;
}

//
// Get the name of the user or group, given
// the id. On success, a new string stores
// the value and the old string is deleted.
//

Int ProcessGetUserName(Char *&name, Int id)
{
  #ifdef TEST
  Log() << "Process: Getting user name for "
        << "user id " << Itk(id) << ".\n";
  #endif

  #if !defined(WIN32) && !defined(__BIONIC__)

  Char *buffer;

  ProcessUserInformation ui;

  if (ProcessGetUserInformation(id, &ui, buffer) == 1)
  {
    StringSet(name, ui.pw_name);

    #ifdef TEST
    Log() << "Process: Returning name " << name
          << " for user id " << Itk(id) << ".\n";
    #endif

    delete [] buffer;

    return 1;
  }

  #else // #if !defined(WIN32) && !defined(__BIONIC__) ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessGetUserId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif // #if !defined(WIN32) && !defined(__BIONIC__) ... #else ...

  return -1;
}

Int ProcessGetGroupName(Char *&name, Int id)
{
  #ifdef TEST
  Log() << "Process: Getting group name for "
        << "group id " << Itk(id) << ".\n";
  #endif

  #if !defined(WIN32) && !defined(__BIONIC__)

  Char *buffer;

  ProcessGroupInformation gi;

  if (ProcessGetGroupInformation(id, &gi, buffer) == 1)
  {
    StringSet(name, gi.gr_name);

    #ifdef TEST
    Log() << "Process: Returning name " << name
          << " for group id " << Itk(id) << ".\n";
    #endif

    delete [] buffer;

    return 1;
  }

  #else // #if !defined(WIN32) && !defined(__BIONIC__) ...

  #ifdef ERROR
  Log() << "Process: ERROR! Function ProcessGetGroupId() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif // #if !defined(WIN32) && !defined(__BIONIC__) ... #else ...

  return -1;
}
