/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Process_H
#define Process_H

#ifndef WIN32

#include <sys/wait.h>

#else // #ifdef WIN32 ...

#include <process.h>

#endif // #ifdef WIN32 ... #else ...

#include "File.h"

#define ProcessId  pid_t
#define ProcessPid pid_t

//
// Use the most appropriate default based
// on the host operating system.
//

#define ProcessPriorityLow       100
#define ProcessPriorityNormal    101
#define ProcessPriorityHigh      102
#define ProcessPriorityRealtime  103

//
// Return an error on Windows. On Cygwin the
// fork is repeated multiple times, since it
// can fail at random on some recent versions
// of Windows.
//

Int ProcessFork();

//
// Create a new process. On Unix-like systems, the
// function is implemented by a fork and a execvp.
// The new process will use the standard descriptors
// passed as parameters, if valid. Otherwise it will
// inherit the descriptors from the parent process.
// If the preserve flag is not set, the child will
// drop the additional privileges inherited from the
// parent. On Windows, the interactive flag tells if
// the console window must be displayed or not for
// the new process.
//

Int ProcessCreate(const Char *const file, const Char *const parameters[],
                      const Char *const environment[], Int fdin, Int fdout,
                          Int fderr, Int priority = 0, Int preserve = 0,
                              Int interactive = 0, Int inherit = 1,
                                  Int group = 0);

//
// Compatibility versions of the previous
// function.
//

inline Int ProcessCreate(const Char *const file, Int count,
                             const Char *const parameters[],
                                 Int interactive = 0)
{
  return ProcessCreate(file, parameters, NULL, -1, -1, -1,
                           0, 0, interactive);
}

inline Int ProcessCreate(const Char *const file, Int count,
                             const Char *const parameters[], Int fdin, Int fdout,
                                 Int fderr, Int preserve = 0, Int interactive = 0,
                                     Int inherit = 1, Int group = 0)
{
  return ProcessCreate(file, parameters, NULL, fdin, fdout, fderr, 0,
                           preserve, interactive, inherit, group);
}

//
// Replace the image of the current process
// with the given program.
//

Int ProcessExec(const Char *const file, Int count,
                    const Char *const parameters[],
                        const Char *const environment[] = NULL);

//
// Detach itself from the controlling terminal
// and run in the background as system daemon.
//

Int ProcessDaemon();

//
// Send a signal to the process, by eventually
// terminating it.
//

Int ProcessSignal(Int id, Int signal);

//
// Send a signal to the current process.
//

Int ProcessRaise(Int signal);

//
// Sleep for the given amount of milliseconds
// in the current thread.
//

void ProcessSleep(Int ms);

//
// Exit from the current program by returning
// an exit code.
//

void NORETURN ProcessExit(Int code);

//
// Exit from the current program, depending on
// the platform, by generating a core dump or
// by invoking the JIT debugger.
//

void NORETURN ProcessAbort();

//
// Terminate the process immediately.
//

void NORETURN ProcessKill();

//
// Return the id of the current process.
//

#define ProcessGetId()         getpid()

inline Int ProcessGetPid()
{
  return ProcessGetId();
}

//
// Return the system pid based on the id. The
// id is normally a handle on Windows and the
// real process id is different.
//

#ifndef WIN32

inline Int ProcessGetPid(Int id)
{
  return id;
}

#else // #ifndef WIN32 ...

Int ProcessGetPid(Int id);

//
// Return a handle of an arbitrary external
// process based on the unique process id.
//

HANDLE ProcessGetHandle(Int pid);

#endif // #ifndef WIN32 ... #else ...

//
// Return the unique system id of the parent
// of the current process. Emulate the Posix
// function on Windows.
//

#ifndef WIN32

#define ProcessGetParentPid()  getppid()

#else // #ifndef WIN32 ...

Int ProcessGetParentPid();

#endif // #ifndef WIN32 ... #else ...

//
// Test whether a process with the given id
// is running in the system. Return 1 if such
// a process exists, 0 if not, -1 in the case
// of an error.
//

Int ProcessIsRunning(Int id);

//
// Test the exit code of the child process, by
// optionally waiting until the child process
// terminates.
//

#ifndef WIN32

#define ProcessWait(id, status, options)  waitpid((id), (status), (options))

#else // #ifndef WIN32 ...

#define WNOHANG   1
#define WUNTRACED 2

Int ProcessWait(Int id, Int *status, Int options);

#endif // #ifndef WIN32 ... #else ...

//
// Test whether the child process is running.
// Return 1 if the child did not exit yet, 0
// if the child exited, -1 in the case of an
// error.
//

Int ProcessIsChildRunning(Int id);

//
// Find out the error code based on the exit
// status.
//

Int ProcessGetError(Int status);

//
// True if the child has exited by calling exit
// or by returning from the main function.
//

Int ProcessIsExited(Int status);

//
// If the above function is true, return the exit
// code of the program.
//

Int ProcessGetExitCode(Int status);

//
// Return true if the child was terminated by a
// signal.
//

Int ProcessIsSignaled(Int status);

//
// If the above function is true, return the signal
// number that terminated the program.
//

Int ProcessGetExitSignal(Int status);

const Char *ProcessGetExitSignalName(Int status);

//
// True if the child is still running but it was
// stopped by a signal.
//

Int ProcessIsStopped(Int status);

//
// The signal that determined the child to stop.
//

Int ProcessGetStopSignal(Int status);

const Char *ProcessGetStopSignalName(Int status);

//
// Environment manipulations utilities.
//

#define ProcessGetEnvironment(name)  getenv((name))

#if defined(__sun) || defined(WIN32)

Int ProcessSetEnvironment(const Char *name, const Char *value);

#else // #if defined(__sun) || defined(WIN32) ...

#define ProcessSetEnvironment(name, value)  setenv((name), (value), 1)

#endif // #if defined(__sun) || defined(WIN32) ... #else ...

//
// Change the scheduling priority of the process.
//

Int ProcessChangePriority(Int value);

//
// Set the scheduling priority of the process.
//

Int ProcessSetPriority(Int value);

//
// Return a platform-specific value.
//

Int ProcessTranslatePriority(Int value);

//
// Query the environment to find the name of the
// local computer or the host name to be used to
// refer to the loopback network interface.
// 

const Char *ProcessGetComputerName();
const Char *ProcessGetLocalhostName();

//
// Return true if the provided name refer to
// the local computer.
//

Int ProcessNameIsLocal(const Char *name);

//
// Check the temporary directory name in the
// environmentor return a default.
//

const Char *ProcessGetTemporaryDirectory();

//
// Return a random name that can be used to
// create a temporary file.
//

Char *ProcessGetTemporaryName(const Char *directory, const Char *prefix);

inline Char *ProcessGetTemporaryName(const Char *prefix)
{
  return ProcessGetTemporaryName(NULL, prefix);
}

//
// Parse the command line arguments by using
// the BSD getopt() function.
//

Int ProcessParseArgs(Int argc, Char * const *argv, const Char *args);

//
// Replace the characters used internally to
// pass string options between programs.
//

void ProcessPurgeArg(Char *arg);
void ProcessUnpurgeArg(Char *arg);

//
// Perform a basic validation of the string
// argument and ensure that the length of the
// string doesn't exceed length - 1, or 1023,
// if a 0 length is given.
//

Int ProcessCheckArg(const Char *arg, Int length = 0);
Int ProcessCheckArg(const Char *name, const Char *arg, Int length = 0);

//
// Validate a numeric argument so that it
// results in a positive integer.
//

Int ProcessParseArg(const Char *arg);

//
// Compare two strings where the second is
// the string to match and the first is
// either null terminated or delimited by
// a comma character.
//

Int ProcessCompareArg(const Char *first, const Char *second);

//
// If the given value is not an empty string,
// replace the symbol string with the value
// in the arg string.
//

Int ProcessReplaceArg(Char *&arg, const Char *symbol, const Char *value);

//
// Manipulate the user and group id of the
// current process.
//

Int ProcessGetUserId();
Int ProcessGetGroupId();

Int ProcessSetUserId(Int id);
Int ProcessSetGroupId(Int id);

//
// Get the numeric id of the user or group
// given the name string.
//

Int ProcessGetUserId(const Char *name);
Int ProcessGetGroupId(const Char *name);

//
// Get the name of the user or group given
// the id. On success, a new string stores
// the value and the old string is deleted.
//

Int ProcessGetUserName(Char *&name, Int id);
Int ProcessGetGroupName(Char *&name, Int id);

//
// Print the current process id.
//

#define Pid()  ProcessGetId()

#endif /* Process_H */
