/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include <stdlib.h>

#include "Time.h"
#include "Thread.h"
#include "Random.h"
#include "Error.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// Global seed for the random number
// generator.
//

#ifndef WIN32

static SYSTEM Unsigned Int RandomSeed;

#endif

//
// Avoid the compilation warning due
// to the value taken from the stack.
//

#if (__GNUC__ > 4) || (__GNUC__ == 4 && __GNUC_MINOR__ > 1)

#pragma GCC diagnostic ignored "-Wuninitialized"

#endif

//
// Set up the random number generator.
//

void RandomInit()
{
  //
  // Add a tiny bit of entropy by starting
  // with whatever is in the stack. Still a
  // very poor way to initialize the random
  // generator.
  //

  Char garbage[sizeof(Unsigned Int)];

  Unsigned Int *pointer = (Unsigned Int *) garbage;

  Unsigned Int seed = *pointer;

  if (seed == 0)
  {
    seed = 1;
  }

  //
  // We may need to generate a fake thread
  // pid on some platforms.
  //

  Int pid = ThreadGetPid();

  if (pid != 0)
  {
    seed *= pid;
  }

  //
  // Give a chance to the number to be 0.
  //

  Timeval time = getTimeval();

  seed *= time.tv_usec;

  #ifdef TEST
  Log() << "Random: Generated random seed "
        << seed << ".\n";
  #endif

  #ifndef WIN32

  RandomSeed = seed;

  #else

  srand(seed);

  #endif
}

Int RandomGet()
{
  //
  // Check if RAND_MAX is limited to 16
  // bits and combine 2 numbers together.
  //

  #if !defined( WIN32 ) && !defined(__BIONIC__)

  #if RAND_MAX < 0x7fffffff

  Int number = (rand_r(&RandomSeed) << 16) | rand_r(&RandomSeed);

  #else 

  Int number = rand_r(&RandomSeed);

  #endif

  #else // #ifndef WIN32 ...

  #ifdef __BIONIC__

  #ifdef WARNING
  Log() << "RandomGet: WARNING! Reentrant randr_r() "
        << "is not implemented on Bionic.\n";
  #endif

  LogWarning() << "Reentrant randr_r() is not "
               << "implemented on Bionic.\n";

  SetError(ENOSYS);

  Int number = -1;

  #else

  Int number = (rand() << 16) | rand();

  #endif // #ifndef WIN32 ... #else ...

  #ifdef TEST
  Log() << "Random: Generated random number "
        << number << ".\n";
  #endif

  #endif

  return number;
}

Md5Digest *RandomGetMd5(Md5Digest *digest)
{
  Int values[4];

  for (Int i = 0; i < 4; i++)
  {
    #if !defined( WIN32 ) && !defined(__BIONIC__)

    values[i] = rand_r(&RandomSeed);

    #else

    #ifdef __BIONIC__

    #ifdef WARNING
    Log() << "RandomGetMd5: WARNING! Reentrant randr_r() "
          << "is not implemented on Bionic.\n";
    #endif

    LogWarning() << "Reentrant randr_r() is not "
                 << "implemented on Bionic.\n";

    SetError(ENOSYS);

    values[i] = -1;

    #else

    values[i] = rand();

    #endif

    #endif
  }

  Md5Get(digest, (Md5Data *) &values, sizeof(values));

  #ifdef TEST

  Char string[Md5Length * 2 + 1];

  Md5GetString(digest, string);

  Log() << "Random: Generated random digest "
        << string << ".\n";
  #endif

  return digest;
}
