/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

#include "Log.h"
#include "Error.h"
#include "String.h"
#include "Signal.h"
#include "Memory.h"
#include "Thread.h"

#ifdef WIN32

Int SignalEmptyMask(const SignalMask *mask)
{
  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalEmptyMask() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int SignalFillMask(const SignalMask *mask)
{
  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalFillMask() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int SignalAddMask(SignalMask *mask, Int signal)
{
  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalAddMask() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int SignalRemoveMask(SignalMask *mask, Int signal)
{
  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalRemoveMask() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int SignalCheckMask(const SignalMask *mask, Int signal)
{
  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalCheckMask() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int SignalSend(ProcessId pid, Int signal)
{
  if (signal != SIGKILL)
  {
    #ifdef WARNING
    Log() << "Signal: WARNING! Can't send signal " << signal
          << " '" << SignalGetName(signal) << "'.\n";
    #endif

    SetError(ENOSYS);

    return -1;
  }

  SetError(0);

  Int result = TerminateProcess((HANDLE) pid, 127);

  //
  // The system error seems to be 5
  // if the process is not running.
  // 

  if (result == 0)
  {
    if (GetError() == 5)
    {
      SetError(ESRCH);
    }
    else
    {
      #ifdef ERROR
      Log() << "Signal: ERROR! Function TerminateProcess() "
            << "failed. Error is " << GetError() << " '"
            << GetErrorString() << "'.\n";
      #endif
    }

    return -1;
  }

  return 1;
}

Int SignalRaise(Int signal)
{
  return SignalSend(ProcessGetPid(), signal);
}

#endif // #ifdef WIN32 ...

Int SignalSetMask(Int how, const SignalMask *mask, SignalMask *oldmask)
{
  Int result;

  #ifndef WIN32

  //
  // On Mac OSX pthread_sigmask() doesn't
  // appear to work in the main thread, so
  // use sigprocmask().
  //

  #ifdef __APPLE__

  #ifdef TEST
  Log() << "Signal: Assuming pthread_sigmask() in a "
        << "child thread.\n";
  #endif

  result = pthread_sigmask(how, mask, oldmask);

  #else // #ifdef __APPLE__ ...

  result = pthread_sigmask(how, mask, oldmask);

  #endif // #ifdef __APPLE__ ... #else ...

  if (result != 0)
  {
    #ifdef ERROR
    Log() << "Signal: ERROR! Function pthread_sigmask "
          << "failed. Error is " << GetError() << " '"
          << GetErrorString() << "'.\n";
    #endif
  }

  #else // #ifndef WIN32 ...

  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalSetMask() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  result = -1;

  #endif // #ifndef WIN32 ... #else ...

  return result;
}

void SignalLogMask(const SignalMask *mask, const Char *label, Logger *logger)
{
  if (label == NULL)
  {
    label = "Signal";
  }

  if (logger == NULL)
  {
    logger = GetLogger();
  }

  for (Int i = 0; i < 32; i++)
  {
    if (StringCompare(SignalGetName(i), "None") != 0)
    {
      Log(logger) << label << "/Signal: " << i
                  << ", " << SignalGetName(i) << "\t";

      //
      // On Android this may be decleared
      // non-const.
      //

      if (SignalCheckMask((SignalMask *) mask, i) != 0)
      {
        LogMore(logger) << "Yes\n";
      }
      else
      {
        LogMore(logger) << "No\n";
      }
    }
  }
}

Int SignalSetAction(Int signal, const SignalAction *action,
                        SignalAction *oldaction)
{
  #ifndef WIN32

  if (sigaction(signal, action, oldaction) < 0)
  {
    #ifdef ERROR
    Log() << "Signal: ERROR! Call to sigaction failed. "
          << "Error is " << GetError() << " '" << GetErrorString()
          << "'.\n";
    #endif

    return -1;
  }

  return 0;

  #else

  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalSetAction() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #endif
}

extern Int SignalSetTimer(Int which, const SignalTimer *value,
                              SignalTimer *oldvalue)
{
  #ifndef WIN32

  #if __GNUC__ < 3

  if (setitimer(which, (SignalTimer *) value, oldvalue) == -1)

  #else

  if (setitimer(which, value, oldvalue) == -1)

  #endif
  {
    #ifdef ERROR
    Log() << "Signal: ERROR! Call to setitimer failed. "
          << "Error is " << GetError() << " '"
          << GetErrorString() << "'.\n";
    #endif

    return -1;
  }

  return 0;

  #else

  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalSetTimer() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #endif
}

void SignalEmptyAction(SignalAction *action)
{
  #ifndef WIN32

  MemorySet(action, 0, sizeof(SignalAction));

  #else

  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalEmptyAction() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif
}

void SignalEmptyTimer(SignalTimer *timer)
{
  #ifndef WIN32

  MemorySet(timer, 0, sizeof(SignalTimer));

  #else

  #ifdef WARNING
  Log() << "Signal: WARNING! Function SignalEmptyTimer() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  #endif
}

SignalFunction SignalSet(Int value, SignalFunction function)
{
  //
  // Some versions of Windows omit to
  // set errno in the case of invalid
  // signals.
  // 

  #ifdef WIN32

  SetError(EINVAL);

  #endif

  SignalFunction previous = signal(value, function);

  if (previous == SIG_ERR)
  {
    #ifdef ERROR
    Log() << "Signal: ERROR! Call to signal failed. "
          << "Error is " << GetError() << " '"
          << GetErrorString() << "'.\n";
    #endif

    return SIG_DFL;
  }

  return previous;
}

const Char *SignalGetName(Int signal)
{
  switch (signal)
  {
    case SIGCHLD:
    {
      return "SIGCHLD";
    }
    case SIGCONT:
    {
      return "SIGCONT";
    }
    case SIGSTOP:
    {
      return "SIGSTOP";
    }
    case SIGUSR1:
    {
      return "SIGUSR1";
    }
    case SIGUSR2:
    {
      return "SIGUSR2";
    }
    case SIGHUP:
    {
      return "SIGHUP";
    }
    case SIGINT:
    {
      return "SIGINT";
    }
    case SIGTERM:
    {
      return "SIGTERM";
    }
    case SIGPIPE:
    {
      return "SIGPIPE";
    }
    case SIGALRM:
    {
      return "SIGALRM";
    }
    case SIGQUIT:
    {
      return "SIGQUIT";
    }
    case SIGVTALRM:
    {
      return "SIGVTALRM";
    }
    case SIGWINCH:
    {
      return "SIGWINCH";
    }
    case SIGPOLL:
    {
      return "SIGPOLL";
    }
    case SIGTSTP:
    {
      return "SIGTSTP";
    }
    case SIGABRT:
    {
      return "SIGABRT";
    }
    case SIGFPE:
    {
      return "SIGFPE";
    }
    case SIGILL:
    {
      return "SIGILL";
    }
    case SIGSEGV:
    {
      return "SIGSEGV";
    }

    #ifdef WIN32

    case SIGBREAK:
    {
      return "SIGBREAK";
    }

    #endif

    case SIGKILL:
    {
      return "SIGKILL";
    }

    //
    // These overlap with SIGBREAK and
    // SIGABRT on Windows.
    //

    #ifndef WIN32

    case SIGTTIN:
    {
      return "SIGTTIN";
    }

    case SIGTTOU:
    {
      return "SIGTTOU";
    }

    #endif

    default:
    {
      return "None";
    }
  }
}

const Char *SignalGetHandlerName(SignalFunction function)
{
  if (function == SIG_IGN)
  {
    return "SIG_IGN";
  }
  else if (function == SIG_DFL)
  {
    return "SIG_DFL";
  }
  else
  {
    return "User defined";
  }
}

