/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Signal_H
#define Signal_H

#include <sys/types.h>
#include <signal.h>

#ifndef WIN32

#include <sys/time.h>

#ifdef __APPLE__

#ifndef SIGPOLL
#define SIGPOLL SIGIO
#endif

#endif // #ifdef __APPLE__ ...

#else // #ifndef WIN32 ...

#define SIGHUP        1
#define SIGQUIT       3
#define SIGKILL       9
#define SIGPIPE      13
#define SIGALRM      14
#define SIGUSR1      10
#define SIGUSR2      12
#define SIGCHLD      17
#define SIGCONT      18
#define SIGSTOP      19
#define SIGTSTP      20

#define SIGBUS        7
#define SIGPROF      27
#define SIGTRAP       5
#define SIGURG       23
#define SIGVTALRM    26
#define SIGXCPU      24
#define SIGXFSZ      25
#define SIGIOT        6
#define SIGSTKFLT    16
#define SIGPOLL      29
#define SIGPWR       30
#define SIGWINCH     28
#define SIGUNUSED    31

#ifdef _MSC_VER

#ifndef _SIGSET_T_
#define _SIGSET_T_
typedef int sigset_t;
#endif

#endif // #ifdef _MSC_VER ...

#endif // #ifndef WIN32 ... #else ...

#include "Types.h"
#include "Process.h"

#ifndef WIN32

#define SignalMask      sigset_t

#define SignalAction    struct sigaction
#define SignalTimer     struct itimerval

#define SignalEmptyMask(mask)  sigemptyset((mask))
#define SignalFillMask(mask)   sigfillset((mask))

#define SignalAddMask(mask, signal)     sigaddset((mask), (signal))
#define SignalRemoveMask(mask, signal)  sigdelset((mask), (signal))
#define SignalCheckMask(mask, signal)   sigismember((mask), (signal))

#define SignalSend(pid, signal)  ::kill((pid), (signal))
#define SignalRaise(signal)      ::raise((signal))

#else // #ifndef WIN32 ...

#define SignalMask      void

#define SignalAction    void
#define SignalTimer     void

Int SignalEmptyMask(SignalMask *mask);
Int SignalFillMask(SignalMask *mask);

Int SignalAddMask(SignalMask *mask, Int signal);
Int SignalRemoveMask(SignalMask *mask, Int signal);
Int SignalCheckMask(const SignalMask *mask, Int signal);

Int SignalSend(ProcessId pid, Int signal);
Int SignalRaise(Int signal);

#endif // #ifndef WIN32 ... #else ...

//
// Posix compliant signal manipulation
// functions.
//

Int SignalSetMask(Int how, const SignalMask *mask, SignalMask *oldmask);

void SignalLogMask(const SignalMask *mask, const Char *label = NULL, Logger *logger = NULL);

void SignalEmptyAction(SignalAction *action);

Int SignalSetAction(Int signal, const SignalAction *action, SignalAction *oldaction);

void SignalEmptyTimer(SignalTimer *timer);

Int SignalSetTimer(Int which, const SignalTimer *value, SignalTimer *oldvalue);

//
// Legacy signal support used on Windows.
//

typedef void (*SignalFunction)(Int value);

SignalFunction SignalSet(Int value, SignalFunction function);

//
// Return the conventional name of the
// signal.
//

const Char *SignalGetName(Int signal);

//
// Return the conventional name of the
// handler or null, if the handler is
// a custom function.
//

const Char *SignalGetHandlerName(SignalFunction function);

//
// Print the name of the system signal.
//

#define Sig(value) SignalGetName((value))

#endif /* Signal_H */
