/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include <sys/types.h>

#ifndef WIN32
#include <sys/utsname.h>
#endif

#if defined(__CYGWIN32__) || defined(__APPLE__) || \
        defined(__FreeBSD__) || defined(__sun)
#include <netinet/in_systm.h>
#endif

#ifdef __sun
#include <unistd.h>
#include <sys/termios.h>
#endif

#if defined(__APPLE__) || defined(__sun)
#include <netinet/in.h>
#endif

#ifndef WIN32
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <fcntl.h>

//
// System specific defines.
//

#if defined(__APPLE__) || defined(__FreeBSD__) || \
    defined(__sun)
#define SOL_IP  IPPROTO_IP
#endif

//
// The TIOCOUTQ ioctl is not implemented on
// Cygwin. Note also that on Windows, Mac OSX
// and Cygwin we force minimum kernel support
// and thus the TIOCOUTQ and IPTOS_LOWDELAY
// calls are disabled.
//

#ifdef __CYGWIN32__
#define TIOCOUTQ  ((unsigned int) -1)
#endif

//
// Redefine ugly bionic macro with glibc
// version to fix error:
// Socket.cpp:3474:11: error: dereferencing type-punned pointer will break strict-aliasing rules [-Werror=strict-aliasing]
//

#ifdef __BIONIC__
#undef IN6_IS_ADDR_V4MAPPED
# define IN6_IS_ADDR_V4MAPPED(a) \
  (__extension__							      \
   ({ __const struct in6_addr *__a = (__const struct in6_addr *) (a);	      \
      __a->s6_addr32[0] == 0						      \
      && __a->s6_addr32[1] == 0						      \
      && __a->s6_addr32[2] == htonl (0xffff); }))

#endif

//
// NX includes.
//

#include "Socket.h"
#include "Error.h"
#include "Memory.h"
#include "String.h"
#include "Random.h"
#include "Thread.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// Here the result can be a function call
// so be sure the parameter is evaluated
// only once.
//

#define SOCKET_RESULT(operation, error) \
{\
  Int value = (operation); \
\
  if (value == WSAEWOULDBLOCK) \
  { \
    errno = EAGAIN; \
\
    return -1; \
  } \
  else if (value == (Int) (error)) \
  { \
    SocketTranslateError(); \
\
    return -1; \
  } \
\
  return value; \
}

#ifdef WIN32

static SYSTEM Int SocketStartupFlag = 0;

/*
FIXME: Statically allocated string good for returning
       the error message built from FormatMessage()
       in SocketGetErrorString(). Double the size for
       safety since the returned string is in WCHARS
       format.
*/
/*
FIXME: This must be protected by allocating the data
       in Thread Local Storage.
*/
#define SocketErrorStringSize 1024

static Char SocketErrorString[2 * SocketErrorStringSize] = { 0 };

#endif // #ifdef WIN32 ...

//
// Static buffer to convert an IP from
// integer in network byte order to a
// string.
//
/*
FIXME: This must be protected by allocating the data
       in Thread Local Storage.
*/
static SYSTEM Char SocketIPString[16];

//
// Set this only once by querying the OS
// version.
//

static SYSTEM Int SocketKernelStep = -1;

static void SocketSetKernelStep();

Int SocketStartup()
{
  //
  // Nothing to do on Unix-like environments.
  //

  #ifdef WIN32

  WSADATA wsaData;

  if (SocketStartupFlag == 0)
  {
    Int result = WSAStartup(MAKEWORD(2, 2), &wsaData);

    if (result != 0)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Network startup failed. "
            << "Error is " << GetError() << " "
            << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Network startup failed. Error is "
                 << GetError() << " " << Stk(GetErrorString())
                 << ".\n";

      return -1;
    }

    SocketStartupFlag = 1;

    //
    // Windows can leave errno set to
    // ERANGE even on success.
    //

    SetError(0);
  }

  #endif // #ifdef WIN32 ...

  if (SocketKernelStep == -1)
  {
    SocketSetKernelStep();
  }

  return 1;
}
  
Int SocketCleanup()
{
  //
  // Nothing to do on Unix-like environments.
  //

  #ifdef WIN32

  if (SocketStartupFlag == 1)
  {
    Int result = WSACleanup();

    if (result != 0)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Network cleanup failed. "
            << "Error is " << GetError() << " "
            << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Network cleanup failed. Error is "
                 << GetError() << " " << Stk(GetErrorString())
                 << ".\n";

      return -1;
    }

    SocketStartupFlag = 0;
  }

  #endif // #ifdef WIN32 ...

  return 1;
}

#ifdef WIN32

Int SocketCreate(Int domain, Int type, Int protocol)
{
  //
  // The standard socket() call makes the
  // descriptor not inheritable by a child
  // process.
  //

  SOCKET_RESULT(WSASocket(domain, type, protocol, NULL, 0, 0), INVALID_SOCKET);
}

Int SocketBind(Int sockfd, const SocketAddress *addr, SocketAddressLength addrlen)
{
  SOCKET_RESULT(bind(sockfd, addr, addrlen), SOCKET_ERROR);
}

Int SocketListen(Int sockfd, Int backlog)
{
  SOCKET_RESULT(listen(sockfd, backlog), SOCKET_ERROR);
}

Int SocketConnect(Int sockfd, const SocketAddress *addr, SocketAddressLength addrlen)
{
  SOCKET_RESULT(connect(sockfd, addr, addrlen), SOCKET_ERROR);
}

Int SocketAccept(Int sockfd, SocketAddress *addr, SocketAddressLength *addrlen)
{
  SOCKET_RESULT(accept(sockfd, addr, addrlen), SOCKET_ERROR);
}

Int SocketRead(Int fd, void *buffer, Int size)
{
  #ifdef DEBUG
  Log() << "Socket: Reading " << size << " bytes "
        << "from FD#" << fd << ".\n";
  #endif

  WSABUF vector;

  vector.buf = (Char *) buffer;
  vector.len = size;

  DWORD flags = 0;
  DWORD total;

  Int result = WSARecv(fd, &vector, 1, &total, &flags, NULL, NULL);

  if (result == 0)
  {
    #ifdef DEBUG
    Log() << "Socket: Read result is " << result
          << " with " << total << " bytes read.\n";
    #endif

    return total;
  }

  #ifdef DEBUG

  Int error = SocketGetError();

  const Char *string = SocketGetErrorString(error);

  if (error != WSAEWOULDBLOCK)
  {
    Log() << "Socket: ERROR! Read error is " << error
          << " " << Stk(string) << ".\n";
  }
  else
  {
    Log() << "Socket: Reading from FD#" << fd
          << " would block.\n";
  }

  #endif

  SOCKET_RESULT(result, SOCKET_ERROR);
}

Int SocketWrite(Int fd, const void *buffer, Int size)
{
  #ifdef DEBUG
  Log() << "Socket: Writing " << size << " bytes "
        << "to FD#" << fd << ".\n";
  #endif

  WSABUF vector;

  vector.buf = (Char *) buffer;
  vector.len = size;

  DWORD total;

  Int result = WSASend(fd, &vector, 1, &total, 0, NULL, NULL);

  #ifdef DEBUG
  Log() << "Socket: Write result is " << result
        << " with " << total << " bytes written.\n";
  #endif

  if (result == 0)
  {
    return total;
  }

  #ifdef DEBUG

  Int error = SocketGetError();

  const Char *string = SocketGetErrorString(error);

  if (error != WSAEWOULDBLOCK)
  {
    Log() << "Socket: ERROR! Write error is " << error
          << " " << Stk(string) << ".\n";
  }

  #endif

  SOCKET_RESULT(result, SOCKET_ERROR);
}

Int SocketShutdown(Int fd, Int how)
{
  SOCKET_RESULT(shutdown(fd, how), SOCKET_ERROR);
}

Int SocketClose(Int fd)
{
  //
  // If the socket was set to non-blocking mode
  // and we set a linger timeout, the close can
  // fail with 'Resource temporarily unavailable'.
  // If this is the case, wait until the close
  // succeeds or fail with some other error.
  //
  // SOCKET_RESULT(closesocket(fd), SOCKET_ERROR);
  //

  Int result;

  while ((result = closesocket(fd)) == SOCKET_ERROR)
  {
    if (SocketGetError() == WSAEWOULDBLOCK)
    {
      #ifdef WARNING
      Log() << "Socket: WARNING! Retrying close on "
            << "socket FD#" << fd << ".\n";
      #endif

      ThreadSleep(50);

      continue;
    }

    SocketTranslateError();

    return -1;
  }

  return 0;
}

Int SocketReadVector(Int fd, const SocketVector *vector, Int count)
{
  #ifdef DEBUG
  Log() << "Socket: Reading " << count << " buffers "
        << "from FD#" << fd << ".\n";
  #endif

  //
  // Must convert the I/O vector to an
  // array of WSABUF structures.
  //

  Int result;

  //
  // Apparently the structures can't be used
  // interchangeably, otherwise we could save
  // the initialization. Note also that the
  // vector pointer is const, while the call
  // requires a non-const WSABUF pointer.
  //

  WSABUF buffers[count];

  for (result = 0; result < count; result++, vector++)
  {
    buffers[result].len = vector -> iov_len;
    buffers[result].buf = (Char *) vector -> iov_base;
  }

  DWORD total;

  DWORD flags = 0;

  result = WSARecv(fd, buffers, count, &total, &flags, NULL, NULL);

  if (result == 0)
  {
    #ifdef DEBUG
    Log() << "Socket: Read result is " << result
          << " with " << total << " bytes read.\n";
    #endif

    return total;
  }

  #ifdef DEBUG

  Int error = SocketGetError();

  const Char *string = SocketGetErrorString(error);

  if (error != WSAEWOULDBLOCK)
  {
    Log() << "Socket: ERROR! Read error is " << error
          << " " << Stk(string) << ".\n";
  }
  else
  {
    Log() << "Socket: Reading from FD#" << fd
          << " would block.\n";
  }

  #endif

  SOCKET_RESULT(result, SOCKET_ERROR);
}

Int SocketWriteVector(Int fd, const SocketVector *vector, Int count)
{
  #ifdef DEBUG
  Log() << "Socket: Writing " << count << " buffers "
        << "to FD#" << fd << ".\n";
  #endif

  //
  // Convert the I/O vector to an array
  // of WSABUF structures.
  //

  Int result;

  WSABUF buffers[count];

  for (result = 0; result < count; result++, vector++)
  {
    buffers[result].len = vector -> iov_len;
    buffers[result].buf = (Char *) vector -> iov_base;
  }

  DWORD total;

  result = WSASend(fd, buffers, count, &total, 0, NULL, NULL);

  #ifdef DEBUG
  Log() << "Socket: Write result is " << result
        << " with " << total << " bytes written.\n";
  #endif

  if (result == 0)
  {
    return total;
  }

  #ifdef DEBUG

  Int error = SocketGetError();

  const Char *string = SocketGetErrorString(error);

  if (error != WSAEWOULDBLOCK)
  {
    Log() << "Socket: ERROR! Write error is " << error
          << " " << Stk(string) << ".\n";
  }

  #endif

  SOCKET_RESULT(result, SOCKET_ERROR);
}

Int SocketReadDatagram(Int fd, void *buffer, Int size, SocketAddress *address,
                           SocketAddressLength *addresslen)
{
  #ifdef DEBUG
  Log() << "Socket: Reading " << size << " datagram "
        << "bytes from FD#" << fd << ".\n";
  #endif

  WSABUF vector;

  vector.buf = (Char *) buffer;
  vector.len = size;

  DWORD flags = 0;
  DWORD total;

  Int result = WSARecvFrom(fd, &vector, 1, &total, &flags,
                               address, addresslen, NULL, NULL);

  if (result == 0)
  {
    #ifdef DEBUG
    Log() << "Socket: Read result is " << result
          << " with " << total << " bytes read.\n";
    #endif

    return total;
  }

  #ifdef DEBUG

  Int error = SocketGetError();

  const Char *string = SocketGetErrorString(error);

  if (error != WSAEWOULDBLOCK)
  {
    Log() << "Socket: ERROR! Read error is " << error
          << " " << Stk(string) << ".\n";
  }
  else
  {
    Log() << "Socket: Reading from FD#" << fd
          << " would block.\n";
  }

  #endif

  SOCKET_RESULT(result, SOCKET_ERROR);
}

Int SocketWriteDatagram(Int fd, const void *buffer, Int size,
                            const SocketAddress *address, SocketAddressLength addresslen)
{
  #ifdef DEBUG
  Log() << "Socket: Writing " << size << " datagram "
        << "bytes to FD#" << fd << ".\n";
  #endif

  WSABUF vector;

  vector.buf = (Char *) buffer;
  vector.len = size;

  DWORD total;

  Int result = WSASendTo(fd, &vector, 1, &total, 0, address,
                             addresslen, NULL, NULL);

  #ifdef DEBUG
  Log() << "Socket: Write result is " << result
        << " with " << total << " bytes written.\n";
  #endif

  if (result == 0)
  {
    return total;
  }

  #ifdef DEBUG

  Int error = SocketGetError();

  const Char *string = SocketGetErrorString(error);

  if (error != WSAEWOULDBLOCK)
  {
    Log() << "Socket: ERROR! Write error is " << error
          << " " << Stk(string) << ".\n";
  }

  #endif

  SOCKET_RESULT(result, SOCKET_ERROR);
}

//
// Make use of the standard DuplicateHandle()
// instead of WSADuplicateSocket() followed by
// WSASocket(). In this way the function will
// work even if the handle is not a socket.
// The produced handle may fail to work with
// alternate Winsock stacks, but this can be
// considered a corner case.
//

Int SocketDuplicate(Int fd)
{
  // 
  // The value passed as 6th parameter
  // makes the new socket inheritable
  // by default.
  // 

  Int duplicate;

  if (DuplicateHandle(GetCurrentProcess(), HandleCast(fd),
                          GetCurrentProcess(), (HANDLE *) &duplicate, 0,
                              TRUE, DUPLICATE_SAME_ACCESS) == 0)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to duplicate FD#"
          << fd << ". Error is " << GetError() << " "
          << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to duplicate FD#" << fd
               << ". Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    return -1;
  }

  return duplicate;
}

Int SocketClone(Int fd1, Int fd2)
{
  //
  // Windows doesn't have support for assigning
  // a specific number to a duplicated handle.
  // Still we should be able to use the native
  // _dup2() if the descriptor to be duplicated
  // is one of the standard descriptors.
  //

  if (HandleCast(fd1) == GetStdHandle(STD_ERROR_HANDLE) ||
          HandleCast(fd1) == GetStdHandle(STD_OUTPUT_HANDLE) ||
              HandleCast(fd1) == GetStdHandle(STD_INPUT_HANDLE))
  {
    if (_dup2(fd1, fd2) == -1)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Failed to duplicate the "
            << "standard FD#" << fd1 << ". Error is "
            << GetError() << " " << Stk(GetErrorString())
            << ".\n";
      #endif

      LogError() << "Failed to duplicate the standard FD#"
                 << fd1 << ". Error is " << GetError()
                 << " " << Stk(GetErrorString())
                 << ".\n";

      return -1;
    }

    return 0;
  }
  else if (HandleCast(fd2) == GetStdHandle(STD_ERROR_HANDLE) ||
               HandleCast(fd2) == GetStdHandle(STD_OUTPUT_HANDLE) ||
                   HandleCast(fd2) == GetStdHandle(STD_INPUT_HANDLE))
  {
    Int fd;

    if ((fd = SocketDuplicate(fd1)) == -1)
    {
      //
      // The error will be reported by the
      // above function.
      //

      return -1;
    }

    if (HandleCast(fd2) == GetStdHandle(STD_ERROR_HANDLE))
    {
      if (SetStdHandle(STD_ERROR_HANDLE, HandleCast(fd)) == 0)
      {
        #ifdef ERROR
        Log() << "Socket: ERROR! Failed to assign the "
              << "standard error to FD#" << fd << ". Error is "
              << GetError() << " " << Stk(GetErrorString())
              << ".\n";
        #endif

        LogError() << "Failed to assign the standard error "
                   << "to FD#" << fd << ". Error is " << GetError()
                   << " " << Stk(GetErrorString()) << ".\n";

        return -1;
      }
    }

    if (HandleCast(fd2) == GetStdHandle(STD_OUTPUT_HANDLE))
    {
      if (SetStdHandle(STD_OUTPUT_HANDLE, HandleCast(fd)) == 0)
      {
        #ifdef ERROR
        Log() << "Socket: ERROR! Failed to assign the "
              << "standard output to FD#" << fd << ". Error is "
              << GetError() << " " << Stk(GetErrorString())
              << ".\n";
        #endif

        LogError() << "Failed to assign the standard output "
                   << "to FD#" << fd << ". Error is " << GetError()
                   << " " << Stk(GetErrorString()) << ".\n";

        return -1;
      }
    }

    if (HandleCast(fd2) == GetStdHandle(STD_INPUT_HANDLE))
    {
      if (SetStdHandle(STD_INPUT_HANDLE, HandleCast(fd)) == 0)
      {
        #ifdef ERROR
        Log() << "Socket: ERROR! Failed to assign the "
              << "standard input to FD#" << fd << ". Error is "
              << GetError() << " " << Stk(GetErrorString())
              << ".\n";
        #endif

        LogError() << "Failed to assign the standard input "
                   << "to FD#" << fd << ". Error is " << GetError()
                   << " " << Stk(GetErrorString()) << ".\n";

        return -1;
      }
    }

    return 1;
  }
  else
  {
    SetError(EBADF);

    #ifdef ERROR
    Log() << "Socket: ERROR! Can't duplicate the "
          << "non-standard FD#" << fd1 << ". Error is "
          << GetError() << " " << Stk(GetErrorString())
          << ".\n";
    #endif

    LogError() << "Can't duplicate the non-standard "
               << "FD#" << fd1 << ". Error is " << GetError()
               << " " << Stk(GetErrorString()) << ".\n";

    return -1;
  }
}

//
// Error traslation utilities.
//

typedef struct
{
  Int        error;
  const Char *string;
  Int        translation;
}
SocketWSAError;

//
// Missing in the MinGW headers.
//

#ifndef WSA_QOS_EUNKOWNPSOBJ
#define WSA_QOS_EUNKOWNPSOBJ  11024
#endif

SocketWSAError SocketWSAErrorList[] =
{
/*
FIXME: ERROR_FILE_NOT_FOUND and ERROR_PATH_NOT_FOUND
       were added to the list to handle the process
       creation errors. This may possibly masquerade
       some other error.
*/
  { ERROR_FILE_NOT_FOUND, "The system cannot find the file specified", ENOENT },
  { ERROR_PATH_NOT_FOUND, "The system cannot find the path specified", ENOENT },
  { WSA_INVALID_HANDLE, "Specified event object handle is invalid", -1 },
  { WSA_NOT_ENOUGH_MEMORY, "Insufficient memory available", -1 },
  { WSA_INVALID_PARAMETER, "One or more parameters are invalid", -1 },
  { WSA_OPERATION_ABORTED, "Overlapped operation aborted", -1 },
  { WSA_IO_INCOMPLETE, "Overlapped I/O event object not in signaled state", -1 },
  { WSA_IO_PENDING, "Overlapped operations will complete later", -1 },
  { WSAEINTR, "Interrupted function call", EINTR },
  { WSAEBADF, "File handle is not valid", EBADF },
  { WSAEACCES, "Permission denied", EACCES },
  { WSAEFAULT, "Bad address", EFAULT },
  { WSAEINVAL, "Invalid argument", EINVAL },
  { WSAEMFILE, "Too many open files", EMFILE },
  { WSAEWOULDBLOCK, "Resource temporarily unavailable", EAGAIN },
  { WSAEINPROGRESS, "Operation now in progress", EINPROGRESS },
  { WSAEALREADY, "Operation already in progress", -1 },
  { WSAENOTSOCK, "Socket operation on nonsocket", ENOTSOCK },
  { WSAEDESTADDRREQ, "Destination address required", -1 },
  { WSAEMSGSIZE, "Message too long", -1 },
  { WSAEPROTOTYPE, "Protocol wrong type for socket", -1 },
  { WSAENOPROTOOPT, "Bad protocol option", -1 },
  { WSAEPROTONOSUPPORT, "Protocol not supported", EPROTO },
  { WSAESOCKTNOSUPPORT, "Socket type not supported", -1 },
  { WSAEOPNOTSUPP, "Operation not supported", ENOTSUP },
  { WSAEPFNOSUPPORT, "Protocol family not supported", -1 },
  { WSAEAFNOSUPPORT, "Address family not supported by protocol family", -1 },
  { WSAEADDRINUSE, "Address already in use", -1 },
  { WSAEADDRNOTAVAIL, "Cannot assign requested address", EADDRNOTAVAIL },
  { WSAENETDOWN, "Network is down", -1 },
  { WSAENETUNREACH, "Network is unreachable", -1 },
  { WSAENETRESET, "Network dropped connection on reset", -1 },
  { WSAECONNABORTED, "Software caused connection abort", ECONNABORTED },
  { WSAECONNRESET, "Connection reset by peer", ECONNRESET },
  { WSAENOBUFS, "No buffer space available", -1 },
  { WSAEISCONN, "Socket is already connected", EISCONN },
  { WSAENOTCONN, "Socket is not connected", -1 },
  { WSAESHUTDOWN, "Cannot send after socket shutdown", -1 },
  { WSAETOOMANYREFS, "Too many references", -1 },
  { WSAETIMEDOUT, "Connection timed out", ETIMEDOUT },
  { WSAECONNREFUSED, "Connection refused", ECONNREFUSED },
  { WSAELOOP, "Cannot translate name", -1 },
  { WSAENAMETOOLONG, "Name too long", ENAMETOOLONG },
  { WSAEHOSTDOWN, "Host is down", -1 },
  { WSAEHOSTUNREACH, "No route to host", -1 },
  { WSAENOTEMPTY, "Directory not empty", ENOTEMPTY },
  { WSAEPROCLIM, "Too many processes", -1 },
  { WSAEUSERS, "User quota exceeded", -1 },
  { WSAEDQUOT, "Disk quota exceeded", -1 },
  { WSAESTALE, "Stale file handle reference", -1 },
  { WSAEREMOTE, "Item is remote", -1 },
  { WSASYSNOTREADY, "Network subsystem is unavailable", -1 },
  { WSAVERNOTSUPPORTED, "Winsock.dll version out of range", -1 },
  { WSANOTINITIALISED, "Successful WSAStartup not yet performed", -1 },
  { WSAEDISCON, "Graceful shutdown in progress", -1 },
  { WSAENOMORE, "No more results", -1 },
  { WSAECANCELLED, "Call has been cancelled", -1 },
  { WSAEINVALIDPROCTABLE, "Procedure call table is invalid", -1 },
  { WSAEINVALIDPROVIDER, "Service provider is invalid", -1 },
  { WSAEPROVIDERFAILEDINIT, "Service provider failed to initialize", -1 },
  { WSASYSCALLFAILURE, "System call failure", -1 },
  { WSASERVICE_NOT_FOUND, "Service not found", -1 },
  { WSATYPE_NOT_FOUND, "Class type not found", -1 },
  { WSA_E_NO_MORE, "No more results", -1 },
  { WSA_E_CANCELLED, "Call was cancelled", -1 },
  { WSAEREFUSED, "Database query was refused", -1 },
  { WSAHOST_NOT_FOUND, "Host not found", -1 },
  { WSATRY_AGAIN, "Nonauthoritative host not found", -1 },
  { WSANO_RECOVERY, "This is a nonrecoverable error", -1 },
  { WSANO_DATA, "Valid name, no data record of requested type", EADDRNOTAVAIL },
  { WSA_QOS_RECEIVERS, "QOS receivers", -1 },
  { WSA_QOS_SENDERS, "QOS senders", -1 },
  { WSA_QOS_NO_SENDERS, "No QOS senders", -1 },
  { WSA_QOS_NO_RECEIVERS, "QOS no receivers", -1 },
  { WSA_QOS_REQUEST_CONFIRMED, "QOS request confirmed", -1 },
  { WSA_QOS_ADMISSION_FAILURE, "QOS admission error", -1 },
  { WSA_QOS_POLICY_FAILURE, "QOS policy failure", -1 },
  { WSA_QOS_BAD_STYLE, "QOS bad style", -1 },
  { WSA_QOS_BAD_OBJECT, "QOS bad object", -1 },
  { WSA_QOS_TRAFFIC_CTRL_ERROR, "QOS traffic control error", -1 },
  { WSA_QOS_GENERIC_ERROR, "QOS generic error", -1 },
  { WSA_QOS_ESERVICETYPE, "QOS service type error", -1 },
  { WSA_QOS_EFLOWSPEC, "QOS flowspec error", -1 },
  { WSA_QOS_EPROVSPECBUF, "Invalid QOS provider buffer", -1 },
  { WSA_QOS_EFILTERSTYLE, "Invalid QOS filter style", -1 },
  { WSA_QOS_EFILTERTYPE, "Invalid QOS filter type", -1 },
  { WSA_QOS_EFILTERCOUNT, "Incorrect QOS filter count", -1 },
  { WSA_QOS_EOBJLENGTH, "Invalid QOS object length", -1 },
  { WSA_QOS_EFLOWCOUNT, "Incorrect QOS flow count", -1 },
  { WSA_QOS_EUNKOWNPSOBJ, "Unrecognized QOS object", -1 },
  { WSA_QOS_EPOLICYOBJ, "Invalid QOS policy object", -1 },
  { WSA_QOS_EFLOWDESC, "Invalid QOS flow descriptor", -1 },
  { WSA_QOS_EPSFLOWSPEC, "Invalid QOS provider-specific flowspec", -1 },
  { WSA_QOS_EPSFILTERSPEC, "Invalid QOS provider-specific filterspec", -1 },
  { WSA_QOS_ESDMODEOBJ, "Invalid QOS shape discard mode object", -1 },
  { WSA_QOS_ESHAPERATEOBJ, "Invalid QOS shaping rate object", -1 },
  { WSA_QOS_RESERVED_PETYPE, "Reserved policy QOS element type", -1 }
};

Int SocketWSAErrorListSize = sizeof(SocketWSAErrorList) / sizeof(SocketWSAError);

const Char *SocketGetErrorString(Int error)
{
  if (error == 0)
  {
    return "No error";
  }

  for (Int i = 0; i < SocketWSAErrorListSize; i++)
  {
    if (error == SocketWSAErrorList[i].error)
    {
      return SocketWSAErrorList[i].string;
    }
  }

  //
  // Return a formatted message in the
  // system locale language.
  //

  if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL, error, MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
                            (LPTSTR) SocketErrorString, SocketErrorStringSize,
                                NULL ) != 0)
  {
    return SocketErrorString;
  }
  else
  {
    WSASetLastError(error);
  }

  return "Unknown error";
}

const Char *SocketGetErrorName(Int error)
{
  if (error == 0)
  {
    return "Success";
  }

  return "Unknown";
}

Int SocketTranslateError(Int error)
{
  for (Int i = 0; i < SocketWSAErrorListSize; i++)
  {
    if (error == SocketWSAErrorList[i].error)
    {
      if (SocketWSAErrorList[i].translation != -1)
      {
        return SocketWSAErrorList[i].translation;
      }
      else
      {
        break;
      }
    }
  }

  //
  // No correspondence found for this
  // error. Return the WSA error.
  //

  return error;
}

void SocketTranslateError()
{
  Int error = WSAGetLastError();

  for (Int i = 0; i < SocketWSAErrorListSize; i++)
  {
    if (error == SocketWSAErrorList[i].error)
    {
      //
      // Check if a standard correspondence
      // exists for this error.
      //

      if (SocketWSAErrorList[i].translation != -1)
      {
        errno = SocketWSAErrorList[i].translation;

        return;
      }
      else
      {
        break;
      }
    }
  }

  //
  // No correspondence found. Force the
  // GetError() function to return the
  // WSA error.
  //

  errno = 0;
}

#endif // #ifdef WIN32 ...

static void SocketSetKernelStep()
{
  #if defined(__CYGWIN32__) || defined(WIN32)

  #ifdef TEST
  Log() << "Socket: Assuming level 0 system support.\n";
  #endif

  SocketKernelStep = 0;

  #elif defined(__APPLE__)

  #ifdef TEST
  Log() << "Socket: Assuming level 2 system support.\n";
  #endif

  SocketKernelStep = 2;

  #else

  struct utsname buffer;

  if (uname(&buffer) == -1)
  {
    #ifdef WARNING
    Log() << "Socket: WARNING! Failed to get system info. Error is "
          << GetError() << " " << Stk(GetErrorString()) << ".\n";

    Log() << "Socket: WARNING! Assuming lowest system support.\n";
    #endif

    LogWarning() << "Failed to get system info. Error is "
                 << GetError() << " " << Stk(GetErrorString())
                 << ".\n";

    LogWarning() << "Assuming lowest system support.\n";

    SocketKernelStep = 0;
  }
  else
  {
    #ifdef TEST
    Log() << "Socket: System is " << Stk(buffer.sysname)
          << " nodename " << Stk(buffer.nodename) << " release "
          << Stk(buffer.release) << ".\n";

    Log() << "Socket: Version is " << Stk(buffer.version)
          << " machine " << Stk(buffer.machine) << ".\n";
    #endif

    //
    // Should test support on other operating systems.
    //

    if (StringCompare(buffer.sysname, "Linux") == 0)
    {
      if (StringSizeCompare(buffer.release, "2.0.", 4) == 0 ||
              StringSizeCompare(buffer.release, "2.2.", 4) == 0)
      {
        #ifdef TEST
        Log() << "Socket: Assuming level 2 system support.\n";
        #endif

        SocketKernelStep = 2;
      }
      else if (StringSizeCompare(buffer.release, "2.4.", 4) == 0)
      {
        #ifdef TEST
        Log() << "Socket: Assuming level 3 system support.\n";
        #endif

        SocketKernelStep = 3;
      }
      else if (StringSizeCompare(buffer.release, "2.6.", 4) == 0 ||
                   StringSizeCompare(buffer.release, "2.8.", 4) == 0)
      {
        #ifdef TEST
        Log() << "Socket: Assuming level 4 system support.\n";
        #endif

        SocketKernelStep = 4;
      }
      else
      {
        #ifdef TEST
        Log() << "Socket: Assuming level 5 system support.\n";
        #endif

        SocketKernelStep = 5;
      }
    }
    else if (StringCompare(buffer.sysname, "SunOS") == 0)
    {
      #ifdef TEST
      Log() << "Socket: Assuming level 1 system support.\n";
      #endif

      SocketKernelStep = 1;
    }
    else
    {
      #ifdef TEST
      Log() << "Socket: Assuming level 0 system support.\n";
      #endif

      SocketKernelStep = 0;
    }
  }

  #endif // #if defined(__CYGWIN32__) || defined(__APPLE__) || defined(WIN32)
}

Int SocketQuery(Int fd, Int option)
{
  Int value;

  SocketLength length = sizeof(value);

  Int result;

  switch (option)
  {
    case SocketOptionNoDelay:
    {
      result = getsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
                              (Char *) &value, &length);

      break;
    }
    case SocketOptionNoBlocking:
    {
      //
      // On Windows there is no way to query
      // the OS whether or not the socket is
      // in non-blocking mode. A state flag
      // must be handled at a higher level.
      //
      
      #ifndef WIN32

      result = fcntl(fd, F_GETFL);

      if (result != -1)
      {
        value = ((result & O_NONBLOCK) != 0);
      }

      break;

      #endif
    }
    default:
    {
      #ifdef WARNING
      Log() << "Socket: WARNING! Query for option "
            << option << " not implemented.\n";
      #endif

      SetError(ENOSYS);

      result = -1;

      break;
    }
  }

  if (result == 0)
  {
    #ifdef TEST
    Log() << "Socket: Query for option " << option
          << " result " << value << ".\n";
    #endif

    return value;
  }

  #ifdef WIN32

  SocketTranslateError();

  #endif

  #ifdef ERROR
  Log() << "Socket: ERROR! Query failed for option "
        << option << " on FD#" << fd << ".\n";
  #endif

  LogError() << "Query failed for option " << option
             << " on FD#" << fd << ".\n";

  #ifdef ERROR
  Log() << "Socket: ERROR! Error is " << GetError()
        << " " << Stk(GetErrorString()) << ".\n";
  #endif

  LogError() << "Error is " << GetError() << " "
             << Stk(GetErrorString()) << ".\n";

  return -1;
}

Int SocketNoBlocking(Int fd, Int value)
{
  #ifndef WIN32

  Int result = fcntl(fd, F_GETFL);

  if (result != -1)
  {
    if (value == ((result & O_NONBLOCK) != 0))
    {
      #ifdef TEST
      Log() << "Socket: Nothing to do for FD#" << fd
            << " with non-blocking mode " << value
            << ".\n";
      #endif

      return 1;
    }

    #ifdef TEST
    Log() << "Socket: Changing non-blocking mode for "
          << "FD#" << fd << " to " << value << ".\n";
    #endif

    if (value == 0)
    {
      result &= ~O_NONBLOCK;
    }
    else
    {
      result |= O_NONBLOCK;
    }

    result = fcntl(fd, F_SETFL, result);
  }

  #else

  #ifdef TEST
  Log() << "Socket: Changing non-blocking mode for "
        << "FD#" << fd << " to " << value << ".\n";
  #endif

  unsigned Long mode = value;

  Int result = ioctlsocket(fd, FIONBIO, &mode);

  #endif

  if (result == -1)
  {
    #ifdef WIN32

    SocketTranslateError();

    #endif

    #ifdef ERROR
    Log() << "Socket: ERROR! Can't set non-blocking mode "
          << "on FD#" << fd << " to " << value << ". Error is "
          << GetError() << " " << Stk(GetErrorString())
          << ".\n";
    #endif

    LogError() << "Can't set non-blocking mode on FD#"
               << fd << " to " << value << ". Error is "
               << GetError() << " " << Stk(GetErrorString())
               << ".\n";

    return -1;
  }
  #ifdef TEST
  else
  {
    Log() << "Socket: Set non-blocking mode on FD#"
          << fd << " to " << value << ".\n";
  }
  #endif

  return 1;
}

Int SocketNoInherit(Int fd, Int value)
{
  #ifdef TEST
  Log() << "Socket: Setting no-inherit to " << value
        << " on FD#" << fd << ".\n";
  #endif

  #ifndef WIN32

  if (fcntl(fd, F_SETFD, value) == 0)
  {
    return 1;
  }

  #else

  Int flag = (value != 0 ? 0 : HANDLE_FLAG_INHERIT);

  if (SetHandleInformation(HandleCast(fd), HANDLE_FLAG_INHERIT, flag) != 0)
  {
    return 1;
  }

  #endif

  #ifdef ERROR
  Log() << "Socket: ERROR! Can't set no-inherit mode "
        << "on FD#" << fd << " to " << value << ". Error is "
        << GetError() << " " << Stk(GetErrorString())
        << ".\n";
  #endif

  LogError() << "Can't set no-inherit mode on FD#" << fd
             << " to " << value << ". Error is " << GetError()
             << " " << Stk(GetErrorString()) << ".\n";

  return -1;
}

//
// Set the TCP_NODELAY option. Note that on old Linux
// kernels turning off the Nagle algorithm didn't work
// when the proxy was run through a PPP link. Trying
// to do so caused the kernel to stop delivering data
// if a serious network congestion was encountered.
//

Int SocketNoDelay(Int fd, Int value)
{
  Int result = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
                              (Char *) &value, sizeof(value));

  #ifdef TEST
  Log() << "Socket: Setting no-delay mode on FD#"
        << fd << " to " << value << " returned "
        << result << " with error " << GetError()
        << " " << Stk(GetErrorString()) << ".\n";
  #endif

  if (result == 0)
  {
    result = 1;
  }

  #ifndef WIN32

  else if (result == -1)
  {
    //
    // Is it become a different error on
    // Mac OSX 10.4?
    //

    #if defined(__APPLE__)

    result = 0;

    #endif

    #if defined(__sun)

    if (GetError() == ENOPROTOOPT)
    {
      result = 0;
    }

    #endif

    #if !defined(__APPLE__) && !defined(__sun)

    if (GetError() == EOPNOTSUPP || GetError() == ENOPROTOOPT)
    {
      result = 0;
    }

    #endif
  }

  #endif

  if (result == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Can't set no-delay mode "
          << "on FD#" << fd << " to " << value << ". Error is "
          << GetError() << " " << Stk(GetErrorString())
          << ".\n";
    #endif

    LogError() << "Can't set no-delay mode on FD#"
               << fd << " to " << value << ". Error is "
               << GetError() << " " << Stk(GetErrorString())
               << ".\n";

    return -1;
  }
  else if (result == 0)
  {
    #ifdef TEST
    Log() << "Socket: Option no-delay not supported "
          << "on FD#" << fd << ".\n";
    #endif

    return -1;
  }
  else
  {
    #ifdef TEST
    Log() << "Socket: Set no-delay mode on FD#"
          << fd << " to " << value << ".\n";
    #endif

    return 1;
  }
}

//
// Set the TCP_CORK option on Linux. Only recent
// kernels support using the option together with
// TCP_NODELAY. Other operating systems may have
// a similar option, but note that TCP_NOPUSH is
// not working reliably Mac OSX.
//

Int SocketNoFlush(Int fd, Int value)
{
  #ifdef __linux__

  if (SocketKernelStep > 2)
  {
    Int result = setsockopt(fd, IPPROTO_TCP, TCP_CORK,
                                (Char *) &value, sizeof(value));

    if (result != -1)
    {
      #ifdef TEST
      Log() << "Socket: Set TCP_CORK to " << value
            << " on FD#" << fd << ".\n";
      #endif

      return 1;
    }

    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to set TCP_CORK to "
          << value << " on FD#" << fd << ".\n";

    Log() << "Socket: ERROR! Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to set TCP_CORK to "
               << value << " on FD#" << fd << ".\n";

    LogError() << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    return -1;
  }

  #else // #ifdef __linux__ ...

  #ifdef TEST
  Log() << "Socket: Option TCP_CORK not supported "
        << "on FD#" << fd << ".\n";
  #endif

  #endif // #ifdef __linux__ ... #else ...

  return 0;
}

Int SocketReceiveBuffer(Int fd, Int size)
{
  if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF,
                     (Char *) &size, sizeof(size)) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to set SO_RCVBUF "
          << "size to " << size << " on FD#" << fd
          << ". Error is " << GetError() << " "
          << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to set SO_RCVBUF size to "
               << size << " on FD#" << fd << ". Error is "
               << GetError() << " " << Stk(GetErrorString())
               << ".\n";

    return -1;
  }
  #ifdef TEST
  else
  {
    Log() << "Socket: Set SO_RCVBUF on FD#" << fd
          << " to " << size << " bytes.\n";
  }
  #endif

  return 1;
}

Int SocketSendBuffer(Int fd, Int size)
{
  if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF,
                     (Char *) &size, sizeof(size)) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to set SO_SNDBUF "
          << "size to " << size << " on FD#" << fd
          << ". Error is " << GetError() << " "
          << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to set SO_SNDBUF size to "
               << size << " on FD#" << fd << ". Error is "
               << GetError() << " " << Stk(GetErrorString())
               << ".\n";

    return -1;
  }
  #ifdef TEST
  else
  {
    Log() << "Socket: Set SO_SNDBUF on FD#" << fd
          << " to " << size << " bytes.\n";
  }
  #endif

  return 1;
}

Int SocketLowDelay(Int fd, Int value)
{
  #ifndef WIN32

  if (SocketKernelStep > 0)
  {
    Int value = IPTOS_LOWDELAY;

    if (setsockopt(fd, SOL_IP, IP_TOS, &value, sizeof(value)) == -1)
    {
      if (GetError() == EOPNOTSUPP)
      {
        #ifdef TEST
        Log() << "Socket: Option IPTOS_LOWDELAY not supported "
              << "on FD#" << fd << ".\n";
        #endif

        return -1;
      }
      else
      {
        #ifdef WARNING
        Log() << "Socket: WARNING! Failed to set IPTOS_LOWDELAY "
              << "on FD#" << fd << " to " << value << ". Error is "
              << GetError() << " " << Stk(GetErrorString())
              << ".\n";
        #endif

        LogWarning() << "Failed to set IPTOS_LOWDELAY on FD#"
                     << fd << " to " << value << ". Error is "
                     << GetError() << " " << Stk(GetErrorString())
                     << ".\n";

        return -1;
      }
    }
    else
    {
      #ifdef TEST
      Log() << "Socket: Set IPTOS_LOWDELAY on FD#" << fd
            << ".\n";
      #endif

      return 1;
    }
  }

  #endif

  #ifdef TEST
  Log() << "Socket: Option IPTOS_LOWDELAY not implemented "
        << "on FD#" << fd << ".\n";
  #endif

  return 0;
}

Int SocketKeepAlive(Int fd, Int value)
{
  if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
                     (Char *) &value, sizeof(value)) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to set SO_KEEPALIVE "
          << "on FD#" << fd << " to " << value << ". Error is "
          << GetError() << " " << Stk(GetErrorString())
          << ".\n";
    #endif

    LogError() << "Failed to set SO_KEEPALIVE on FD#"
               << fd << " to " << value << ". Error is "
               << GetError() << " " << Stk(GetErrorString())
               << ".\n";

    return -1;
  }
  #ifdef TEST
  else
  {
    Log() << "Socket: Set SO_KEEPALIVE on FD#" << fd
          << ".\n";
  }
  #endif

  return 1;
}

Int SocketLingerTimeout(Int fd, Int timeout)
{
  struct linger value;

  if (timeout > 0)
  {
    value.l_onoff  = 1;
    value.l_linger = timeout;
  }
  else
  {
    value.l_onoff  = 0;
    value.l_linger = 0;
  }

  if (setsockopt(fd, SOL_SOCKET, SO_LINGER,
                     (Char *) &value, sizeof(value)) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to set SO_LINGER values to "
          << value.l_onoff << " and " << value.l_linger
          << " on FD#" << fd << ". Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to set SO_LINGER values to "
               << value.l_onoff << " and " << value.l_linger
               << " on FD#" << fd << ". Error is " << GetError()
               << " " << Stk(GetErrorString()) << ".\n";

    return -1;
  }
  #ifdef TEST
  else
  {
    Log() << "Socket: Set SO_LINGER values to "
          << value.l_onoff << " and " << value.l_linger
          << " on FD#" << fd << ".\n";
  }
  #endif

  return 1;
}

Int SocketReuseAddress(Int fd, Int value)
{
  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
                     (Char *) &value, sizeof(value)) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to set SO_REUSEADDR "
          << "on FD#" << fd << " to " << value << ". Error is "
          << GetError() << " " << Stk(GetErrorString())
          << ".\n";
    #endif

    LogError() << "Failed to set SO_REUSEADDR on FD#"
               << fd << " to " << value << ". Error is "
               << GetError() << " " << Stk(GetErrorString())
               << ".\n";

    return -1;
  }
  #ifdef TEST
  else
  {
    Log() << "Socket: Set SO_REUSEADDR on FD#"
          << fd << ".\n";
  }
  #endif

  return 1;
}

Int SocketBytesReadable(Int fd)
{
  unsigned Long readable = 0;

  #ifndef WIN32

  if (ioctl(fd, FIONREAD, &readable) == -1)

  #else

  if (ioctlsocket(fd, FIONREAD, &readable) == -1)

  #endif
  {
    #ifdef WIN32

    SocketTranslateError();

    #endif

    #ifdef TEST
    Log() << "Socket: WARNING! Failed to get bytes readable "
          << "from FD#" << fd << ". Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    return -1;
  }

  #ifdef TEST
  Log() << "Socket: Returning " << (Int) readable
        << " bytes readable on FD#" << fd << ".\n";
  #endif

  return (Int) readable;
}

Int SocketBytesQueued(Int fd)
{
  //
  // The TIOCOUTQ ioctl returns the space
  // available on the Linux kernels 2.0
  // and 2.2.
  //

  #ifndef WIN32

  if (SocketKernelStep > 2)
  {
    Long queued = 0;

    if (ioctl(fd, TIOCOUTQ, (void *) &queued) == -1)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Failed to get bytes queued "
            << "on FD#" << fd << ". Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Failed to get bytes queued "
                 << "on FD#" << fd << ". Error is " << GetError()
                 << " " << Stk(GetErrorString()) << ".\n";

      return -1;
    }
    else
    {
      #ifdef TEST
      Log() << "Socket: Returning " << (Int) queued
            << " bytes queued on FD#" << fd << ".\n";
      #endif

      return (Int) queued;
    }
  }

  #endif

  #ifdef TEST
  Log() << "Socket: Option TIOCOUTQ not implemented "
        << "on FD#" << fd << ",\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

Int SocketBytesWritable(Int fd)
{
  #ifndef WIN32

  if (SocketKernelStep > 2)
  {
    //
    // Get the bytes queued first.
    //

    Int queued;
    Int size;

    SocketLength len = sizeof(size);

    if ((queued = SocketBytesQueued(fd)) == -1)
    {
      return -1;
    }

    //
    // Subtract from the size of the
    // TCP buffer.
    //

    if (getsockopt(fd, SOL_SOCKET, SO_SNDBUF,
                       (Char *) &size, &len) == -1)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Failed to get the SO_SNDBUF "
            << "size on FD#" << fd << ". Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Failed to get the SO_SNDBUF "
                 << "size on FD#" << fd << ". Error is "
                 << GetError() << " " << Stk(GetErrorString())
                 << ".\n";

      return -1;
    }

    #ifdef TEST
    Log() << "Socket: Size of the send buffer "
          << "for FD#" << fd << " is " << size
          << ".\n";
    #endif

    //
    // Return 0 if there aren't at least
    // 16K available.
    //

    size -= (queued + 16384);

    if (size > 0)
    {
      #ifdef TEST
      Log() << "Socket: Returning " << size
            << " bytes writable on FD#" << fd
            << ".\n";
      #endif

      return size;
    }
    else
    {
      #ifdef TEST
      Log() << "Socket: Returning no bytes "
            << "writable on FD#" << fd << ".\n";
      #endif

      return 0;
    }
  }

  #endif

  #ifdef TEST
  Log() << "Socket: Option TIOCOUTQ not implemented "
        << "on FD#" << fd << ",\n";
  #endif

  SetError(ENOSYS);

  return -1;
}

//
// Set the TCP_QUICKACK option on Linux.
//

Int SocketQuickAck(Int fd, Int value)
{
  #ifdef __linux__

  if (SocketKernelStep > 2)
  {
    Int result = setsockopt(fd, IPPROTO_TCP, TCP_QUICKACK,
                                (Char *) &value, sizeof(value));

    if (result != -1)
    {
      #ifdef TEST
      Log() << "Socket: Set TCP_QUICKACK to " << value
            << " on FD#" << fd << ".\n";
      #endif

      return 1;
    }

    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to set TCP_QUICKACK to "
          << value << " on FD#" << fd << ".\n";

    Log() << "Socket: ERROR! Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to set TCP_QUICKACK to "
               << value << " on FD#" << fd << ".\n";

    LogError() << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    return -1;
  }

  #else // #ifdef __linux__ ...

  #ifdef TEST
  Log() << "Socket: Option TCP_QUICKACK not supported "
        << "on FD#" << fd << ".\n";
  #endif

  #endif // #ifdef __linux__ ... #else ...

  return 0;
}

//
// Replace the socketpair() on Windows by listening
// on a random TCP port and by creating a thread
// connecting to this listening port. Compared to
// using a named pipe, this has the advantage of
// being a real socket pair, so that it can be used
// with all the winsock functions, including select.
//

#ifdef WIN32

typedef struct
{
  Int listenFd_;
  Int connectFd_;
  Int acceptFd_;

  Int listenIp_;
  Int listenPort_;

  SocketAddressInet listenAddr_;
  SocketAddressInet acceptAddr_;

  SocketAddressLength acceptLen_;

  Char acceptData_[256];

} SocketPairData;


DWORD WINAPI SocketPairConnector(PVOID parameter)
{
  ThreadInit();

  SocketPairData *data = (SocketPairData *) parameter;

  Unsigned Int cookieData;
  Unsigned Int cookieIndex;

  #ifdef TEST
  Log() << "Socket: Using connecting socket "
        << "FD#" << data -> connectFd_ << ".\n";
  #endif

  #ifdef TEST
  Log() << "Socket: Connecting to port "
        << data -> listenPort_ << ".\n";
  #endif

  if (SocketConnect(data -> connectFd_,
                        (SocketAddress *) &data -> listenAddr_,
                            sizeof(SocketAddress)) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Connection to port "
          << data -> listenPort_ << " failed. Error is "
          << GetError() << " " << Stk(GetErrorString())
          << ".\n";
    #endif

    LogError() << "Connection to port " << data -> listenPort_
               << " failed. Error is " << GetError()
               << " " << Stk(GetErrorString()) << ".\n";

    return 0;
  }

  if (SocketNoInherit(data -> connectFd_, 1) == -1)
  {
    return 0;
  }

  #ifdef TEST
  Log() << "Socket: Connected to port "
        << data -> listenPort_ << ".\n";
  #endif

  StringPrint(data -> acceptData_, "NXCORE cookie=");

  cookieData = RandomGet();

  for (cookieIndex = 0; cookieIndex < 16; cookieIndex++)
  {
    if (cookieIndex % 4 == 0)
    {
      cookieData = RandomGet();
    }

    StringSizePrint(data -> acceptData_ + StringLength("NXCORE cookie=") +
                        2 * cookieIndex, 3, "%02x", cookieData & 0xff);

    cookieData >>= 8;
  }

  StringPrint(data -> acceptData_ +
                  StringLength("NXCORE cookie=") + 32, " ");

  #ifdef TEST
  Log() << "Socket: Writing authentication data "
        << Stk(data -> acceptData_) << ".\n";
  #endif

  if (SocketWrite(data -> connectFd_, data -> acceptData_,
                      StringLength(data -> acceptData_)) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to write the "
          << "authentication data. Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to write the authentication data. "
               << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    return 0;
  }

  return 1;
}

Int SocketPairListener(SocketPairData *data, Int fds[2])
{
  HANDLE threadHandle;
  DWORD  threadId;
  DWORD  threadResult;

  data -> listenIp_ = SocketIPAddress("127.0.0.1");

  data -> listenAddr_.sin_family = AF_INET;

  data -> listenAddr_.sin_port = SocketNetworkShort(data -> listenPort_);

  data -> listenAddr_.sin_addr.s_addr = data -> listenIp_;

  const Char *connectedHost;

  Char readData[DEFAULT_STRING_LENGTH];

  #ifdef TEST
  Log() << "Socket: Creating the listening socket.\n";
  #endif

  SetError(0);

  if ((data -> listenFd_ = SocketCreate(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1 ||
          (data -> connectFd_ = SocketCreate(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1 ||
              SocketBind(data -> listenFd_, (SocketAddress *) &data -> listenAddr_,
                  sizeof(data -> listenAddr_)) == -1 || SocketListen(data ->
                      listenFd_, 1) == -1)
  {
    #ifdef WARNING
    Log() << "Socket: WARNING! Can't create the "
          << "listening socket. Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    goto MakePipeListenerError;
  }

  #ifdef TEST
  Log() << "Socket: Using listening socket "
        << "FD#" << data -> listenFd_ << ".\n";
  #endif

  if (SocketNoInherit(data -> listenFd_, 1) == -1)
  {
    goto MakePipeListenerError;
  }

  #ifdef TEST
  Log() << "Socket: Creating the connecting thread.\n";
  #endif

  if (ThreadCreate(&threadHandle, &threadId, SocketPairConnector, data) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Can't create the "
          << "connecting thread. Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Can't create the connecting thread. "
               << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    goto MakePipeListenerError;
  }

  #ifdef TEST
  Log() << "Socket: Connecting thread created.\n";
  #endif

  if (ThreadJoin(threadHandle, &threadResult) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Can't join the "
          << "connecting thread. Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Can't join the connecting thread. "
               << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    goto MakePipeListenerError;
  }

  #ifdef TEST
  Log() << "Socket: Connecting thread exited with "
        << "code " << threadResult << ".\n";
  #endif

  data -> acceptLen_ = sizeof(data -> acceptAddr_);

  if (threadResult == 0 || (data -> acceptFd_ = SocketAccept(data ->
          listenFd_, (SocketAddress *) &data -> acceptAddr_, &data ->
              acceptLen_)) == -1)
  {
    if (threadResult == 0)
    {
      SetError(WSAECONNREFUSED);
    }

    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to create the "
          << "pipe connection. Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to create the pipe connection. "
               << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    goto MakePipeListenerError;
  }

  #ifdef TEST
  Log() << "Socket: Using accepted socket "
        << "FD#" << data -> acceptFd_ << ".\n";
  #endif

  if (SocketNoInherit(data -> acceptFd_, 1) == -1)
  {
    goto MakePipeListenerError;
  }

  //
  // Check if the connection is actually
  // from localhost.
  //

  connectedHost = SocketConvertHostAddress(data -> acceptAddr_.sin_addr);

  if ((Int) data -> acceptAddr_.sin_addr.s_addr != data -> listenIp_)
  {
    SetError(WSAECONNREFUSED);

    #ifdef ERROR
    Log() << "Socket: ERROR! Refusing the pipe "
          << "connection from host " << Stk(connectedHost)
          << ".\n";
    #endif

    LogError() << "Failed to accept the pipe connection from "
               << "host " << Stk(connectedHost) << ".\n";

    goto MakePipeListenerError;
  }

  //
  // Now authenticate the remote.
  //

  SocketRead(data -> acceptFd_, readData, StringLength(data -> acceptData_));

  if (StringSizeCompare(readData, data -> acceptData_,
                            StringLength(data -> acceptData_)) != 0)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to authenticate the "
          << "pipe connection from host " << Stk(connectedHost)
          << ".\n";
    #endif

    LogError() << "Failed to authenticate the pipe connection "
               << "from host " << Stk(connectedHost) << ".\n";

    goto MakePipeListenerError;
  }

  #ifdef TEST

  Unsigned Short Int connectedPort;

  connectedPort = SocketHostShort(data -> acceptAddr_.sin_port);

  Log() << "Socket: Pipe connection created with host "
        << Stk(connectedHost) << " and port " << connectedPort
        << ".\n";

  Log() << "Socket: Pipe descriptors are listener "
        << data -> listenFd_ << " connected " << data -> connectFd_
        << " accepted " << data -> acceptFd_ << ".\n";

  #endif

  SocketClose(data -> listenFd_);

  fds[0] = data -> connectFd_;
  fds[1] = data -> acceptFd_;

  data -> listenFd_  = -1;
  data -> connectFd_ = -1;
  data -> acceptFd_  = -1;

  return 1;

MakePipeListenerError:

  if (data -> listenFd_ != -1)
  {
    SocketClose(data -> listenFd_);

    data -> listenFd_ = -1;
  }

  if (data -> connectFd_ != -1)
  {
    SocketClose(data -> connectFd_);

    data -> connectFd_ = -1;
  }

  if (data -> acceptFd_ != -1)
  {
    SocketClose(data -> acceptFd_);

    data -> acceptFd_ = -1;
  }

  return -1;
}

Int SocketPair(Int domain, Int type, Int protocol, Int fds[2])
{
  //
  // Unix domain socket are obviously not
  // supported. Force the connected socket
  // to be in the INET domain.
  //

  SocketPairData data;

  data.listenFd_  = -1;
  data.connectFd_ = -1;
  data.acceptFd_  = -1;

  data.listenIp_   = -1;
  data.listenPort_ = -1;

  MemoryZero(&data.listenAddr_, sizeof(SocketAddressInet));
  MemoryZero(&data.acceptAddr_, sizeof(SocketAddressInet));

  data.acceptLen_ = sizeof(data.acceptAddr_);

  *data.acceptData_ = '\0';

  for (Int i = 0; i < 20; i++)
  {
    data.listenPort_ = (RandomGet() + 1025) & 0xffff;

    #ifdef TEST
    Log() << "Socket: Trying with port " << data.listenPort_
          << ".\n";
    #endif

    if (SocketPairListener(&data, fds) == 1)
    {
      //
      // Set the no-delay flag on both ends
      // of the socket pair.
      //

      SocketNoDelay(fds[0], 1);
      SocketNoDelay(fds[1], 1);

      goto SocketPairEnd;
    }
  }

  goto SocketPairError;

SocketPairEnd:

  #ifdef TEST
  Log() << "Socket: Pair created with FD#" << fds[0]
        << " and FD#" << fds[1] << ".\n";
  #endif

  return 0;

SocketPairError:

  #ifdef ERROR
  Log() << "Socket: ERROR! Failed to create the socket pair.\n";
  #endif

  LogError() << "Failed to create the socket pair.\n\n";

  return -1;
}

//
// Write a datagram made of multiple
// fragments on Windows.
//

Int SocketWriteDatagramVector(Int fd, const SocketVector *vector, Int count,
                                  const SocketAddress *address, SocketAddressLength addresslen)
{
  #ifdef DEBUG
  Log() << "Socket: Writing " << count << " datagram "
        << "fragments to FD#" << fd << ".\n";
  #endif

  //
  // Must convert the I/O vector to an
  // array of WSABUF structures.
  //

  Int result;

  WSABUF buffers[count];

  for (result = 0; result < count; result++, vector++)
  {
    buffers[result].len = vector -> iov_len;
    buffers[result].buf = (Char *) vector -> iov_base;
  }

  DWORD total;

  result = WSASendTo(fd, buffers, count, &total, 0, address,
                         addresslen, NULL, NULL);

  #ifdef DEBUG
  Log() << "Socket: Write result is " << result
        << " with " << total << " datagram bytes "
        << "written.\n";
  #endif

  if (result == 0)
  {
    return total;
  }

  #ifdef DEBUG

  Int error = SocketGetError();

  const Char *string = SocketGetErrorString(error);

  if (error != WSAEWOULDBLOCK)
  {
    Log() << "Socket: ERROR! Write error is " << error
          << " " << Stk(string) << ".\n";
  }

  #endif

  SOCKET_RESULT(result, SOCKET_ERROR);
}

#endif // #ifdef WIN32 ...

Int SocketPipe(Int fds[2], Int size)
{
  fds[0] = -1;
  fds[1] = -1;

  if (SocketPair(AF_UNIX, SOCK_STREAM, 0, fds) == -1 ||
          SocketNoInherit(fds[0], 1) == -1 ||
              SocketNoInherit(fds[1], 1) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Can't create the socket "
          << "pipe. Error is " << GetError() << " "
          << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Can't create the socket pipe. "
               << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    if (fds[0] != -1)
    {
      SocketClose(fds[0]);
    }

    if (fds[1] != -1)
    {
      SocketClose(fds[1]);
    }

    return -1;
  }

  //
  // Set the size of the underlying
  // operating system buffer.
  //

  SocketReceiveBuffer(fds[0], size);
  SocketReceiveBuffer(fds[1], size);

  SocketSendBuffer(fds[0], size);
  SocketSendBuffer(fds[1], size);

  return 0;
}

unsigned Int SocketHostAddress(const Char *name)
{
  #ifdef TEST
  Log() << "Socket: Querying IP address of host "
        << Stk(name) << ".\n";
  #endif

  //
  // Set an error and silently return a null
  // address if the name is empty.
  //

  if (name == NULL || *name == '\0')
  {
    #ifdef TEST
    Log() << "Socket: WARNING! Returning an empty IP "
          << "for " << Stk(name) << ".\n";
    #endif

    SetError(EADDRNOTAVAIL);

    return 0;
  }

  hostent *host = gethostbyname(name);

  if (host == NULL)
  {
    //
    // On some Unices gethostbyname() doesn't
    // accept numerical dotted IP addresses,
    // so try inet_addr().
    //

    unsigned Int address = SocketIPAddress(name);

    if (address == INADDR_NONE)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Failed to resolve address "
            << Stk(name) << ".\n";
      #endif

      LogError() << "Failed to resolve address "
                 << Stk(name) << ".\n";

      #ifdef WIN32

      SocketTranslateError();

      #ifdef ERROR
      Log() << "Socket: ERROR! Error is " << GetError()
            << ", " << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Error is " << GetError() << ", "
                 << Stk(GetErrorString()) << ".\n";

      #else

      const Char *label;

      switch (h_errno)
      {
        case HOST_NOT_FOUND:
        {
          label = "The specified host is unknown";

          break;
        }
        case NO_ADDRESS:
        {
          label = "Host does not have an IP address";

          break;
        }
        case NO_RECOVERY:
        {
          label = "Non-recoverable name server error";

          break;
        }
        case TRY_AGAIN:
        {
          label = "Temporary name server error";

          break;
        }
        default:
        {
          //
          // The function can fail with an unknown
          // error if we went out of descriptors.
          //

          if (GetError() == EMFILE)
          {
            h_errno = GetError();

            label = GetErrorString();
          }
          else
          {
            label = "Unknown";
          }

          break;
        }
      }

      #ifdef ERROR
      Log() << "Socket: ERROR! Error is " << h_errno
            << " " << Stk(label) << ".\n";
      #endif

      LogError() << "Error is " << h_errno << " "
                 << Stk(label) << ".\n";

      #endif

      //
      // Set the reported error to "Address
      // not available" since the name query
      // normally reports ETIMEDOUT and this
      // may confuse the user.
      // 

      SetError(EADDRNOTAVAIL);

      return 0;
    }

    return (unsigned Int) address;
  }
  else
  {
    return (*((unsigned Int *) host -> h_addr_list[0]));
  }
}

//
// Retrieve the host name whose IP address is
// represented by the string.
//
/*
FIXME: We should be able to handle both SocketHost adresses
       and addresses in dotted string format. This should
       be a wrapper of the function taking a SocketHost.
*/
const Char *SocketHostName(const Char *ip)
{
  hostent    *host;
  SocketHost address;

  MemoryZero(&address, sizeof(SocketHost));

  address.s_addr = SocketIPAddress(ip);

  if (address.s_addr == INADDR_NONE)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Wrong IPv4 dotted-decimal address "
          << Stk(ip) << ".\n";
    #endif

    LogError() << "Wrong IPv4 dotted-decimal address "
               << Stk(ip) << ".\n";

    return NULL;
  }

  host = gethostbyaddr((Char *) &address, sizeof(SocketHost), AF_INET);

  if (host == NULL)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to resolve address "
          << Stk(ip) << ".\n";
    #endif

    LogError() << "Failed to resolve address "
               << Stk(ip) << ".\n";

    #ifdef WIN32

    #ifdef ERROR
    Log() << "Socket: ERROR! Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    #else

    const Char *label;

    switch (h_errno)
    {
      case HOST_NOT_FOUND:
      {
        label = "The specified host is unknown";

        break;
      }
      case NO_ADDRESS:
      {
        label = "Host does not have an IP address";

        break;
      }
      case NO_RECOVERY:
      {
        label = "Non-recoverable name server error";

        break;
      }
      case TRY_AGAIN:
      {
        label = "Temporary name server error";

        break;
      }
      default:
      {
        //
        // The function can fail with an unknown
        // error if we went out of descriptors.
        //

        if (GetError() == EMFILE)
        {
          h_errno = GetError();

          label = GetErrorString();
        }
        else
        {
          label = "Unknown";
        }

        break;
      }
    }

    #ifdef ERROR
    Log() << "Socket: ERROR! Error is " << h_errno
          << " " << Stk(label) << ".\n";
    #endif

    LogError() << "Error is " << h_errno << " "
               << Stk(label) << ".\n";

    #endif

    return NULL;
  }

/*
FIXME: The hostent structure should be protected by
       allocating the data in Thread Local Storage.
       More simply, the static data can be protected
       by serializing the access to the function in
       a higher level class.
*/
  return host -> h_name;
}

Int SocketHostNumeric(const Char *name)
{
  #ifdef TEST
  Log() << "Socket: Checking host name " << Stk(name)
        << ".\n";
  #endif

  if (name == NULL || *name == '\0')
  {
    #ifdef TEST
    Log() << "Socket: Returning host as invalid.\n";
    #endif

    return -1;
  }

  const Char *next = name;

  while (*next != '\0')
  {
    if ((*next >= '0' && *next <= '9') ||
            *next == '.' || *next == ':')
    {
      next++;
    }
    else
    {
      #ifdef TEST
      Log() << "Socket: Returning host as alphabetic.\n";
      #endif

      return 0;
    }
  }

  #ifdef TEST
  Log() << "Socket: Returning host as numeric.\n";
  #endif

  return 1;
}

const Char *SocketConvertHostAddress(SocketHost host)
{
  //
  // The function uses a static buffer that
  // is overwritten at each call.
  //
/*
FIXME: This must be protected by allocating the data
       in Thread Local Storage.
*/
  return inet_ntoa(host);
}

const Char *SocketConvertIPAddress(unsigned Int ip)
{
  //
  // The function uses a static buffer that
  // is overwritten at each call.
  //
/*
FIXME: This must be protected by allocating the data
       in Thread Local Storage.
*/
  StringPrint(SocketIPString, "%d.%d.%d.%d", (ip & 0xff),
                  (ip >> 8 & 0xff), (ip >> 16 & 0xff),
                      (ip >> 24 & 0xff));

  return SocketIPString;
}

const Char *SocketConvertNameAddress(const Char *name)
{
  return SocketConvertIPAddress(SocketHostAddress(name));
}

Int SocketError(Int fd)
{
  Int error;

  SocketLength len = sizeof(error);

  if (getsockopt(fd, SOL_SOCKET, SO_ERROR,
                     (Char *) &error, &len) == -1)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Failed to get the SO_ERROR option "
          << "for FD#" << fd << ". Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Failed to get the SO_ERROR option "
               << "for FD#" << fd << ". Error is " << GetError()
               << " " << Stk(GetErrorString()) << ".\n";

    return -1;
  }

  #ifdef WIN32

  return SocketTranslateError(error);

  #else

  return error;

  #endif
}

//
// Network information functions.
//

static void NOINLINE SocketNormalizeIPv6Address(struct sockaddr_storage *addr,
                                                    socklen_t *len)
{
  struct sockaddr_in6 *a6 = (struct sockaddr_in6 *) addr;
  struct sockaddr_in *a4  = (struct sockaddr_in *)  addr;

  struct in_addr inaddr;
  uint16_t port;

  if (addr -> ss_family != AF_INET6 ||
          IN6_IS_ADDR_V4MAPPED(&a6 -> sin6_addr) == 0)
  {
    return;
  }

  MemoryCopy(&inaddr, ((Char *) &a6 -> sin6_addr) + 12, sizeof(inaddr));

  port = a6 -> sin6_port;

  MemorySet(a4, 0, sizeof(*a4));

  a4 -> sin_family = AF_INET;

  *len = sizeof(*a4);

  MemoryCopy(&a4 -> sin_addr, &inaddr, sizeof(inaddr));

  a4 -> sin_port = port;
}

//
// Return the local/remote IP-address/hostname
// of the socket. The returned string must be
// freed by the caller.
//

static Char *SocketGetAddress(Int fd, Int local, Int flags)
{
  struct sockaddr_storage addr;
  socklen_t addrlen;

  Char ntop[NI_MAXHOST];
  Char *address = NULL;

  //
  // Get IP address of client.
  //

  addrlen = sizeof(addr);

  MemorySet(&addr, 0, sizeof(addr));

  if (local != 0)
  {
    if (getsockname(fd, (struct sockaddr *) &addr, &addrlen) < 0)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Can't get the local host "
            << "info for FD#" << fd << ".\n";
      #endif

      LogError() << "Can't get the local host info "
                 << "for FD#" << fd << ".\n";

      #ifdef ERROR
      Log() << "Socket: ERROR! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Error is " << GetError() << " "
                 << Stk(GetErrorString()) << ".\n";

      return NULL;
    }
  }
  else
  {
    if (getpeername(fd, (struct sockaddr *) &addr, &addrlen) < 0)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Can't get the remote host "
            << "info for FD#" << fd << ".\n";
      #endif

      LogError() << "Can't get the remote host info "
                 << "for FD#" << fd << ".\n";

      #ifdef ERROR
      Log() << "Socket: ERROR! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Error is " << GetError() << " "
                 << Stk(GetErrorString()) << ".\n";

      return NULL;
    }
  }

  //
  // Work around Linux IPv6 weirdness. 
  //

  if (addr.ss_family == AF_INET6)
  {
    addrlen = sizeof(struct sockaddr_in6);
  }

  SocketNormalizeIPv6Address(&addr, &addrlen);

  //
  // Get the address in literal form.
  //

  Int r;

  if ((r = getnameinfo((struct sockaddr *) &addr, addrlen, ntop,
                           sizeof(ntop), NULL, 0, flags)) != 0)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Can't get the host name "
          << "info for FD#" << fd << ".\n";
    #endif

    LogError() << "Can't get the host name info "
               << "for FD#" << fd << ".\n";

    #ifdef ERROR
    Log() << "Socket: ERROR! Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    return NULL;
  }

  StringSet(address, ntop);

  return address;
}

static Char *SocketGetRemoteAddress(Int fd)
{
  Char *address = NULL;

  if ((address = SocketGetAddress(fd, 0, NI_NUMERICHOST)) == NULL)
  {
    StringSet(address, "UNKNOWN");
  }

  return address;
}

static Char *SocketGetLocalAddress(Int fd)
{
  Char *address = NULL;

  if ((address = SocketGetAddress(fd, 1, NI_NUMERICHOST)) == NULL)
  {
    StringSet(address, "UNKNOWN");
  }

  return address;
}

//
// Return the local or remote port for
// the socket.
//

static Int SocketGetPort(Int fd, Int local)
{
  struct sockaddr_storage from;
  socklen_t fromlen;

  Char strport[NI_MAXSERV];

  //
  // Get IP address of client.
  //
  
  fromlen = sizeof(from);

  MemorySet(&from, 0, sizeof(from));

  if (local != 0)
  {
    if (getsockname(fd, (struct sockaddr *) &from, &fromlen) < 0)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Can't get the local host "
            << "for FD#" << fd << ".\n";
      #endif

      LogError() << "Can't get the local host for FD#"
                 << fd << ".\n";

      #ifdef ERROR
      Log() << "Socket: ERROR! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Error is " << GetError() << " "
                 << Stk(GetErrorString()) << ".\n";

      return -1;
    }
  }
  else
  {
    if (getpeername(fd, (struct sockaddr *) &from, &fromlen) < 0)
    {
      #ifdef ERROR
      Log() << "Socket: ERROR! Can't get the remote host "
            << "for FD#" << fd << ".\n";
      #endif

      LogError() << "Can't get the remote host for FD#"
                 << fd << ".\n";

      #ifdef ERROR
      Log() << "Socket: ERROR! Error is " << GetError()
            << " " << Stk(GetErrorString()) << ".\n";
      #endif

      LogError() << "Error is " << GetError() << " "
                 << Stk(GetErrorString()) << ".\n";

      return -1;
    }
  }

  //
  // Work around Linux IPv6 weirdness.
  //

  if (from.ss_family == AF_INET6)
  {
    fromlen = sizeof(struct sockaddr_in6);
  }

  //
  // Return the port number.
  //

  Int r;

  if ((r = getnameinfo((struct sockaddr *) &from, fromlen, NULL, 0,
          strport, sizeof(strport), NI_NUMERICSERV)) != 0)
  {
    #ifdef ERROR
    Log() << "Socket: ERROR! Can't get the network port.\n";
    #endif

    LogError() << "Can't get the network port.\n";

    #ifdef ERROR
    Log() << "Socket: ERROR! Error is " << GetError()
          << " " << Stk(GetErrorString()) << ".\n";
    #endif

    LogError() << "Error is " << GetError() << " "
               << Stk(GetErrorString()) << ".\n";

    return -1;
  }

  return StringToInt(strport);
}

static Int SocketGetLocalPort(Int fd)
{
  return SocketGetPort(fd, 1);
}

static Int SocketGetRemotePort(Int fd)
{
  return SocketGetPort(fd, 0);
}

Int SocketNetworkInfo(Int fd, Char *&lhost, Int &lport,
                          Char *&rhost, Int &rport)
{
  lhost = SocketGetLocalAddress(fd);
  rhost = SocketGetRemoteAddress(fd);

  lport = SocketGetLocalPort(fd);
  rport = SocketGetRemotePort(fd);

  #ifdef TEST
  Log() << "Socket: Descriptor FD#" << fd
        << " local host " << Stk(lhost) << " port "
        << Itk(lport) << ".\n";

  Log() << "Socket: Descriptor FD#" << fd
        << " remote host " << Stk(rhost) << " port "
        << Itk(rport) << ".\n";
  #endif

  if (lhost != NULL && rhost != NULL &&
          lport != -1 && rport != -1)
  {
    return 1;
  }

  #ifdef ERROR
  Log() << "Socket: ERROR! Failed to collect network "
        << "information for FD#" << fd << ".\n";
  #endif

  LogError() << "Failed to collect network information "
             << "for FD#" << fd << ".\n";

  return -1;
}
