/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Socket_H
#define Socket_H

#ifndef WIN32

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#else

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#ifdef  WINVER
#undef  WINVER
#endif

#define WINVER 0x501

#include <winsock2.h>
#include <windows.h>
#include <ws2tcpip.h>

#ifdef ERROR
#undef ERROR
#endif

#endif

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

#if !defined(__linux__) && !defined(WIN32)
#include <netinet/in_systm.h>
#endif

#if !defined(__CYGWIN32__) && !defined(WIN32)
#include <sys/un.h>
#endif

#if defined(__EMX__ ) || defined(__CYGWIN32__)

struct sockaddr_un
{
  u_short sun_family;
  char sun_path[108];
};

#endif

#ifdef __sun
#include <stropts.h>
#include <sys/filio.h>
#endif

#include "Types.h"

#ifdef DEBUG

#include "Log.h"

#endif

//
// Initialize the networking stack.
//

Int SocketStartup();
Int SocketCleanup();

//
// Missing defines.
//

#ifdef WIN32

#define SHUT_RD     SD_RECEIVE
#define SHUT_WR     SD_SEND
#define SHUT_RDWR   SD_BOTH

struct sockaddr_un { };

#endif

#define SocketShutdownRead       SHUT_RD
#define SocketShutdownWrite      SHUT_WR
#define SocketShutdownReadWrite  SHUT_RDWR

//
// Socket types and structures.
//

#define SocketHost           struct in_addr

#define SocketLength         socklen_t

#define SocketAddress        struct sockaddr
#define SocketAddressInet    struct sockaddr_in
#define SocketAddressUnix    struct sockaddr_un

#define SocketAddressLength  socklen_t

//
// Return the operating system's handle
// associated to the socket descriptor.
//

#define SocketGetId(fd)  (fd)

//
// Vector of data buffers.
//

#ifdef WIN32

typedef struct
{
  void  *iov_base;
  size_t iov_len;
}
DataVector;

#else

typedef struct iovec DataVector;

#endif

//
// Basic socket operations.
//

typedef DataVector SocketVector;

#ifdef WIN32

Int SocketCreate(Int domain, Int type, Int protocol);
Int SocketBind(Int sockfd, const SocketAddress *addr, SocketAddressLength addrlen);
Int SocketListen(Int sockfd, Int backlog);
Int SocketConnect(Int sockfd, const SocketAddress *addr, SocketAddressLength addrlen);
Int SocketAccept(Int sockfd, SocketAddress *addr, SocketAddressLength *addrlen);
Int SocketRead(Int fd, void *buffer, Int size);
Int SocketWrite(Int fd, const void *buffer, Int size);
Int SocketShutdown(Int fd, Int how);
Int SocketClose(Int fd);

Int SocketPair(Int domain, Int type, Int protocol, Int fds[2]);

Int SocketReadVector(Int fd, const SocketVector *vector, Int count);
Int SocketWriteVector(Int fd, const SocketVector *vector, Int count);

Int SocketReadDatagram(Int fd, void *buffer, Int size, SocketAddress *address,
                           SocketAddressLength *addresslen);
Int SocketWriteDatagram(Int fd, const void *buffer, Int size,
                            const SocketAddress *address, SocketAddressLength addresslen);
Int SocketWriteDatagramVector(Int fd, const SocketVector *vector, Int count,
                                  const SocketAddress *address, SocketAddressLength addresslen);

Int SocketDuplicate(Int fd);
Int SocketClone(Int fd1, Int fd2);

#define SocketSetError(e)  WSASetLastError((e))
#define SocketGetError()   WSAGetLastError()

const Char *SocketGetErrorString(Int error);
const Char *SocketGetErrorName(Int error);

Int SocketTranslateError(Int error);

void SocketTranslateError();

#else // #ifdef WIN32 ...

#define SocketCreate(domain, type, protocol)     ::socket((domain), (type), (protocol))
#define SocketBind(sockfd, addr, addrlen)        ::bind((sockfd), (addr), (addrlen))
#define SocketListen(sockfd, backlog)            ::listen((sockfd), (backlog))
#define SocketConnect(sockfd, addr, addrlen)     ::connect((sockfd), (addr), (addrlen))
#define SocketAccept(sockfd, addr, addrlen)      ::accept((sockfd), (addr), (addrlen))
#define SocketShutdown(fd, how)                  ::shutdown((fd), (how))
#define SocketClose(fd)                          ::close((fd))
#define SocketPair(domain, type, protocol, fds)  ::socketpair((domain), (type), (protocol), (fds))

#define SocketRead(fd, buffer, size)   ::recv((fd), (buffer), (size), 0)
#define SocketWrite(fd, buffer, size)  ::send((fd), (buffer), (size), 0)

#define SocketReadDatagram(fd, buffer, size, address, addresslen) \
                                              ::recvfrom((fd), (buffer), (size), 0, (address), (addresslen))
#define SocketWriteDatagram(fd, buffer, size, address, addresslen) \
                                              ::sendto((fd), (buffer), (size), 0, (address), (addresslen))

#define SocketDuplicate(fd)    ::dup((fd))
#define SocketClone(fd1, fd2)  ::dup2((fd1), (fd2))

//
// Implement scattered I/O on Unix
// using recvmsg() and sendmsg().
//

inline Int SocketReadVector(Int fd, const SocketVector *vector, Int count)
{
  struct msghdr message;

  message.msg_name    = NULL;
  message.msg_namelen = 0;

  message.msg_iov    = (SocketVector *) vector;
  message.msg_iovlen = count;

  message.msg_control    = NULL;
  message.msg_controllen = 0;

  message.msg_flags = 0;

  return recvmsg(fd, &message, 0);
}

inline Int SocketWriteVector(Int fd, const SocketVector *vector, Int count)
{
  struct msghdr message;

  message.msg_name    = NULL;
  message.msg_namelen = 0;

  message.msg_iov    = (SocketVector *) vector;
  message.msg_iovlen = count;

  message.msg_control    = NULL;
  message.msg_controllen = 0;

  message.msg_flags = 0;

  return sendmsg(fd, &message, 0);
}

inline Int SocketWriteDatagramVector(Int fd, const SocketVector *vector, Int count,
                                         const SocketAddress *address, SocketAddressLength addresslen)
{
  struct msghdr message;

  message.msg_name    = (SocketAddress *) address;
  message.msg_namelen = addresslen;

  message.msg_iov    = (SocketVector *) vector;
  message.msg_iovlen = count;

  message.msg_control    = NULL;
  message.msg_controllen = 0;

  message.msg_flags = 0;

  return sendmsg(fd, &message, 0);
}

#endif // #ifdef WIN32 ... #else ...

//
// Create a bidirectional communication
// facility used for signaling between
// processes or threads. Both ends have
// the close-on-exec flag set on return.
//

Int SocketPipe(Int fds[2], Int size = 65536);

//
// Return the IP address in network byte
// order of the named host.
//

unsigned Int SocketHostAddress(const Char *name);

//
// Return the name of the host associated
// to the IPv4 address in dotted format.
//

const Char *SocketHostName(const Char *ip);

//
// Return 1 if the name is a numeric IP,
// or otherwise 0.
//

Int SocketHostNumeric(const Char *name);

//
// Convert the IP address from the dotted
// notation to a binary number in network
// byte order.
//

#define SocketIPAddress(address)  ::inet_addr((address))

//
// Return the INADDR_ANY constant converted
// to network byte order. Don't include the
// namespace where the system can implement
// these as macros.
//

#ifdef WIN32

#define SocketAnyAddress()  ::htonl(INADDR_ANY)

#else

#define SocketAnyAddress()  htonl(INADDR_ANY)

#endif

//
// Convert the given integer from host
// byte order to network byte order.
//

#ifdef WIN32

#define SocketNetworkLong(value)   ::htonl((value))
#define SocketNetworkShort(value)  ::htons((value))

#else

#define SocketNetworkLong(value)   htonl((value))
#define SocketNetworkShort(value)  htons((value))

#endif

//
// Convert the given integer from network
// byte order to host byte order.
//

#ifdef WIN32

#define SocketHostLong(value)      ::ntohl((value))
#define SocketHostShort(value)     ::ntohs((value))

#else

#define SocketHostLong(value)      ntohl((value))
#define SocketHostShort(value)     ntohs((value))

#endif

//
// Convert the IP address from an in_addr
// structure to a dotted string.
//

const Char *SocketConvertHostAddress(SocketHost host);

//
// Convert the IP address from an integer
// in network byte order to a string.
//

const Char *SocketConvertIPAddress(unsigned Int ip);

//
// Convert the named DNS host to a dotted
// string.
//

const Char *SocketConvertNameAddress(const Char *name);

//
// Configurable socket options.
//

#define SocketOptionNoBlocking      0
#define SocketOptionNoInherit       1
#define SocketOptionNoDelay         2
#define SocketOptionNoFlush         3

#define SocketOptionReceiveBuffer   4
#define SocketOptionSendBuffer      5
#define SocketOptionLowDelay        6
#define SocketOptionKeepAlive       7
#define SocketOptionLingerTimeout   8
#define SocketOptionReuseAddress    9
#define SocketOptionQuickAck        10

//
// Query the socket option.
//

Int SocketQuery(Int fd, Int option);

//
// Set the specific socket option.
//

Int SocketNoBlocking(Int fd, Int value);
Int SocketNoInherit(Int fd, Int value);
Int SocketNoDelay(Int fd, Int value);
Int SocketNoFlush(Int fd, Int value);

Int SocketReceiveBuffer(Int fd, Int size);
Int SocketSendBuffer(Int fd, Int size);
Int SocketLowDelay(Int fd, Int value);
Int SocketKeepAlive(Int fd, Int value);
Int SocketLingerTimeout(Int fd, Int timeout);
Int SocketReuseAddress(Int fd, Int value);
Int SocketQuickAck(Int fd, Int value);

//
// Get the socket information.
//

Int SocketBytesReadable(Int fd);
Int SocketBytesWritable(Int fd);
Int SocketBytesQueued(Int fd);

//
// Check the result of the asynchronous
// connection attempt or the error state
// of the socket.
//

Int SocketError(Int fd);

//
// Obtain the local and remote host and
// port data about the connected socket.
//

Int SocketNetworkInfo(Int fd, Char *&localHost, Int &localPort,
                          Char *&remoteHost, Int &remotePort);

//
// Inline version of the above function
// providing direct access to the system
// call.
//

inline Int SocketBytesReadable(Int fd, Int *readable)
{
  unsigned long t;

  #ifndef WIN32

  Int result = ioctl(fd, FIONREAD, &t);

  #else

  Int result;

  if ((result = ioctlsocket(fd, FIONREAD, &t)) == SOCKET_ERROR)
  {
    SocketTranslateError();
  }

  #endif

  #ifdef DEBUG
  Log() << "Socket: Bytes readable from FD#"
        << fd << " are " << t << " with result "
        << result << ".\n";
  #endif

  *readable = (int) t;

  return result;
}

//
// Print the IP address in dotted notation.
//

#define Sip(value) SocketConvertIPAddress((value))

#endif /* Socket_H */
