/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Dir.h"
#include "Time.h"
#include "String.h"
#include "Memory.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

#ifdef WIN32

Char *StringToken(Char *string, const Char *delimiter, Char **state)
{
  Char *result;

  if (string == NULL)
  {
    string = *state;
  }

  string += strspn(string, delimiter);

  if (*string == '\0')
  {
    return NULL;
  }

  result = string;

  string += strcspn(string, delimiter);

  if (*string != '\0')
  {
    *string++ = '\0';
  }

  *state = string;

  return result;
}

#endif

void StringAlloc(Char *&destination)
{
  destination = new Char[DEFAULT_STRING_LENGTH];

  *destination = '\0';
}

void StringAlloc(Char *&destination, const Char *source)
{
  StringAlloc(destination);

  if (source != NULL)
  {
    StringSizeCopy(destination, source, DEFAULT_STRING_LENGTH - 1);

    *(destination + DEFAULT_STRING_LENGTH - 1) = '\0';
  }
}

void StringAlloc(Char *&destination, Int size)
{
  destination = new Char[size];

  *destination = '\0';
}

void StringInit(Char *&destination, const Char *source)
{
  destination = NULL;

  if (source != NULL)
  {
    StringSet(destination, source);
  }
}

void StringInit(Char *&destination, const Char *buffer, Int size)
{
  destination = NULL;

  if (buffer != NULL)
  {
    destination = new Char[size + 1];

    //
    // End the copy as soon as a zero
    // is encountered.
    //

    StringSizeCopy(destination, buffer, size);

    //
    // Ensure that the string is null
    // terminated.
    //

    *(destination + size) = '\0';
  }
}

Char *StringInit(const Char *source)
{
  Char *destination;

  StringInit(destination, source);

  return destination;
}

Char *StringInit(const Char *buffer, Int size)
{
  Char *destination;

  StringInit(destination, buffer, size);

  return destination;
}

void StringSet(Char *&destination, const Char *source)
{
  if (destination != source)
  {
    if (destination != NULL)
    {
      delete [] destination;
    }

    if (source == NULL)
    {
      destination = NULL;
    }
    else
    {
      destination = new Char[StringLength(source) + 1];

      StringCopy(destination, source);
    }
  }
}

void StringReset(Char *&destination)
{
  if (destination != NULL)
  {
    delete [] destination;

    destination = NULL;
  }
}

void StringAdd(Char *&destination, const Char *source1, const Char *source2,
                   const Char *source3, const Char *source4)
{
  Int length = (destination == NULL ? 0 : StringLength(destination));

  length += (source1 == NULL ? 0 : StringLength(source1));
  length += (source2 == NULL ? 0 : StringLength(source2));
  length += (source3 == NULL ? 0 : StringLength(source3));
  length += (source4 == NULL ? 0 : StringLength(source4));

  if (length == 0)
  {
    return;
  }

  Char *data = new Char[length + 1];

  if (destination != NULL)
  {
    StringCopy(data, destination);
  }
  else
  {
    *data = '\0';
  }

  if (source1 != NULL)
  {
    StringCopy(data + StringLength(data), source1);
  }

  if (source2 != NULL)
  {
    StringCopy(data + StringLength(data), source2);
  }

  if (source3 != NULL)
  {
    StringCopy(data + StringLength(data), source3);
  }

  if (source4 != NULL)
  {
    StringCopy(data + StringLength(data), source4);
  }

  delete [] destination;

  destination = data;
}

void StringAdd(Char *&destination, const Char *buffer, Int size)
{
  if (buffer == NULL || size == 0)
  {
    return;
  }

  Int length = (destination == NULL ? 0 : StringLength(destination));

  Char *data = new Char[length + size + 1];

  if (destination != NULL)
  {
    MemoryCopy(data, destination, length);

    MemoryCopy(data + length, buffer, size);

    *(data + length + size) = '\0';
  }
  else
  {
    MemoryCopy(data, buffer, size);

    *(data + size) = '\0';
  }

  delete [] destination;

  destination = data;
}

Char *StringSetIp(Char *&destination, unsigned Int ip)
{
  if (destination != NULL)
  {
    delete [] destination;
  }

  destination = new Char[16];

  return StringCopyIp(destination, ip);
}

Char *StringCopyIp(Char *destination, unsigned Int ip)
{
  StringPrint(destination, "%d", ip & 0xff);
  StringPrint(destination + StringLength(destination), ".%d", ip >> 8 & 0xff);
  StringPrint(destination + StringLength(destination), ".%d", ip >> 16 & 0xff);
  StringPrint(destination + StringLength(destination), ".%d", ip >> 24 & 0xff);

  return destination;
}

Char *StringSetTimeval(Char *&destination, const Timeval &ts)
{
  if (destination != NULL)
  {
    delete [] destination;
  }

  destination = new Char[25];

  return StringCopyTimeval(destination, ts);
}

Char *StringCopyTimeval(Char *destination, const Timeval &ts)
{
  #ifdef WIN32

  StringSizeCopy(destination, ctime((time_t *) &ts.tv_sec), 24);

  *(destination + 24) = '\0';

  #else

  ctime_r((time_t *) &ts.tv_sec, destination);

  *(destination + 24) = '\0';

  #endif

  return destination;
}

Char *StringSetMsTimeval(Char *&destination, const Timeval &ts)
{
  if (destination != NULL)
  {
    delete [] destination;
  }

  destination = new Char[17];

  return StringCopyMsTimeval(destination, ts);
}

Char *StringCopyMsTimeval(Char *destination, const Timeval &ts)
{
  struct tm *time;
/*
FIXME: The _localtime32() and _localtime64()
       functions are not defined on MinGW64,
       apparently due to a conflict with how
       OpenSSL is built.
*/
  #ifdef WIN32

  #ifdef __x86_64__

  __time64_t t = ts.tv_sec;

  time = localtime(&t);

  #else // #ifdef __x86_64__ ...

  time = localtime(&ts.tv_sec);

  #endif // #ifdef __x86_64__ ...

  #else // #ifdef WIN32 ...

  struct tm buffer;

  time = localtime_r((time_t *) &ts.tv_sec, &buffer);

  #endif // #ifdef WIN32 ... #else ...

  StringPrint(destination, "%02d:%02d:%02d %07.3f",
                  time -> tm_hour, time -> tm_min, time -> tm_sec,
                      (Double) ts.tv_usec / 1000);

  return destination;
}

Char *StringPurge(Char *string)
{
  if (string != NULL)
  {
    Char *next = string;
  
    while (*next != '\0')
    {
      //
      // Don't stop at the special characters
      // used as escape in the option string.
      //
 
      switch (*next)
      {
        case 0x1f:
        {
          break;
        }
        case 0x1d:
        {
          break;
        }
        case 0x1e:
        {
          break;
        }
        default:
        {
          if (CharIsPrint(*next) == 0)
          {
            *next = '\0';

            return string;
          }

          break;
        }
      }

      next++;
    }
  }

  return string;
}

//
// String search and replace functions.
//

const Char *StringHead(const Char *string, const Char *head)
{
  #ifdef TEST
  Log() << "String: Finding heading string "
        << Stk(head) << " in " << Stk(string)
        << ".\n";
  #endif

  Int length = StringLength(head);

  if (StringSizeCompare(string, head, length) == 0)
  {
    #ifdef TEST
    Log() << "String: String found with length "
          << length << ".\n";
    #endif

    return string;
  }

  return NULL;
}

const Char *StringTrail(const Char *string, const Char *trail)
{
  #ifdef TEST
  Log() << "String: Finding trailing string "
        << Stk(trail) << " in " << Stk(string)
        << ".\n";
  #endif

  Int length = StringLength(trail);
  Int offset = StringLength(string);

  if (offset < length)
  {
    return NULL;
  }

  offset -= length;

  if (StringSizeCompare(string + offset, trail, length) == 0)
  {
    #ifdef TEST
    Log() << "String: String found with offset "
          << offset << ".\n";
    #endif

    return string + offset;
  }

  return NULL;
}

Int StringReplace(Char *&string, const Char *first, const Char *second)
{
  #ifdef TEST
  Log() << "String: Replacing " << Stk(first)
        << " with " << Stk(second) << " in "
        << Stk(string) << ".\n";
  #endif

  Int stringlen = StringLength(string);
  Int firstlen  = StringLength(first);
  Int secondlen = StringLength(second);

  Int found = 0;

  const Char *pstring = string;
  const Char *pfirst;

  //
  // Find how many times the requested
  // string occurs in the source string.
  //

  for (;;)
  {
    pfirst = StringFind(pstring, first);

    if (pfirst == 0)
    {
      break;
    }

    found++;

    pstring = pfirst + firstlen;

    if (pstring >= string + stringlen)
    {
      break;
    }
  }

  if (found == 0)
  {
    #ifdef TEST
    Log() << "String: Pattern " << Stk(first)
          << " not found.\n";
    #endif

    return 0;
  }

  //
  // Create a new buffer to hold the
  // destination string.
  //

  Int replacelen = stringlen + found * (secondlen - firstlen) + 1;

  Char *replace = new Char[replacelen];

  //
  // Copy the original string to the
  // destination while replacing all
  // the instances of the requested
  // string.
  //

  Char *preplace = replace;

  Int skiplen;

  pstring = string;

  for (;;)
  {
    pfirst = StringFind(pstring, first);

    if (pfirst == 0)
    {
      break;
    }

    skiplen = pfirst - pstring;

    //
    // Copy the section until the next
    // occurence of the first string.
    //

    StringSizeCopy(preplace, pstring, skiplen);

    preplace += skiplen;

    //
    // Copy the replacement string.
    //

    MemoryCopy(preplace, second, secondlen);

    preplace += secondlen;

    pstring = pfirst + firstlen;

    if (pstring >= string + stringlen)
    {
      break;
    }
  }

  //
  // Copy the remaining part of the
  // string.
  //

  StringCopy(preplace, pstring);

  delete [] string;

  string = replace;

  #ifdef TEST
  Log() << "String: Found " << found
        << " occurrences new string "
        << Stk(string) << ".\n";
  #endif

  return found;
}

Int StringHeadReplace(Char *&string, const Char *first, const Char *second)
{
  #ifdef TEST
  Log() << "String: Replacing " << Stk(first)
        << " with " << Stk(second) << " at "
        << "head of " << Stk(string) << ".\n";
  #endif

  Int stringlen = StringLength(string);
  Int firstlen  = StringLength(first);
  Int secondlen = StringLength(second);

  if (StringSizeCompare(string, first, firstlen) == 0)
  {
    Int replacelen = stringlen - firstlen + secondlen + 1;

    Char *replace = new Char[replacelen];

    StringCopy(replace, second);
    StringCopy(replace + secondlen, string + firstlen);

    delete [] string;

    string = replace;

    #ifdef TEST
    Log() << "String: Found occurrence "
          << "new string " << Stk(string)
          << ".\n";
    #endif

    return 1;
  }

  #ifdef TEST
  Log() << "String: Pattern " << Stk(first)
        << " not found.\n";
  #endif

  return 0;
}

Int StringVersion(const Char *string, Int &major, Int &minor, Int &patch, Int &build)
{
  #ifdef TEST
  Log() << "String: Parsing version in "
        << Stk(string) << ".\n";
  #endif

  major = -1;
  minor = -1;
  patch = -1;
  build = -1;

  Int found = StringScan(string, "%i.%i.%i_%i", &major, &minor,
                             &patch, &build);

  #ifdef TEST
  Log() << "String: Found version major " << major
        << " minor " << minor << " patch " << patch
        << " build " << build << " result " << found
        << ".\n";
  #endif

  return found;
}


//
// Other string utility functions.
//

Int StringCount(const Char *buffer, Int length, Char separator)
{
  const Char *start = buffer;
  const Char *end   = start + length;
  const Char *next  = start;

  Int count = 0;

  do
  {
    if (*next++ == separator)
    {
      count++;
    }
  }
  while (next < end);

  //#ifdef TEST
  Log() << "String: Found " << count << " lines "
        << "with separator " << Int(separator)
        << ".\n";
  //#endif

  return count;
}

Int StringIsBackDirectory(const Char *string)
{
  Char path[7];

  const Char *slash = DirSlashString;

  StringPrint(path, "..%s", slash);

  Int stringlen = StringLength(string);
  Int pathlen   = StringLength(path);

  if (pathlen > stringlen)
  {
    return 0;
  }
  else if (StringSizeCompare(string, path, pathlen) == 0)
  {
    return 1;
  }

  StringPrint(path, "%s..%s", slash, slash);

  return (StringFind(string + pathlen, path) != NULL);
}

Int StringIsAbsolutePath(const Char *string)
{
  Char path[3];

  const Char *slash = DirSlashString;

  StringPrint(path, "%s", slash);

  Int pathlen = StringLength(path);

  return (StringSizeCompare(string, path, pathlen) == 0);
}
