/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef String_H
#define String_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "Types.h"
#include "Time.h"

//
// Size of a pre-allocated string buffer.
//

#define DEFAULT_STRING_LENGTH  1024

#define StringCopy(destination, source) strcpy((destination), (source))
#define StringSizeCopy(destination, source, size) strncpy((destination), (source), (size))

#define StringCat(destination, source) strcat((destination), (source))
#define StringSizeCat(destination, source, size) strncat((destination), (source), (size))

#define StringCompare(first, second) strcmp((first), (second))
#define StringSizeCompare(first, second, size) strncmp((first), (second), (size))

#define StringCaseCompare(first, second) strcasecmp((first), (second))
#define StringSizeCaseCompare(first, second, size) strncasecmp((first), (second), (size))

#define StringLength(string) strlen((string))

#define StringFind(string, substring) strstr((string), (substring))
#define StringCaseFind(string, substring) strcasestr((string), (substring))

#define StringToInt(string)       atoi((string))
#define StringToLong(string)      atol((string))
#define StringToLongLong(string)  atoll((string))
#define StringToFloat(string)     atof((string))

#define StringChar(string, value) strchr((string), (value))
#define StringRightChar(string, value) strrchr((string), (value))

#define StringPrint     sprintf
#define StringSizePrint snprintf
#define StringScan      sscanf

#define CharIsAlpha(value) isalpha((value))
#define CharIsAlnum(value) isalnum((value))
#define CharIsDigit(value) isdigit((value))
#define CharIsPrint(value) isprint((value))
#define CharIsHex(value)   isxdigit((value))

#define CharToLower(value) tolower((value))
#define CharToUpper(value) toupper((value))

//
// Check if the the string is null or empty.
//

inline Int StringEmpty(const Char *string)
{
  return (string == NULL || *string == '\0');
}

//
// Null terminate the buffer for safety at
// the end of the known storage.
//

inline void StringEnd(Char *string, Int size)
{
  *(string + size - 1) = '\0';
}

//
// Reimplement strtok_r() on Windows since the
// equivalent strtok_s() function may be not
// available in the runtime.
//

#ifndef WIN32

#define StringToken(string, delimiter, state) strtok_r((string), (delimiter), (state))

#else

Char *StringToken(Char *string, const Char *delimiter, Char **state);

#endif

//
// Create an empty string with a fixed
// storage size.
//

void StringAlloc(Char *&destination);

//
// Create an empty string of fixed size
// and copy the content from the source.
//

void StringAlloc(Char *&destination, const Char *source);

//
// Create an empty string of the given
// storage size.
//

void StringAlloc(Char *&destination, Int size);

//
// Set the destination to NULL, then copy
// from the source by allocating memory.
//

void StringInit(Char *&destination, const Char *source);

//
// Make a null terminated string from a
// buffer.
//
 
void StringInit(Char *&destination, const Char *buffer, Int size);

//
// Versions returning the new string.
//

Char *StringInit(const Char *source);

Char *StringInit(const Char *buffer, Int size);

//
// Copy a string by allocating memory for
// the destination.
//

void StringSet(Char *&destination, const Char *source);

//
// Reset the given string and deallocate
// the memory
//

void StringReset(Char *&destination);

//
// Concatenate the strings reallocating
// memory for the destination.
//
 
void StringAdd(Char *&destination, const Char *source1 = NULL,
                   const Char *source2 = NULL, const Char *source3 = NULL,
                       const Char *source4 = NULL);

void StringAdd(Char *&destination, const Char *buffer, Int size);

//
// Allocate memory for the string and fill
// it with the given IP address in dotted
// notation.
//

Char *StringSetIp(Char *&destination, unsigned Int ip);
Char *StringCopyIp(Char *destination, unsigned Int ip);

//
// Write the timestamp to the string.
//

Char *StringSetTimeval(Char *&destination, const Timeval &ts);
Char *StringCopyTimeval(Char *destination, const Timeval &ts);

Char *StringSetMsTimeval(Char *&destination, const Timeval &ts);
Char *StringCopyMsTimeval(Char *destination, const Timeval &ts);

//
// Purge the non-printable characters from the
// string. Accept the special characters used
// to exchange options. Terminate the string
// with zero at the first offending character.
//

Char *StringPurge(Char *string);

//
// Extract the version information from the
// string.
//

Int StringVersion(const Char *string, Int &major, Int &minor, Int &patch, Int &build);

//
// String search and replace functions.
//

const Char *StringHead(const Char *string, const Char *head);

const Char *StringTrail(const Char *string, const Char *trail);

Int StringReplace(Char *&string, const Char *first, const Char *second);

Int StringHeadReplace(Char *&string, const Char *first, const Char *second);

//
// Return how many lines of text are in the
// buffer, based on the requesed separator.
// By passing '\0' the function will return
// the number of null terminated strings.
//

Int StringCount(const Char *buffer, Int length, Char separator = '\0');

inline Int StringLines(const Char *buffer, Int length)
{
  return StringCount(buffer, length, '\n');
}

//
// Search the "../" or "..\\" sub-strings in
// the given string. True if any such string
// is found.
//

Int StringIsBackDirectory(const Char *string);

//
// True if the string represents an absolute
// path on the filesystem.
//

Int StringIsAbsolutePath(const Char *string);

inline Int StringIsRelativePath(Char *string)
{
  return (StringIsAbsolutePath(string) == 0);
}

#endif /* String_H */
