/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "StringList.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

StringList::StringList(Int exclusive)
{
  exclusive_ = exclusive;
}

StringList::~StringList()
{
  removeStrings();
}

StringList::Iterator StringList::getIterator(Int index)
{
  Iterator j;

  Int size = getSize();
  Int i;

  if (index < 0 || index >= size)
  {
    return getEnd();
  }
  else if (index == 0)
  {
    j = getBegin();
  }
  else if (index == size - 1)
  {
    j = --(getEnd());
  }
  else if (index > (size >> 1))
  {
    i = size - 1;

    j = getEnd();

    while (i-- >= index) j--;
  }
  else
  {
    i = 0;

    j = getBegin();

    while (i++ < index) j++;
  }

  return j;
}

const Char *StringList::getString(Int index) const
{
  StringList *list;
  StringList::Iterator j;

  list = NonConstCast(StringList *, this);

  j = list -> getIterator(index);

  return getString(j);
}

void StringList::addString(const Char *string)
{
  if (exclusive_ == 1)
  {
    for (Iterator j = getBegin(); j != getEnd(); j++)
    {
      if (StringCompare(getString(j), string) == 0)
      {
        #ifdef WARNING
        Log() << "StringList: WARNING! String " << Stk(string)
              << " already present in the list.\n";
        #endif

        return;
      }
    }
  }

  Char *buffer = NULL;

  StringSet(buffer, string);

  VoidList::addValue(VoidCast(buffer));
}

void StringList::insertString(const Char *string, Iterator iterator)
{
  if (exclusive_ == 1)
  {
    for (Iterator j = getBegin(); j != getEnd(); j++)
    {
      if (StringCompare(getString(j), string) == 0)
      {
        #ifdef WARNING
        Log() << "StringList: WARNING! String " << Stk(string)
              << " already present in the list.\n";
        #endif

        return;
      }
    }
  }

  Char *buffer = NULL;

  StringSet(buffer, string);

  VoidList::insertValue(VoidCast(buffer), iterator);
}

void StringList::removeString(Iterator i)
{
  Char *buffer = (Char *) getString(i);

  if (buffer == NULL)
  {
    return;
  }

  StringReset(buffer);

  VoidList::removeValue(i);
}

void StringList::removeString(const Char *string)
{
  Iterator j = findString(string);

  if (j == getEnd())
  {
    return;
  }

  Char *buffer = (Char *) getString(j);

  StringReset(buffer);

  VoidList::removeValue(j);
}

void StringList::removeStrings()
{
  Iterator i = getBegin();

  while (i != getEnd())
  {
    Iterator j = i++;

    removeString(j);
  }

  VoidList::removeValues();
}

void StringList::replaceString(Iterator iterator, const Char *to)
{
  if (exclusive_ == 1)
  {
    for (Iterator j = getBegin(); j != getEnd(); j++)
    {
      if (StringCompare(getString(j), to) == 0)
      {
        #ifdef WARNING
        Log() << "StringList: WARNING! String " << Stk(to)
              << " already present in the list.\n";
        #endif

        return;
      }
    }
  }

  Char *from = (Char *) *iterator;
  Char *buffer;

  StringInit(buffer, to);

  *iterator = buffer;

  StringReset(from);
}

StringList::Iterator StringList::findString(const Char *string)
{
  if (string == NULL)
  {
    return getEnd();
  }

  for (Iterator j = getBegin(); j != getEnd(); j++)
  {
    if (StringCompare(getString(j), string) == 0)
    {
      return j;
    }
  }

  return getEnd();
}

StringList *StringList::findSubString(const Char *substring,
                                          StringList *list)
{
  if (list == NULL)
  {
    list = new StringList();
  }

  for (Iterator j = getBegin(); j != getEnd(); j++)
  {
    if (StringFind(getString(j), substring) != NULL)
    {
      list -> addString(getString(j));
    }
  }

  return list;
}

StringList *StringList::split(Iterator iterator, const Char *sep,
                                  SplitType type)
{
  Char *string = (Char *) getString(iterator);

  if (string == NULL)
  {
    return NULL;
  }

  return split(string, sep, type);
}

StringList *StringList::split(const Char *string, const Char *sep,
                                  SplitType type)
{
  if (string == NULL)
  {
    return NULL;
  }

  if (type == KeepEmpty)
  {
    return splitAndKeepEmpty(string, sep);
  }
  else if (type == SkipEmpty)
  {
    return splitAndSkipEmpty(string, sep);
  }

  #ifdef WARNING
  Log() << "StringList: WARNING! Split type " << Itk(type)
        << " unknown.\n";
  #endif

  return NULL;
}

StringList *StringList::splitAndKeepEmpty(const Char *string, const Char *sep)
{
  if (string == NULL || sep == NULL)
  {
    return NULL;
  }

/*
FIXME: Extend this method to support a string
       separator. Now it only supports a single
       character.
*/
  Char separator;

  if (StringLength(sep) > 1)
  {
    Log() << "StringList: WARNING! The KeepEmpty split mode currently "
          << "only support a single character separator.\n";
    Log() << "StringList: WARNING! Converting separator " << Stk(sep)
          << " to '" << *sep << "'.\n";
  }

  separator = *sep;

  Int stringLength = StringLength(string);

  StringList *list = new StringList();

  //
  // Can't use StringToken() because it
  // doesn't support empty fields.
  //

  Char field[stringLength + 1];
  Int fieldPos;

  *field = '\0';
  fieldPos = 0;

  for (int i = 0; i < stringLength; i++)
  {
    if (string[i] == separator)
    {
      field[fieldPos] = '\0';

      list -> addString(field);

      *field = '\0';
      fieldPos = 0;

      continue;
    }

    field[fieldPos++] = string[i];
  }

  field[fieldPos] = '\0';

  list -> addString(field);

  return list;
}

//
// StringToken skips the empty parts.
//

StringList *StringList::splitAndSkipEmpty(const Char *string, const Char *sep)
{
  if (string == NULL || sep == NULL)
  {
    return NULL;
  }

  StringList *list = new StringList();

  Char str[StringLength(string) + 1];
  Char *field;
  Char *state;

  StringCopy(str, string);

  field = StringToken(str, sep, &state);

  while (field != NULL)
  {
    list -> addString(field);

    field = StringToken(NULL, sep, &state);
  }

  return list;
}

void StringList::merge(Char *&buffer, const Char *sep)
{
  Iterator k;
  const Char *separator;

  separator = (sep == NULL ? "" : sep);

  if (buffer != NULL)
  {
    StringReset(buffer);
  }

  for (Iterator j = getBegin(); j != getEnd(); )
  {
    k = j++;

    if (j == getEnd())
    {
      StringAdd(buffer, getString(k));
    }
    else
    {
      StringAdd(buffer, getString(k), separator);
    }
  }
}

Char *StringList::merge(const Char *sep)
{
  Char *merged = NULL;

  merge(merged, sep);

  return merged;
}

StringList *StringList::copyList()
{
  StringList *copyList = new StringList(exclusive_);

  for (Iterator j = getBegin(); j != getEnd(); j++)
  {
    copyList -> addString(getString(j));
  }

  return copyList;
}


/*
FIXME: Methods below need to be removed/
       reviewed/rewritten.
*/

/*
FIXME: This sort algorythm is quite slow. Maybe
       the List class in libnx should be modified
       to export the STL sort method.
*/
void StringList::sort()
{
  Int j = getSize() - 1;

  while (j > 0)
  {
    for (Int i = 0; i < j; i++)
    {
      if (StringCompare(getString(i), getString(i + 1)) > 0)
      {
        swapString(getIterator(i), getIterator(i + 1));
      }
    }

    j--;
  }
}

void StringList::swapString(Iterator i, Iterator j)
{
  void *temporary = *i;

  *i = *j;

  *j = temporary;
}

/*
FIXME: This method must be moved elsewhere because
       it is not general enough to stay here.
*/
StringList *StringList::sortByItem(int item, const Char *sep)
{
  StringList *list;

  list = split(getBegin(), sep, KeepEmpty);

  if (list -> getIterator(item) == list -> getEnd())
  {
    #ifdef WARNING
    Log() << "StringList: WARNING! The item doesn't exist.\n";
    #endif

    delete list;

    return NULL;
  }

  StringList *listSorted = new StringList();
  StringList *listSortedByItem = new StringList();

  listSorted -> addString(list -> getString(item));
  listSortedByItem -> addString(getFirst());

  delete list;

  for (int i = 1; i < getSize(); i++)
  {
    list = split(getIterator(i), sep, KeepEmpty);

    int pos = listSorted -> insertString(list -> getString(item));

    if ((listSorted -> getIterator(pos) == listSorted -> findSubStringIterator(listSorted -> getLast())))
    {
      listSortedByItem -> addString(getString(i));
    }
    else
    {
      listSortedByItem -> insertString(getString(i), listSortedByItem -> getIterator(pos));
    }

    delete list;
  }

  delete listSorted;

  #ifdef DEBUG
  Log() << "StringList: Printing the sorted list.\n";

  for (int i = 0; i < listSortedByItem -> getSize(); i++)
  {
    Log() << "StringList: " << Stk(listSortedByItem -> getString(i)) << "\n";
  }

  Log() << "StringList: End list.\n";
  #endif

  return listSortedByItem;
}

/*
FIXME: A real insert method should accept the
       position where to insert the string.
       What this method is useful for?
*/
Int StringList::insertString(const Char *string)
{
  if (string == NULL)
  {
    return -1;
  }

  Int index = 0;

  Iterator j = getBegin();

  while (j != getEnd())
  {
    if (StringCompare(string, getString(j)) < 0)
    {
      insertString(string, j);

      return index;
    }

    index++;
    j++;
  }

  addString(string);

  return getSize() - 1;
}

void StringList::moveString(Iterator from, Iterator to)
{
  insertString(getString(from), to);

  removeString(from);
}

/*
FIXME: This method is implicitly wrong, because if
       there are more items with the same substring,
       only the first one will be returned. Maybe
       its name should be changed to something
       reflecting this.
*/
StringList::Iterator StringList::findSubStringIterator(const Char *substring,
                                                           Int *index)
{
  Int position = 0;

  for (Iterator j = getBegin(); j != getEnd(); j++)
  {
    if (StringFind(getString(j), substring) != NULL)
    {
      if (index != NULL)
      {
        *index = position;
      }

      return j;
    }

    position++;
  }

  if (index != NULL)
  {
    *index = -1;
  }

  return getEnd();
}


