/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Log.h"
#include "Error.h"
#include "Random.h"
#include "Signal.h"
#include "Thread.h"

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

//
// Thread creation functions.
//

Int ThreadCreate(ThreadHandle *handle, ThreadId *id,
                     ThreadReturn ThreadConvention (ThreadFunction *function)(void *),
                         void *parameter)
{
  #ifndef WIN32

  pthread_attr_t attributes;

  pthread_attr_init(&attributes);

  pthread_attr_setdetachstate(&attributes, PTHREAD_CREATE_JOINABLE);

  //
  // Set the default signal mask to be used by
  // the child thread. The mask should block
  // all signals handled by System in the main
  // thread. The thread should be interrupted
  // only by a quit signal, when such a signal
  // is explicitly enabled.
  //

  Int result = pthread_create(id, &attributes, function, parameter);

  pthread_attr_destroy(&attributes);

  if (result != 0)
  {
    SetError(result);

    goto ThreadCreateError;
  }

  //
  // The thread handle and the thread
  // id are the same value.
  //

  *handle = *id;

  #else // #ifndef WIN32 ...

  //
  // The parameters are, in order, default security
  // attributes, default stack size, the thread entry
  // point, thread function arguments, default creation
  // flags, thread identifier returned by the create
  // function. Note that the thread identifier is not
  // used.
  //

  *handle = CreateThread(NULL, 0, function, parameter, 0, id);

  if (*handle == NULL )
  {
    goto ThreadCreateError;
  }

  #endif // #ifndef WIN32 ... #else ...

  return 1;

ThreadCreateError:

  #ifdef ERROR
  Log() << "Thread: ERROR! Thread creation failed. "
        << "Error is " << GetError() << " "
        << Stk(GetErrorString()) << ".\n";
  #endif

  *handle = ThreadHandleNone;
  *id     = ThreadIdNone;

  return -1;
}

//
// Thread signaling and cleanup
// functions.
//

Int ThreadRunning(ThreadHandle handle)
{
  #ifndef WIN32

  Int error;

  #ifdef TEST
  Log() << "Thread: Checking thread " << handle
        << ".\n";
  #endif

  if ((error = pthread_kill(handle, 0)) == 0)
  {
    #ifdef TEST
    Log() << "Thread: Thread " << handle << " is "
          << "running.\n";
    #endif

    return 1;
  }
  else if (error == ESRCH)
  {
    #ifdef TEST
    Log() << "Thread: Thread " << handle << " is "
          << "not running.\n";
    #endif

    return 0;
  }

  SetError(error);

  #ifdef ERROR
  Log() << "Thread: ERROR! Thread check failed. "
        << "Error is " << GetError() << " "
        << Stk(GetErrorString()) << ".\n";
  #endif

  #else // #ifndef WIN32 ...

  #ifdef WARNING
  Log() << "Thread: WARNING! Thread signaling not "
        << "available on Windows.\n";
  #endif

  #endif // #ifndef WIN32 ... #else ...

  return -1;
}

Int ThreadSignal(ThreadHandle handle, Int signal)
{
  #ifndef WIN32

  Int error;

  #ifdef TEST
  Log() << "Thread: Signaling thread " << handle
        << " with signal " << signal << " "
        << Sig(signal) << ".\n";
  #endif

  if ((error = pthread_kill(handle, signal)) == 0)
  {
    return 1;
  }

  SetError(error);

  if (error != ESRCH)
  {
    #ifdef ERROR
    Log() << "Thread: ERROR! Thread signal failed. "
          << "Error is " << GetError() << " "
          << Stk(GetErrorString()) << ".\n";
    #endif
  }
  #ifdef TEST
  else
  {
    Log() << "Thread: Thread " << handle << " is "
          << "not running.\n";
  }
  #endif

  #else // #ifndef WIN32 ...

  #ifdef WARNING
  Log() << "Thread: WARNING! Thread signaling not "
        << "available on Windows.\n";
  #endif

  #endif // #ifndef WIN32 ... #else ...

  return -1;
}

Int ThreadJoin(ThreadHandle handle, ThreadReturn *code)
{
  #ifndef WIN32

  Int error;

  if ((error = pthread_join(handle, code)) == 0)
  {
    return 1;
  }

  SetError(error);

  #else // #ifndef WIN32 ...

  for (;;)
  {
    WaitForSingleObject(handle, INFINITE);

    #ifdef TEST
    Log() << "Thread: Wait for the thread completed.\n";
    #endif

    if (GetExitCodeThread(handle, code) != 0)
    {
      if (*code != STILL_ACTIVE)
      {
        CloseHandle(handle);

        return 1;
      }

      #ifdef WARNING
      Log() << "Thread: WARNING! Can't join the thread yet.\n";
      #endif
    }
    else
    {
      CloseHandle(handle);

      break;
    }
  }

  #endif // #ifndef WIN32 ... #else ...

  #ifdef ERROR
  Log() << "Thread: ERROR! Thread join failed. "
        << "Error is " << GetError() << " "
        << Stk(GetErrorString()) << ".\n";
  #endif

  return -1;
}

void ThreadExit(ThreadReturn code)
{
  #ifndef WIN32

  pthread_exit(VoidCast(code));

  #else

  ExitThread(code);

  #endif
}

//
// Thread cancelation functions.
//

void ThreadCancelability(Int type)
{
  #ifndef WIN32

  #ifndef __BIONIC__

  switch (type)
  {
    case ThreadCancelabilityDeferred:
    {
      pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

      break;
    }
    case ThreadCancelabilityImmediate:
    {
      pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

      break;
    }
    default:
    {
      #ifdef ERROR
      Log() << "Thread: ERROR! Invalid thread cancelability "
            << "type " << Itk(type) << ".\n";
      #endif

      break;
    }
  }

  #else // #ifndef __BIONIC__ ...

  #ifdef WARNING
  Log() << "Thread: WARNING! Function ThreadCancelability() "
        << "not implemented.\n";
  #endif

  #endif // #ifndef __BIONIC__ ... #else ...

  #endif // #ifndef WIN32 ...
}

Int ThreadCancel(ThreadHandle handle)
{
  #ifndef WIN32

  #ifdef __BIONIC__

  #ifdef WARNING
  Log() << "Thread: WARNING! Function ThreadCancel() "
        << "not implemented.\n";
  #endif

  SetError(ENOSYS);

  return -1;

  #else // #ifdef __BIONIC__ ...

  //
  // There is no way to set the exit code
  // of a terminated thread.
  //

  Int error;

  if ((error = pthread_cancel(handle)) == 0)
  {
    return 1;
  }
  else if (error == ESRCH)
  {
    #ifdef TEST
    Log() << "Thread: Thread " << handle << " is "
          << "not running.\n";
    #endif

    return 0;
  }

  SetError(error);

  #endif // #ifdef __BIONIC__ ... #else ...

  #else // #ifndef WIN32 ...

  if (TerminateThread(handle, ThreadReturnNone) != 0)
  {
    return 1;
  }

  #endif // #ifndef WIN32 .. #else ...

  #ifdef ERROR
  Log() << "Thread: ERROR! Thread cancel failed. "
        << "Error is " << GetError() << " "
        << Stk(GetErrorString()) << ".\n";
  #endif

  return -1;
}

Int ThreadReset(ThreadHandle handle)
{
  //
  // We presently don't have a way to
  // test if the thread is running on
  // Windows.
  //

  #ifndef WIN32

  if (ThreadRunning(handle) == 0)
  {
    return 0;
  }

  #endif

  #ifdef TEST
  Log() << "Thread: Resetting thread " << handle
        << ".\n";
  #endif

  //
  // On Linux64 the code below gives an
  // invalid read in pthread_cancel.c if
  // executed after a fork, rather than
  // returning an error.
  //

  ThreadCancel(handle);

  return ThreadJoin(handle);
}

//
// Thread utility functions.
//

void ThreadYield()
{
  #ifndef WIN32

  #if defined(__APPLE__) || defined(__sun)

  sched_yield();

  #else // #if defined(__APPLE__) || ...

  #ifndef __BIONIC__

  pthread_yield();

  #else // #ifndef __BIONIC__

  #ifdef WARNING
  Log() << "Thread: WARNING! Function ThreadYield() "
        << "not implemented.\n";
  #endif

  #endif // #ifndef __BIONIC__ ... #else ...

  #endif // // #if defined(__APPLE__) || ... #else ... 

  #else // #ifndef WIN32 ...

  SwitchToThread();

  #endif // #ifndef WIN32 ... #else ...
}

//
// Thread initialization functions.
//

void ThreadInit()
{
  #ifdef DEBUG
  Log() << "Thread: Starting initialization "
        << "of thread " << Tid() << ".\n";
  #endif

  #ifndef WIN32

  //
  // Make the new thread cancelable at
  // any time. Except where this is not
  // possible at all.
  //

  #ifndef __BIONIC__

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  #endif // #ifndef __BIONIC__ ...

  #else // #ifndef WIN32 ...

  //
  // Windows maintains a different seed
  // for each thread.
  //

  RandomInit();

  #endif // #ifndef WIN32 ... #else ...

  #ifdef DEBUG
  Log() << "Thread: Completed initialization "
        << "of thread " << Tid() << ".\n";
  #endif
}
