/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Thread_H
#define Thread_H

#ifndef WIN32

#include <sys/types.h>

#ifdef __linux__
#include <sys/syscall.h>
#endif

#include <unistd.h>
#include <pthread.h>

#else

#ifdef ERROR
#undef ERROR
#endif

#endif

#include "Types.h"

//
// Make the definitions independent
// from the platform.
//

#ifndef WIN32

typedef pthread_t ThreadHandle;
typedef pthread_t ThreadId;

typedef void *    ThreadReturn;

#if defined(__x86_64__) || defined(__aarch64__)
#define ThreadReturnCast(value) ((ThreadReturn) ((Long) (value)))
#else
#define ThreadReturnCast(value) ((ThreadReturn) ((Int) (value)))
#endif

#define ThreadFunction
#define ThreadConvention

#else // #ifndef WIN32 ...

typedef HANDLE ThreadHandle;
typedef DWORD  ThreadId;

typedef DWORD  ThreadReturn;

#define ThreadReturnCast(value) ((ThreadReturn) (value))

#ifndef _MSC_VER

#define ThreadFunction
#define ThreadConvention WINAPI

#else // #ifndef _MSC_VER ...

#define ThreadFunction   WINAPI
#define ThreadConvention

#ifndef _PID_T_
#define _PID_T_
typedef int pid_t;
#endif

#endif // #ifndef _MSC_VER ... #else ...

#endif // #ifndef WIN32 ... #else ...

#define ThreadPid        pid_t

#define ThreadReturnNone ((ThreadReturn) -1)
#define ThreadHandleNone ((ThreadHandle) -1)
#define ThreadIdNone     ((ThreadId)  -1)
#define ThreadPidNone    ((ThreadPid) -1)

//
// Thread cancelability options.
//

#define ThreadCancelabilityDeferred   0
#define ThreadCancelabilityImmediate  1

//
// Set a default context used to
// abort the thread.
//

#define ThreadContext() \
\
  SaveContext(*GetSystem() -> getContextData())

#define ThreadBegin() \
\
  if (ThreadContext() != 0) \
  { \
    GetSystem() -> abort(GetError()); \
  } \
\
  ThreadInit();

#define ThreadEnd(result) \
\
  return ((ThreadReturn) (result));

//
// Thread creation functions.
//

Int ThreadCreate(ThreadHandle *handle, ThreadId *id,
                     ThreadReturn ThreadConvention (ThreadFunction *function)(void *),
                             void *parameter);

//
// Thread signaling and cleanup
// functions.
//

Int ThreadRunning(ThreadHandle handle);

Int ThreadSignal(ThreadHandle handle, Int signal);

Int ThreadJoin(ThreadHandle handle, ThreadReturn *code);

inline Int ThreadJoin(ThreadHandle handle)
{
  ThreadReturn code;

  return ThreadJoin(handle, &code);
}

void ThreadExit(ThreadReturn code);

//
// Thread cancelation functions.
//

void ThreadCancelability(Int type);

Int ThreadCancel(ThreadHandle handle);

//
// Cancel the thread if it is running.
//

Int ThreadReset(ThreadHandle handle);

//
// Tell the OS to yield to another
// thread.
//

void ThreadYield();

//
// Sleep in the current thread for
// the given amount of milliseconds.
//

#ifndef WIN32

#define ThreadSleep(ms)  usleep((ms) * 1000)

#else

#define ThreadSleep(ms)  Sleep(ms)

#endif

//
// Install the default thread signal mask.
// This will block all signals handled by
// the main thread. A child thread should
// explicitly enable the quit signal when
// an interrupt is required to terminate
// a wait.
//

void ThreadInit();

//
// Get the current thread id and,
// where available, the thread pid.
//

inline ThreadId ThreadGetId()
{
  #ifndef WIN32

  return pthread_self();

  #else

  return GetCurrentThreadId();

  #endif
}

inline Int ThreadGetPid()
{
  #ifndef WIN32
  
  #if defined(__linux__) && defined(SYS_gettid)

  return (Int) syscall(SYS_gettid);

  #else

  //
  // The thread id is actually a pointer
  // to a structure. Force it to resemble
  // a small integer.
  //

  return (IntCast(ThreadGetId()) & 0xffff);

  #endif

  #else

  //
  // On Windows use the system id.
  //

  return (Int) ThreadGetId();

  #endif
}

inline Int ThreadRaise(Int signal)
{
  #ifndef WIN32

  return ThreadSignal(ThreadGetId(), signal);

  #else

  return ThreadSignal(ThreadHandleNone, signal);

  #endif
}

//
// Print the current thread pid.
//

#define Thr() ThreadGetId()
#define Tid() ThreadGetPid()

#endif /* Thread_H */
