/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Time_H
#define Time_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>

#ifndef _MSC_VER

#include <sys/time.h>

#endif

#include "Log.h"

//
// Type used for timeout manipulation.
//

typedef struct timeval Timeval;

//
// Use nanoseconds instead of microseconds.
// This type is mainly required by pthread
// and is not present on Windows.
//

#ifndef WIN32

typedef struct timespec Timespec;

#endif

//
// A Win32 API implementation of the POSIX
// gettimeofday() function.
//

#ifdef _MSC_VER

struct timezone
{
  Int  tz_minuteswest;
  Int  tz_dsttime;
};

Int gettimeofday(struct timeval *tv, struct timezone *tz);

#endif

//
// Return the current timestamp taken from
// the system.
//

inline Timeval getTimeval()
{
  Timeval ts;

  gettimeofday(&ts, NULL);

  return ts;
}

//
// A 0 value on both seconds and microseconds
// fields means that timestamp is invalid or
// not set.
//

inline Timeval nullTimeval()
{
  Timeval ts;

  ts.tv_sec  = 0;
  ts.tv_usec = 0;

  return ts;
}

inline Int isTimeval(const Timeval &ts)
{
  if (ts.tv_sec == 0 && ts.tv_usec == 0)
  {
     return 0;
  }

  return 1;
}

//
// Return the difference in milliseconds
// between the two timestamps.
//

inline Long diffMsTimeval(const Timeval &ts1, const Timeval &ts2)
{
  //
  // Add 500 microseconds to round up
  // to the nearest millisecond.
  //

  return ((ts2.tv_sec * 1000000 + ts2.tv_usec + 500) -
              (ts1.tv_sec * 1000000 + ts1.tv_usec + 500)) / 1000;
}

//
// Return a timestamp corresponding to
// the given amount of milliseconds.
//

inline Timeval getTimeval(Long ms)
{
  struct timeval ts;

  ts.tv_sec  = ms / 1000;
  ts.tv_usec = (ms % 1000) * 1000;

  return ts;
}

//
// Calculate the time elapsed since
// the timestamp.
//

inline Long diffMsTimeval(const Timeval &ts)
{
  return diffMsTimeval(ts, getTimeval());
}

//
// The same in microseconds. It doesn't
// round the value.
//

inline Long diffUsTimeval(const Timeval &ts1, const Timeval &ts2)
{
  return (ts2.tv_sec * 1000000 + ts2.tv_usec) -
              (ts1.tv_sec * 1000000 + ts1.tv_usec);
}

inline Long diffUsTimeval(const Timeval &ts)
{
  return diffUsTimeval(ts, getTimeval());
}

inline Timeval setMsTimeval(Long ms)
{
  if (ms <= 0)
  {
    return nullTimeval();
  }
  else
  {
    Timeval ts;

    ts.tv_sec  = ms / 1000;
    ts.tv_usec = (ms % 1000) * 1000;

    return ts;
  }
}

inline void setMsTimeval(Timeval &ts, Long ms)
{
  if (ms <= 0)
  {
    ts.tv_sec  = 0;
    ts.tv_usec = 0;
  }
  else
  {
    ts.tv_sec  = ms / 1000;
    ts.tv_usec = (ms % 1000) * 1000;
  }
}

inline Timeval setUsTimeval(Long us)
{
  if (us <= 0)
  {
    return nullTimeval();
  }
  else
  {
    Timeval ts;

    ts.tv_sec  = us / 1000000;
    ts.tv_usec = us % 1000000;

    return ts;
  }
}

inline void setUsTimeval(Timeval &ts, Long us)
{
  if (us <= 0)
  {
    ts.tv_sec  = 0;
    ts.tv_usec = 0;
  }
  else
  {
    ts.tv_sec  = us / 1000000;
    ts.tv_usec = us % 1000000;
  }
}

//
// Return the smaller between two timestamps.
//

inline Timeval &setMinTimeval(Timeval &ts, Long ms)
{
  if ((ts.tv_sec * 1000 + ts.tv_usec / 1000) > ms)
  {
    ts.tv_sec  = ms / 1000;
    ts.tv_usec = (ms % 1000) * 1000;
  }

  return ts;
}

inline Timeval &setMinTimeval(Timeval &ts1, Timeval &ts2)
{
  if ((ts1.tv_sec * 1000000 + ts1.tv_usec) >
           (ts2.tv_sec * 1000000 + ts2.tv_usec))
  {
    ts1.tv_sec  = ts2.tv_sec;
    ts1.tv_usec = ts2.tv_usec;
  }

  return ts1;
}

//
// Convert a timestamp in the total number
// of milliseconds.
//

inline Long getMsTimeval(const Timeval &ts)
{
  return ts.tv_sec * 1000 + ts.tv_usec / 1000;
}

//
// Add or subtract a number of milliseconds
// from a timestamp.
//

inline void addMsTimeval(Timeval &ts, Long ms)
{
  ts.tv_sec  += ms / 1000;
  ts.tv_usec += (ms % 1000) * 1000;

  if (ts.tv_usec >= 1000000)
  {
    ts.tv_sec++;
    ts.tv_usec -= 1000000;
  }
}

inline void subMsTimeval(Timeval &ts, Long ms)
{
  ts.tv_sec -= ms / 1000;

  if (ts.tv_sec < 0)
  {
    ts.tv_sec  = 0;
    ts.tv_usec = 0;
  }
  else
  {
    ts.tv_usec -= (ms % 1000) * 1000;

    if (ts.tv_usec < 0)
    {
      ts.tv_usec = 0;
    }
  }
}

//
// The same with microseconds.
//

inline void addUsTimeval(Timeval &ts, Long us)
{
  ts.tv_sec  += us / 1000000;
  ts.tv_usec += (us % 1000000) * 1000;

  if (ts.tv_usec >= 1000000)
  {
    ts.tv_usec -= 1000000;
    ts.tv_sec++;
  }
}

inline void subUsTimeval(Timeval &ts, Long us)
{
  ts.tv_sec -= us / 1000000;

  if (ts.tv_sec < 0)
  {
    ts.tv_sec  = 0;
    ts.tv_usec = 0;
  }
  else
  {
    ts.tv_usec -= (us % 1000000) * 1000000;

    if (ts.tv_usec < 0)
    {
      ts.tv_usec = 0;
    }
  }
}

//
// Check the difference between timestamps.
// Return 0 if the system time went backward
// compared to the second timestamp, or the
// difference between the timestamps exceeds
// the given number of milliseconds.
//

inline Int checkDiffTimeval(const Timeval &ts1, const Timeval &ts2,
                                  Long drift = 86400000)
{
  Long diffTs = diffMsTimeval(ts1, ts2);

  if (diffTs < 0 || diffTs > drift)
  {
    return 0;
  }

  return 1;
}

//
// Return a string representing the timestamp.
//

const Char *strTimeval(const Timeval &ts);
const Char *strMsTimeval(const Timeval &ts);

inline const Char *strTimeval()
{
  return strTimeval(getTimeval());
}

inline const Char *strMsTimeval()
{
  return strMsTimeval(getTimeval());
}

inline const Char *strMsTimeval(const Timeval *ts)
{
  return strMsTimeval(*ts);
}

//
// Manipulate the timespec structure.
//

#ifndef WIN32

inline Timespec nullTimespec()
{
  Timespec ts;

  ts.tv_sec  = 0;
  ts.tv_nsec = 0;

  return ts;
}

inline Int isTimespec(const Timespec &ts)
{
  if (ts.tv_sec == 0 && ts.tv_nsec == 0)
  {
    return 0;
  }

  return 1;
}

inline Timespec getTimespec()
{
  Timeval tv = getTimeval();

  Timespec ts;

  ts.tv_sec  = tv.tv_sec;
  ts.tv_nsec = tv.tv_usec * 1000;

  return ts;
}

inline Timespec getTimespec(const Timeval &tv)
{
  Timespec ts;

  ts.tv_sec  = tv.tv_sec;
  ts.tv_nsec = tv.tv_usec * 1000;

  return ts;
}

inline Timespec setMsTimespec(Long ms)
{
  if (ms <= 0)
  {
    return nullTimespec();
  }
  else
  {
    Timespec ts;

    ts.tv_sec  = ms / 1000;
    ts.tv_nsec = (ms % 1000) * 1000000;

    return ts;
  }
}

inline void setMsTimespec(Timespec &ts, Long ms)
{
  if (ms <= 0)
  {
    ts.tv_sec  = 0;
    ts.tv_nsec = 0;
  }
  else
  {
    ts.tv_sec  = ms / 1000;
    ts.tv_nsec = (ms % 1000) * 1000000;
  }
}

inline void addMsTimespec(Timespec &ts, Long ms)
{
  ts.tv_sec  += ms / 1000;
  ts.tv_nsec += (ms % 1000) * 1000000;

  if (ts.tv_nsec >= 1000000000)
  {
    ts.tv_sec++;
    ts.tv_nsec -= 1000000000;
  }
}

//
// Convert from a Timespec structure.
//

inline Timeval getTimeval(const Timespec &ts)
{
  Timeval tv;

  tv.tv_sec  = ts.tv_sec;
  tv.tv_usec = ts.tv_nsec / 1000;

  return tv;
}

#endif // #ifndef WIN32 ...

//
// Print a formatted timestamp.
//

#define Stn()    strTimeval()
#define Stm(ts)  strMsTimeval(ts)

#endif /* Time_H */
