/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* The NX library and any derived work however based on this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for further options.     */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef Types_H
#define Types_H

#include <sys/types.h>
#include <sys/cdefs.h>

#include <limits.h>
#include <string.h>

#include <iostream>
#include <fstream>

using namespace std;

//
// If defined, use custom allocators
// for new and delete calling malloc
// and free.
//

#undef TYPES_USE_MALLOC

//
// Include the Windows header early
// to avoid conficts.
//

#ifdef WIN32

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#ifdef  WINVER
#undef  WINVER
#endif

#define WINVER 0x501

#include <winsock2.h>
#include <windows.h>

#ifdef ERROR
#undef ERROR
#endif

#endif

//
// Optional attributes for variables
// and function declarations.
//

#ifdef WIN32

#ifdef ERROR
#undef ERROR
#endif

#define IMPORT   __declspec(dllimport)
#define EXPORT   __declspec(dllexport)

#define NORETURN __declspec(noreturn)
#define NOINLINE __declspec(noinline)

#define WEAK

#else // #ifdef WIN32 ...

#define IMPORT
#define EXPORT

#define NORETURN __attribute__((noreturn))
#define NOINLINE __attribute__((noinline))

#define WEAK     __attribute__((weak))

#endif // #ifdef WIN32 ... #else ...

//
// Determine whether the compiler is
// run with C++ exceptions enabled.
//

#ifdef EXCEPTIONS
#undef EXCEPTIONS
#endif

#ifdef _MSC_VER

#ifdef  _CPPUNWIND
#define EXCEPTIONS
#endif

#else // #ifdef _MSC_VER ...

#ifdef  __EXCEPTIONS
#define EXCEPTIONS
#endif

#endif // #ifdef _MSC_VER ... #else ...

//
// Attributes for global variables.
//

#define SYSTEM 

//
// Base types.
//

#define Char      char
#define Int       int
#define Short     short
#define Long      long

#define Float     float
#define Double    double

#define Size      size_t
#define Access    mode_t
#define Offset    off_t

#ifdef WIN32

#define Uid       int
#define Gid       int

#else // #ifdef WIN32 ...

#define Uid       uid_t
#define Gid       gid_t

#endif // #ifdef WIN32 ... #else ...

#define Unsigned  unsigned

#if defined(WIN32) || defined(__sun)

#include <stdint.h>

#define Int8  int8_t
#define Int16 int16_t
#define Int32 int32_t
#define Int64 int64_t

#else // #if defined(WIN32) || defined(__sun) ...

//
// Find Int8.
//
 
#if defined(USHRT_CHAR_MAX) && (USHRT_CHAR_MAX == 0xff)
#define Int8 short char
#elif defined(UCHAR_MAX) && (UCHAR_MAX == 0xff)
#define Int8 char
#elif defined(USHRT_SHRT_MAX) && (USHRT_SHRT_MAX == 0xff)
#define Int8 short short int
#elif defined(USHRT_MAX) && (USHRT_MAX == 0xff)
#define Int8 short int
#elif defined(UINT_MAX) && (UINT_MAX == 0xff)
#define Int8 int
#else
#error "ERROR: Can't find the Int8 type."
#endif

//
// Find Int16.
//

#if defined(USHRT_CHAR_MAX) && (USHRT_CHAR_MAX == 0xffff)
#define Int16 short char
#elif defined(UCHAR_MAX) && (UCHAR_MAX == 0xffff)
#define Int16 char
#elif defined(USHRT_SHRT_MAX) && (USHRT_SHRT_MAX == 0xffff)
#define Int16 short short int
#elif defined(USHRT_MAX) && (USHRT_MAX == 0xffff)
#define Int16 short int
#elif defined(UINT_MAX) && (UINT_MAX == 0xffff)
#define Int16 int
#else
#error "ERROR: Can't find the Int16 type."
#endif

//
// Find Int32.
//

#if defined(USHRT_SHRT_MAX) && (USHRT_SHRT_MAX == 0xffffffff)
#define Int32 short short int
#elif defined(USHRT_MAX) && (USHRT_MAX == 0xffffffff)
#define Int32 short int
#elif defined(UINT_MAX) && (UINT_MAX == 0xffffffff)
#define Int32 int
#elif defined(ULONG_MAX) && (ULONG_MAX == 0xffffffff)
#define Int32 long int
#else
#error "ERROR: Can't find the Int32 type."
#endif

//
// Find Int64.
//

#if defined(UINT_MAX) && (UINT_MAX == 0xffffffffffffffff)
#define Int64 int
#elif defined(ULONG_MAX) && (ULONG_MAX == 0xffffffffffffffff)
#define Int64 long int
#elif defined(ULONG_LONG_MAX) && (ULONG_LONG_MAX == 0xffffffffffffffff)
#define Int64 long long int
#else
#error "ERROR: Can't find the Int64 type."
#endif

#endif // #if defined(WIN32) || defined(__sun) ... #else ...

//
// Static cast utilities.
//

#if defined(__x86_64__) || defined(__aarch64__)
#define IntCast(value)         ((Int)          ((Long Long)  (value)))
#define UnsignedIntCast(value) ((Unsigned Int) ((Unsigned Long Long) (value)))
#else
#define IntCast(value)         ((Int) (value))
#define UnsignedIntCast(value) ((Unsigned Int) (value))
#endif

#define UnsignedCast(value) UnsignedIntCast((value))

#define LongCast(value)     ((Long) (value))
#define LongLongCast(value) ((Long Long) (value))

#define UnsignedLongCast(value)     ((Unsigned Long) (value))
#define UnsignedLongLongCast(value) ((Unsigned Long Long) (value))

#define CharCast(value)       ((Char) (value))
#define StringCast(value)     ((Char *) (value))

#if defined(__x86_64__) || defined(__aarch64__)
#define VoidCast(value)           ((void *) ((Long Long) (value)))
#define VoidReferenceCast(value)  ((void *&) ((Long Long &) (value)))
#else
#define VoidCast(value)           ((void *) ((Int) (value)))
#define VoidReferenceCast(value)  ((void *&) ((Int &) (value)))
#endif

#define DoubleCast(value)   ((Double) (value))

#ifdef WIN32
#define HandleCast(value)   ((HANDLE) ((Long Long) (value)))
#else
#define HandleCast(value)   ((void *) (value))
#endif

#define VoidFunctionCast(value) ((void (*)(void)) (value))
#define IntFunctionCast(value) ((void (*)(Int)) (value))

//
// Suppress a warning about an unused
// function parameter.
//

#define UnusedCast(value)  ((void) value)

//
// Force a const object pointer to be
// non-const and viceversa.

#define NonConstCast(type, value)  ((type) value)
#define ConstCast(type, value)     ((const type) value)

//
// Static object cast utilities.
//

#if defined(__x86_64__) || defined(__aarch64__)

#define ObjectCast(value) ((Object *) ((Long) (value)))
#define WidgetCast(value) ((Widget *) ((Long) (value)))
#define TimerCast(value)  ((Timer *)  ((Long) (value)))

#define BufferCast(value)     ((Buffer *)     ((Long) (value)))
#define BufferListCast(value) ((BufferList *) ((Long) (value)))

#define RunnableCast(value)   ((Runnable *)   ((Long) (value)))
#define WaitableCast(value)   ((Waitable *)   ((Long) (value)))
#define ThreadableCast(value) ((Threadable *) ((Long) (value)))

#define SchedulerCast(value)  ((Scheduler *)  ((Long) (value)))
#define DispatcherCast(value) ((Dispatcher *) ((Long) (value)))

#define AlarmableCast(value)     ((Alarmable *)     ((Long) (value)))
#define InterruptibleCast(value) ((Interruptible *) ((Long) (value)))

#else

#define ObjectCast(value) ((Object *) ((Int) (value)))
#define WidgetCast(value) ((Widget *) ((Int) (value)))
#define TimerCast(value)  ((Timer *)  ((Int) (value)))

#define BufferCast(value)     ((Buffer *)     ((Int) (value)))
#define BufferListCast(value) ((BufferList *) ((Int) (value)))

#define RunnableCast(value)   ((Runnable *)   ((Int) (value)))
#define WaitableCast(value)   ((Waitable *)   ((Int) (value)))
#define ThreadableCast(value) ((Threadable *) ((Int) (value)))

#define SchedulerCast(value)  ((Scheduler *)  ((Int) (value)))
#define DispatcherCast(value) ((Dispatcher *) ((Int) (value)))

#define AlarmableCast(value)     ((Alarmable *)     ((Int) (value)))
#define InterruptibleCast(value) ((Interruptible *) ((Int) (value)))

#endif

//
// Boolean utilities.
//

#define Not(value) ((value) == 0)

//
// Standard C++ streams.
//

#define InputOutputStream  iostream

#define InputStream        istream
#define OutputStream       ostream

#define StreamBuffer       streambuf

#define InputStreamBuffer  streambuf
#define OutputStreamBuffer streambuf

#define InputFileStream    ifstream
#define OutputFileStream   ofstream

#define FileStreamMode     ios::openmode

#define FileStreamIn       ios::in
#define FileStreamOut      ios::out
#define FileStreamAppend   ios::app
#define FileStreamTruncate ios::trunc
#define FileStreamBinary   ios::binary

//
// Overloaded new and delete operators.
//

#ifdef TYPES_USE_MALLOC

void *operator new(Size size) throw(bad_alloc);
void *operator new[](Size size) throw(bad_alloc);

void operator delete(void *data) throw();
void operator delete[](void *data) throw();

#endif

#endif /* Types_H */
